﻿#region Header
// --------------------------------------------------------------------------
// OBD Tools
// ==========================================================================
//
// Support for automotive On-board diagnostics (OBD).
//
// ==========================================================================
// <copyright file="DtcDatabase.cs" company="Tethys">
// Copyright  2014 by Thomas Graf
//            All rights reserved.
//            Licensed under the Apache License, Version 2.0.
//            Unless required by applicable law or agreed to in writing, 
//            software distributed under the License is distributed on an
//            "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
//            either express or implied. 
// </copyright>
// 
// System ... Portable Library
// Tools .... Microsoft Visual Studio 2013
//
// ---------------------------------------------------------------------------
#endregion

namespace Tethys.OBD
{
    using System.Collections.Generic;

    /// <summary>
    /// A database for diagnostic trouble codes (DTC).
    /// </summary>
    public class DtcDatabase
    {
        /// <summary>
        /// The DTC-description dictionary.
        /// </summary>
        private static Dictionary<string, string> dtcDictionary;

        /// <summary>
        /// Initializes the DTC descriptions.
        /// </summary>
        private static void InitDtcDescriptions()
        {
            dtcDictionary = new Dictionary<string, string>(1000);

            // Related to: Fuel and Air Metering
            dtcDictionary.Add("P0001", "Fuel Volume Regulator Control Circuit/Open");
            dtcDictionary.Add("P0002", "Fuel Volume Regulator Control Circuit Range/Performance");
            dtcDictionary.Add("P0003", "Fuel Volume Regulator Control Circuit Low");
            dtcDictionary.Add("P0004", "Fuel Volume Regulator Control Circuit High");
            dtcDictionary.Add("P0005", "Fuel Shutoff Valve Control Circuit/Open");
            dtcDictionary.Add("P0006", "Fuel Shutoff Valve Control Circuit Low");
            dtcDictionary.Add("P0007", "Fuel Shutoff Valve Control Circuit High");
            dtcDictionary.Add("P0008", "Engine Position System Performance (Bank1)");
            dtcDictionary.Add("P0009", "Engine Position System Performance (Bank2)");
            dtcDictionary.Add("P0010", "A Camshaft Position Actuator Circuit (Bank1)");
            dtcDictionary.Add("P0011", "A Camshaft Position - Timing Over-Advanced or System Performance (Bank 1)");
            dtcDictionary.Add("P0012", "A Camshaft Position - Timing Over-Retarded (Bank 1)");
            dtcDictionary.Add("P0013", "B Camshaft Position - Actuator Circuit (Bank1)");
            dtcDictionary.Add("P0014", "B Camshaft Position - Timing Over-Advanced or Sytem Performance (Bank 1)");
            dtcDictionary.Add("P0015", "B Camshaft Position - Timing Over-Retarded (Bank 1)");
            dtcDictionary.Add("P0016", "Crankshaft Position Camshaft Position Correlation (Bank 1 Sensor A)");
            dtcDictionary.Add("P0017", "Crankshaft Position Camshaft Position Correlation (Bank 1 Sensor B)");
            dtcDictionary.Add("P0018", "Crankshaft Position Camshaft Position Correlation (Bank 2 Sensor A)");
            dtcDictionary.Add("P0019", "Crankshaft Position Camshaft Position Correlation (Bank 2 Sensor B)");
            dtcDictionary.Add("P0020", "A Camshaft Position Actuator Circuit (Bank 2)");
            dtcDictionary.Add("P0021", "A Camshaft Position - Timing Over-Advanced or System Performance (Bank 2)");
            dtcDictionary.Add("P0022", "A Camshaft Position - Timing Over-Retarded (Bank 2)");
            dtcDictionary.Add("P0023", "B Camshaft Position - Actuator Circuit (Bank 2)");
            dtcDictionary.Add("P0024", "B Camshaft Position - Timing Over-Advanced or System Performance (Bank 2)");
            dtcDictionary.Add("P0025", "B Camshaft Position - Timing Over-Retarded (Bank 2)");
            dtcDictionary.Add("P0026", "Intake Valve Control Solenoid Circuit Range/Performance (Bank 1)");
            dtcDictionary.Add("P0027", "Exhaust Valve Control Solenoid Circuit Range/Performance (Bank 1)");
            dtcDictionary.Add("P0028", "Intake Valve Control Solenoid Circuit Range/Performance (Bank 2)");
            dtcDictionary.Add("P0029", "Exhaust Valve Control Solenoid Circuit Range/Performance (Bank 2)");
            dtcDictionary.Add("P0030", "HO2S Heater Control Circuit (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0031", "HO2S Heater Control Circuit Low (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0032", "HO2S Heater Control Circuit High (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0033", "Turbo Charger Bypass Valve Control Circuit");
            dtcDictionary.Add("P0034", "Turbo Charger Bypass Valve Control Circuit Low");
            dtcDictionary.Add("P0035", "Turbo Charger Bypass Valve Control Circuit High");
            dtcDictionary.Add("P0036", "HO2S Heater Control C ircuit (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0037", "HO2S Heater Control Circuit Low (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0038", "HO2S Heater Control Circuit High (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0039", "Turbo/Super Charger Bypass Valve Control Circuit Range/Performance");
            dtcDictionary.Add("P0040", "O2 Sensor Signals Swapped Bank 1 Sensor 1/ Bank 2 Sensor 1");
            dtcDictionary.Add("P0041", "O2 Sensor Signals Swapped Bank 1 Sensor 2/ Bank 2 Sensor 2");
            dtcDictionary.Add("P0042", "HO2S Heater Control Circuit (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0043", "HO2S Heater Control Circuit Low (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0044", "HO2S Heater Control Circuit High (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0045", "Turbo/Super Charger Boost Control Solenoid Circuit/Open");
            dtcDictionary.Add("P0046", "Turbo/Super Charger Boost Control Solenoid Circuit Range/Performance");
            dtcDictionary.Add("P0047", "Turbo/Super Charger Boost Control Solenoid Circuit Low");
            dtcDictionary.Add("P0048", "Turbo/Super Charger Boost Control Solenoid Circuit High");
            dtcDictionary.Add("P0049", "Turbo/Super Charger Turbine Overspeed");
            dtcDictionary.Add("P0050", "HO2S Heater Control Circuit (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0051", "HO2S Heater Control Circuit Low (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0052", "HO2S Heater Control Circuit High (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0053", "HO2S Heater Resistance (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0054", "HO2S Heater Resistance (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0055", "HO2S Heater Resistance (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0056", "HO2S Heater Control Circuit (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0057", "HO2S Heater Control Circuit Low (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0058", "HO2S Heater Control Circuit High (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0059", "HO2S Heater Resistance (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0060", "HO2S Heater Resistance (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0061", "HO2S Heater Resistance (Bank 2 Sensor 3)");
            dtcDictionary.Add("P0062", "HO2S Heater Control Circuit (Bank 2 Sensor 3)");
            dtcDictionary.Add("P0063", "HO2S Heater Control Circuit Low (Bank 2 Sensor 3)");
            dtcDictionary.Add("P0064", "HO2S Heater Control Circuit High (Bank 2 Sensor 3)");
            dtcDictionary.Add("P0065", "Air Assisted Injector Control Range/Performance");
            dtcDictionary.Add("P0066", "Air Assisted Injector Control Circuit or Circuit Low");
            dtcDictionary.Add("P0067", "Air Assisted Injector Control Circuit High");
            dtcDictionary.Add("P0068", "MAP/MAF Throttle Position Correlation");
            dtcDictionary.Add("P0069", "Manifold Absolute Pressure Barometric Pressure Correlation");
            dtcDictionary.Add("P0070", "Ambient Air Temperature Sensor Circuit");
            dtcDictionary.Add("P0071", "Ambient Air Temperature Sensor Range/Performance");
            dtcDictionary.Add("P0072", "Ambient Air Temperature Sensor Circuit Low");
            dtcDictionary.Add("P0073", "Ambient Air Temperature Sensor Circuit High");
            dtcDictionary.Add("P0074", "Ambient Air Temperature Sensor Circuit Intermittent");
            dtcDictionary.Add("P0075", "Intake Valve Control Solenoid Circuit (Bank 1)");
            dtcDictionary.Add("P0076", "Intake Valve Control Solenoid Circuit Low (Bank 1)");
            dtcDictionary.Add("P0077", "Intake Valve Control Solenoid Circuit High (Bank 1)");
            dtcDictionary.Add("P0078", "Exhaust Valve Control Solenoid Circuit (Bank 1)");
            dtcDictionary.Add("P0079", "Exhaust Valve Control Solenoid Circuit Low (Bank 1)");
            dtcDictionary.Add("P0080", "Exhaust Valve Control Solenoid C ircuit High (Bank 1)");
            dtcDictionary.Add("P0081", "Intake Valve Control Solenoid Circuit (Bank 2)");
            dtcDictionary.Add("P0082", "Intake Valve Control Solenoid Circuit Low (Bank 2)");
            dtcDictionary.Add("P0083", "Intake Valve Control Solenoid Circuit High (Bank 2)");
            dtcDictionary.Add("P0084", "Exhaust Valve Control Solenoid Circuit (Bank 2)");
            dtcDictionary.Add("P0085", "Exhaust Valve Control Solenoid Circuit Low (Bank 2)");
            dtcDictionary.Add("P0086", "Exhaust Valve Control Solenoid C ircuit High (Bank 2)");
            dtcDictionary.Add("P0087", "Fuel Rail/System Pressure - Too Low");
            dtcDictionary.Add("P0088", "Fuel Rail/System Pressure - Too High");
            dtcDictionary.Add("P0089", "Fuel Pressure Regulator 1 Performance");
            dtcDictionary.Add("P0090", "Fuel Pressure Regulator 1 Control Circuit");
            dtcDictionary.Add("P0091", "Fuel Pressure Regulator 1 Control Circuit Low");
            dtcDictionary.Add("P0092", "Fuel Pressure Regulator 1 Control Circuit High");
            dtcDictionary.Add("P0093", "Fuel System Leak Detected Large Leak");
            dtcDictionary.Add("P0094", "Fuel System Leak Detected Small Leak");
            dtcDictionary.Add("P0095", "Intake Air Temperature Sensor 2 Circuit");
            dtcDictionary.Add("P0096", "Intake Air Temperature Sensor 2 Circuit Range/Performance");
            dtcDictionary.Add("P0097", "Intake Air Temperature Sensor 2 Circuit Low");
            dtcDictionary.Add("P0098", "Intake Air Temperature Sensor 2 Circuit High");
            dtcDictionary.Add("P0099", "Intake Air Temperature Sensor 2 Circuit Intermittent/Erratic");
            dtcDictionary.Add("P0100", "Mass or Volume Air flow Circuit Malfunction");
            dtcDictionary.Add("P0101", "Mass or Volume Air flow Circuit Range/Performance Problem");
            dtcDictionary.Add("P0102", "Mass or Volume Air Flow Circuit low Input");
            dtcDictionary.Add("P0103", "Mass or Volume Air flow Circuit High Input");
            dtcDictionary.Add("P0104", "Mass or Volume Air flow Circuit Intermittent");
            dtcDictionary.Add("P0105", "Manifold Absolute Pressure/Barometric Pressure Circuit Malfunction");
            dtcDictionary.Add(
                "P0106",
                "Manifold Absolute Pressure/Barometric Pressure Circuit Range/Performance Problem");
            dtcDictionary.Add("P0107", "Manifold Absolute Pressure/Barometric Pressure Circuit Low Input");
            dtcDictionary.Add("P0108", "Manifold Absolute Pressure/Barometric Pressure Circuit High Input");
            dtcDictionary.Add("P0109", "Manifold Absolute Pressure/Barometric Pressure Circuit Intermittent");
            dtcDictionary.Add("P0110", "Intake Air Temperature Circuit Malfunction");
            dtcDictionary.Add("P0111", "Intake Air Temperature Circuit Range/Performance Problem");
            dtcDictionary.Add("P0112", "Intake Air Temperature Circuit Low Input");
            dtcDictionary.Add("P0113", "Intake Air Temperature Circuit High Input");
            dtcDictionary.Add("P0114", "Intake Air Temperature Circuit Intermittent");
            dtcDictionary.Add("P0115", "Engine Coolant Temperature Circuit Malfunction");
            dtcDictionary.Add("P0116", "Engine Coolant Temperature Circuit Range/Performance Problem");
            dtcDictionary.Add("P0117", "Engine Coolant Temperature Circuit Low Input");
            dtcDictionary.Add("P0118", "Engine Coolant Temperature Circuit High Input");
            dtcDictionary.Add("P0119", "Engine Coolant Temperature Circuit Intermittent");
            dtcDictionary.Add("P0120", "Throttle Pedal Position Sensor/Switch A Circuit Malfunction");
            dtcDictionary.Add("P0121", "Throttle/Pedal Position Sensor/Switch A Circuit Range/Performance Problem");
            dtcDictionary.Add("P0122", "Throttle/Pedal Position Sensor/Switch A Circuit Low Input");
            dtcDictionary.Add("P0123", "Throttle/Pedal Position Sensor/Switch A Circuit High Input");
            dtcDictionary.Add("P0124", "Throttle/Pedal Position Sensor/Switch A Circuit Intermittent");
            dtcDictionary.Add("P0125", "Insufficient Coolant Temperature for Closed Loop Fuel Control");
            dtcDictionary.Add("P0126", "Insufficient Coolant Temperature for Stable Operation");
            dtcDictionary.Add("P0130", "O2 Sensor Circuit Malfunction (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0131", "O2 Sensor Circuit Low Voltage (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0132", "O2 Sensor Circuit High Voltage (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0133", "O2 Sensor Circuit Slow Response (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0134", "O2 Sensor Circuit No Activity Detected (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0135", "O2 Sensor Heater Circuit Malfunction (Bank 1 Sensor 1)");
            dtcDictionary.Add("P0136", "O2 Sensor Circuit Malfunction (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0137", "O2 Sensor Circuit Low Voltage (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0138", "O2 Sensor Circuit High Voltage (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0139", "O2 Sensor Circuit Slow Response (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0140", "O2 Sensor Circuit No Activity Detected (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0141", "O2 Sensor Heater Circuit Malfunction (Bank 1 Sensor 2)");
            dtcDictionary.Add("P0142", "O2 Sensor Circuit Malfunction (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0143", "O2 Sensor Circuit Low Voltage (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0144", "O2 Sensor Circuit High Voltage (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0145", "O2 Sensor Circuit Slow Response (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0146", "O2 Sensor Circuit No Activity Detected (Bank 1 Sensor 3");
            dtcDictionary.Add("P0147", "O2 Sensor Heater Circuit Malfunction (Bank 1 Sensor 3)");
            dtcDictionary.Add("P0150", "O2 Sensor Circuit Malfunction (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0151", "O2 Sensor Circuit Low Voltage (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0152", "O2 Sensor Circuit High Voltage (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0153", "O2 Sensor Circuit Slow Response (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0154", "O2 Sensor Circuit No Activity Detected (Bank 2 Sensor 1)");
            dtcDictionary.Add("P0155", "O2 Sensor Heater Circuit Malfunction (Bank 2 Sensor 1");
            dtcDictionary.Add("P0156", "O2 Sensor Circuit Malfunction (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0157", "O2 Sensor Circuit Low Voltage (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0158", "O2 Sensor Circuit High Voltage (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0159", "O2 Sensor Circuit Slow Response (Bank 2 Sensor 2");
            dtcDictionary.Add("P0160", "O2 Sensor Circuit No Activity Detected (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0161", "O2 Sensor Heater Circuit Malfunction (Bank 2 Sensor 2)");
            dtcDictionary.Add("P0162", "O2 Sensor Circuit Malfunction (Bank 2 Sensor 3");
            dtcDictionary.Add("P0163", "O2 Sensor Circuit Low Voltage (Bank 2 Sensor 3");
            dtcDictionary.Add("P0164", "O2 Sensor Circuit High Voltage (Bank 2 Sensor 3)");
            dtcDictionary.Add("P0165", "O2 Sensor Circuit Slow Response (Bank 2 Sensor 3)");
            dtcDictionary.Add("P0166", "O2 Sensor Circuit No Activity Detected (Bank 2 Sensor 3");
            dtcDictionary.Add("P0167", "O2 Sensor Heater Circuit Malfunction (Bank 2 Sensor 3)");
            dtcDictionary.Add("P0170", "Fuel Trim Malfunction (Bank 1)");
            dtcDictionary.Add("P0171", "System Too Lean (Bank 1)");
            dtcDictionary.Add("P0172", "System Too Rich (Bank 1)");
            dtcDictionary.Add("P0173", "Fuel Trim Malfunction (Bank 2)");
            dtcDictionary.Add("P0174", "System Too Lean (Bank 2");
            dtcDictionary.Add("P0175", "System Too Rich (Bank 2)");
            dtcDictionary.Add("P0176", "Fuel Composition Sensor Circuit Malfunction");
            dtcDictionary.Add("P0177", "Fuel Composition Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0178", "Fuel Composition Sensor Circuit Low Input");
            dtcDictionary.Add("P0179", "Fuel Composition Sensor Circuit High Input");
            dtcDictionary.Add("P0180", "Fuel Temperature Sensor A Circuit Malfunction");
            dtcDictionary.Add("P0181", "Fuel Temperature Sensor A Circuit Performance");
            dtcDictionary.Add("P0182", "Fuel Temperature Sensor A Circuit low Input");
            dtcDictionary.Add("P0183", "Fuel Temperature Sensor A Circuit High Input");
            dtcDictionary.Add("P0184", "Fuel Temperature Sensor A Circuit Intermittent");
            dtcDictionary.Add("P0185", "Fuel Temperature Sensor B Circuit Malfunction");
            dtcDictionary.Add("P0186", "Fuel Temperature Sensor B Circuit Range/Performance");
            dtcDictionary.Add("P0187", "Fuel Temperature Sensor U Circuit Low Input");
            dtcDictionary.Add("P0188", "Fuel Temperature Sensor B Circuit High Input");
            dtcDictionary.Add("P0189", "Fuel Temperature Sensor B Circuit Intermittent");
            dtcDictionary.Add("P0190", "Fuel Rail Pressure Sensor Circuit Malfunction");
            dtcDictionary.Add("P0191", "Fuel Rail Pressure Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0192", "Fuel Rail Pressure Sensor Circuit Low Input");
            dtcDictionary.Add("P0193", "Fuel Rail Pressure Sensor Circuit High Input");
            dtcDictionary.Add("P0194", "Fuel Rail Pressure Sensor Circuit Intermittent");
            dtcDictionary.Add("P0195", "Engine Oil Temperature Sensor Malfunction");
            dtcDictionary.Add("P0196", "Engine Oil Temperature Sensor Range/Performance");
            dtcDictionary.Add("P0197", "Engine Oil Temperature Sensor Low");
            dtcDictionary.Add("P0198", "Engine Oil Temperature Sensor High");
            dtcDictionary.Add("P0199", "Engine Oil Temperature Sensor Intermittent");
            dtcDictionary.Add("P0200", "Injector Circuit Malfunction");
            dtcDictionary.Add("P0201", "Injector Circuit Malfunction - Cylinder 1");
            dtcDictionary.Add("P0202", "Injector Circuit Malfunction - Cylinder 2");
            dtcDictionary.Add("P0203", "Injector Circuit Malfunction - Cylinder 3");
            dtcDictionary.Add("P0204", "Injector Circuit Malfunction - Cylinder 4");
            dtcDictionary.Add("P0205", "Injector Circuit Malfunction - Cylinder 5");
            dtcDictionary.Add("P0206", "Injector Circuit Malfunction - Cylinder 6");
            dtcDictionary.Add("P0207", "Injector Circuit Malfunction - Cylinder 7");
            dtcDictionary.Add("P0208", "Injector Circuit Malfunction - Cylinder 8");
            dtcDictionary.Add("P0209", "Injector Circuit Malfunction - Cylinder 9");
            dtcDictionary.Add("P0210", "Injector Circuit Malfunction - Cylinder 10");
            dtcDictionary.Add("P0211", "Injector Circuit Malfunction - Cylinder 11");
            dtcDictionary.Add("P0212", "Injector Circuit Malfunction - Cylinder 12");
            dtcDictionary.Add("P0213", "Cold Start Injector 1 Malfunction");
            dtcDictionary.Add("P0214", "Cold Start Injector 2 Malfunction");
            dtcDictionary.Add("P0215", "Engine Shutoff Solenoid Malfunction");
            dtcDictionary.Add("P0216", "Injection Timing Control Circuit Malfunction");
            dtcDictionary.Add("P0217", "Engine Overtemp Condition");
            dtcDictionary.Add("P0218", "Transmission Over Temperature Condition");
            dtcDictionary.Add("P0219", "Engine Over Speed Condition");
            dtcDictionary.Add("P0220", "Throttle/Pedal Position Sensor/Switch B Circuit Malfunction");
            dtcDictionary.Add("P0221", "Throttle/pedal Position Sensor/Switch B Circuit Range/Performance Problem");
            dtcDictionary.Add("P0222", "Throttle/pedal Position Sensor/Switch B Circuit Low Input");
            dtcDictionary.Add("P0223", "Throttle/Pedal Position Sensor/Switch B Circuit High Input");
            dtcDictionary.Add("P0224", "Throttle/Pedal Position Sensor/Switch B Circuit Intermittent");
            dtcDictionary.Add("P0225", "Throttle/Pedal Position Sensor/Switch C Circuit Malfunction");
            dtcDictionary.Add("P0226", "Throttle/Pedal Position Sensor/Switch C Circuit Range/Performance Problem");
            dtcDictionary.Add("P0227", "Throttle/Pedal Position Sensor/Switch C Circuit Low Input");
            dtcDictionary.Add("P0228", "Throttle/Pedal Position Sensor/Switch C Circuit High Input");
            dtcDictionary.Add("P0229", "Throttle/Pedal Position Sensor/Switch C Circuit Intermittent");
            dtcDictionary.Add("P0230", "Fuel Pump Primary Circuit Malfunction");
            dtcDictionary.Add("P0231", "Fuel Pump Secondary Circuit Low");
            dtcDictionary.Add("P0232", "Fuel Pump Secondary Circuit High");
            dtcDictionary.Add("P0233", "Fuel Pump Secondary Circuit Intermittent");
            dtcDictionary.Add("P0234", "Engine Overboost Condition");
            dtcDictionary.Add("P0235", "Turbocharger Boost Sensor A Circuit Malfunction");
            dtcDictionary.Add("P0236", "Turbocharger Boost Sensor A Circuit Range/Performance");
            dtcDictionary.Add("P0237", "Turbocharger Boost Sensor A Circuit Low");
            dtcDictionary.Add("P0238", "Turbocharger Boost Sensor A Circuit High");
            dtcDictionary.Add("P0239", "Turbocharger Boost Sensor B Circuit Malfunction");
            dtcDictionary.Add("P0240", "Turbocharger Boost Sensor B Circuit Range/Performance");
            dtcDictionary.Add("P0241", "Turbocharger Boost Sensor B Circuit Low");
            dtcDictionary.Add("P0242", "Turbocharger Boost Sensor B Circuit High");
            dtcDictionary.Add("P0243", "Turbocharger Wastegate Solenoid A Malfunction");
            dtcDictionary.Add("P0244", "Turbocharger Wastegate Solenoid A Range/Performance");
            dtcDictionary.Add("P0245", "Turbocharger Wastegate Solenoid A low");
            dtcDictionary.Add("P0246", "Turbocharger Wastegate Solenoid A High");
            dtcDictionary.Add("P0247", "Turbocharger Wastegate Solenoid B Malfunction");
            dtcDictionary.Add("P0248", "Turbocharger Wastegate Solenoid B Range/Performance");
            dtcDictionary.Add("P0249", "Turbocharger Wastegate Solenoid B Low");
            dtcDictionary.Add("P0250", "Turbocharger Wastegate Solenoid B High");
            dtcDictionary.Add("P0251", "Injection Pump Fuel Metering Control 'A' Malfunction (Cam/Rotor/Injector)");
            dtcDictionary.Add(
                "P0252",
                "Injection Pump Fuel Metering Control 'A' Range/Performance (Cam/Rotor/Injector)");
            dtcDictionary.Add("P0254", "Injection Pump Fuel Metering Control 'A' High (Cam/Rotor/Injector)");
            dtcDictionary.Add("P0256", "Injection Pump Fuel Metering Control 'B' Malfunction (Cam/Rotor/Injector)");
            dtcDictionary.Add(
                "P0257",
                "Injection Pump Fuel Metering Control 'B' Range/Performance (Cam/Rotor/Injector)");
            dtcDictionary.Add("P0258", "Injection Pump Fuel Metering Control 'B' Low (Cam/Rotor/Injector)");
            dtcDictionary.Add("P0259", "Injection lump Fuel Metering Control 'B' High (Cam/Rotor/Injector)");
            dtcDictionary.Add("P0260", "Injection Pump Fuel Metering Control 'B' Intermittent (Cam/Rotor/Injector)");
            dtcDictionary.Add("P0261", "Cylinder 1 Injector Circuit Low");
            dtcDictionary.Add("P0262", "Cylinder 1 Injector Circuit High");
            dtcDictionary.Add("P0263", "Cylinder 1 Contribution/Balance Fault");
            dtcDictionary.Add("P0264", "Cylinder 2 Injector Circuit Low");
            dtcDictionary.Add("P0265", "Cylinder 2 Injector Circuit High");
            dtcDictionary.Add("P0266", "Cylinder 2 Contribution/Balance Fault");
            dtcDictionary.Add("P0267", "Cylinder 3 Injector Circuit Low");
            dtcDictionary.Add("P0268", "Cylinder 3 Injector Circuit High");
            dtcDictionary.Add("P0269", "Cylinder 3 Contribution/Balance Fault");
            dtcDictionary.Add("P0270", "Cylinder 4 Injector Circuit Low");
            dtcDictionary.Add("P0271", "Cylinder 4 Injector Circuit High");
            dtcDictionary.Add("P0272", "Cylinder 4 Contribution/Balance Fault");
            dtcDictionary.Add("P0273", "Cylinder 5 Injector Circuit Low");
            dtcDictionary.Add("P0274", "Cylinder 5 Injector Circuit High");
            dtcDictionary.Add("P0275", "Cylinder 5 Contribution/Balance Fault");
            dtcDictionary.Add("P0276", "Cylinder 6 Injector Circuit Low");
            dtcDictionary.Add("P0277", "Cylinder 6 Injector Circuit High");
            dtcDictionary.Add("P0278", "Cylinder 6 Contribution/Balance Fault");
            dtcDictionary.Add("P0279", "Cylinder 7 Injector Circuit Low");
            dtcDictionary.Add("P0280", "Cylinder 7 Injector Circuit High");
            dtcDictionary.Add("P0281", "Cylinder 7 Contribution/Balance Fault");
            dtcDictionary.Add("P0282", "Cylinder 8 Injector Circuit Low");
            dtcDictionary.Add("P0283", "Cylinder 8 Injector Circuit High");
            dtcDictionary.Add("P0284", "Cylinder 8 Contribution/Balance Fault");
            dtcDictionary.Add("P0285", "Cylinder 9 Injector Circuit Low");
            dtcDictionary.Add("P0286", "Cylinder 9 Injector Circuit High");
            dtcDictionary.Add("P0287", "Cylinder 9 Contribution/Balance Fault");
            dtcDictionary.Add("P0288", "Cylinder 10 Injector Circuit Low");
            dtcDictionary.Add("P0289", "Cylinder 10 Injector Circuit High");
            dtcDictionary.Add("P0290", "Cylinder 10 Contribution/balance Fault");
            dtcDictionary.Add("P0291", "Cylinder 11 Injector Circuit Low");
            dtcDictionary.Add("P0292", "Cylinder 11 Injector Circuit High");
            dtcDictionary.Add("P0293", "Cylinder 11 Contribution/balance Fault");
            dtcDictionary.Add("P0294", "Cylinder 12 Injector Circuit Low");
            dtcDictionary.Add("P0295", "Cylinder 12 Injector Circuit High");
            dtcDictionary.Add("P0296", "Cylinder 12 Contribution/Balance Fault");

            // Related to: Ignition System and Misfire Detection
            dtcDictionary.Add("P0300", "Random/Multiple Cylinder Misfire Detected");
            dtcDictionary.Add("P0301", "Cylinder 1 Misfire Detected");
            dtcDictionary.Add("P0302", "Cylinder 2 Misfire Detected");
            dtcDictionary.Add("P0303", "Cylinder 3 Misfire Detected");
            dtcDictionary.Add("P0304", "Cylinder 4 Misfire Detected");
            dtcDictionary.Add("P0305", "Cylinder 5 Misfire Detected");
            dtcDictionary.Add("P0306", "Cylinder 6 Misfire Detected");
            dtcDictionary.Add("P0307", "Cylinder 7 Misfire Detected");
            dtcDictionary.Add("P0308", "Cylinder 8 Misfire Detected");
            dtcDictionary.Add("P0309", "Cylinder 9 Misfire Detected");
            dtcDictionary.Add("P0310", "Cylinder 10 Misfire Detected");
            dtcDictionary.Add("P0311", "Cylinder 11 Misfire Detected");
            dtcDictionary.Add("P0312", "Cylinder 12 Misfire Detected");
            dtcDictionary.Add("P0320", "Ignition/Distributor Engine Speed Input Circuit Malfunction");
            dtcDictionary.Add("P0321", "Ignition/Distributor Engine Speed Input Circuit Range/Performance");
            dtcDictionary.Add("P0322", "Ignition/Distributor Engine Speed Input Circuit No Signal");
            dtcDictionary.Add("P0323", "Ignition/Distributor Engine Speed Input Circuit Intermittent");
            dtcDictionary.Add("P0325", "Knock Sensor 1 Circuit Malfunction (Bank 1 or Single Sensor");
            dtcDictionary.Add("P0326", "Knock Sensor 1 Circuit Range/Performance (Bank 1 or Single Sensor)");
            dtcDictionary.Add("P0327", "Knock Sensor 1 Circuit low Input (Bank 1 or Single Sensor)");
            dtcDictionary.Add("P0328", "Knock Sensor 1 Circuit High Input (Bank 1 or Single Sensor)");
            dtcDictionary.Add("P0329", "Knock Sensor 1 Circuit Input Intermittent (Bank 1 or Single Sensor)");
            dtcDictionary.Add("P0330", "Knock Sensor 2 Circuit Malfunction (Bank 2)");
            dtcDictionary.Add("P0331", "Knock Sensor 2 Circuit Range/Performance (Bank 2)");
            dtcDictionary.Add("P0332", "Knock Sensor 2 Circuit Low Input (Bank 2)");
            dtcDictionary.Add("P0333", "Knock Sensor 2 Circuit High Input (Bank 2)");
            dtcDictionary.Add("P0334", "Knock Sensor 2 Circuit Input Intermittent (Bank 2)");
            dtcDictionary.Add("P0335", "Crankshaft Position Sensor A Circuit Malfunction");
            dtcDictionary.Add("P0336", "Crankshaft Position Sensor A Circuit Range/Performance");
            dtcDictionary.Add("P0337", "Crankshaft Position Sensor A Circuit Low Input");
            dtcDictionary.Add("P0338", "Crankshaft Position Sensor A Circuit High Input");
            dtcDictionary.Add("P0339", "Crankshaft Position Sensor A Circuit Intermittent");
            dtcDictionary.Add("P0340", "Camshaft Position Sensor Circuit Malfunction");
            dtcDictionary.Add("P0341", "Camshaft Position Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0342", "Camshaft Position Sensor Circuit Low Input");
            dtcDictionary.Add("P0343", "Camshaft Position Sensor Circuit High Input");
            dtcDictionary.Add("P0344", "Camshaft Position Sensor Circuit Intermittent");
            dtcDictionary.Add("P0350", "Ignition Coil Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0351", "Ignition Coil A Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0352", "Ignition Coil B Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0353", "Ignition Coil C Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0354", "Ignition Coil D Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0355", "Ignition Coil B Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0356", "Ignition Coil F Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0357", "Ignition Coil G Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0358", "Ignition Coil H Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0359", "Ignition Coil I Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0360", "Ignition Coil I Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0361", "Ignition Coil K Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0362", "Ignition Coil L Primary/Secondary Circuit Malfunction");
            dtcDictionary.Add("P0370", "Timing Reference High Resolution Signal A Malfunction");
            dtcDictionary.Add("P0371", "Timing Reference High Resolution Signal A Too Many Pulses");
            dtcDictionary.Add("P0372", "Timing Reference High Resolution Signal A Too Few Pulses");
            dtcDictionary.Add("P0373", "Timing Reference High Resolution Signal A Intermittent/Erratic Pulses");
            dtcDictionary.Add("P0374", "Timing Reference High Resolution Signal A No Pulses");
            dtcDictionary.Add("P0375", "Timing Reference High Resolution Signal B Malfunction");
            dtcDictionary.Add("P0376", "Timing Reference High Resolution Signal B Too Many Pulses");
            dtcDictionary.Add("P0377", "Timing Reference High Resolution Signal B Too Few Pulses");
            dtcDictionary.Add("P0378", "Timing Reference High Resolution Signal B Intermittent/Erratic Pulses");
            dtcDictionary.Add("P0379", "Timing Reference High Resolution Signal B No Pulses");
            dtcDictionary.Add("P0380", "Glow Plug/Heater Circuit 'A' Malfunction");
            dtcDictionary.Add("P0381", "Glow Plug/Heater Indicator Circuit Malfunction");
            dtcDictionary.Add("P0382", "Glow Plug/Heater Circuit 'B' Malfunction");
            dtcDictionary.Add("P0385", "Crankshaft Position Sensor B Circuit Malfunction");
            dtcDictionary.Add("P0386", "Crankshaft Position Sensor B Circuit Range/Performance");
            dtcDictionary.Add("P0387", "Crankshaft Position Sensor B Circuit Low Input");
            dtcDictionary.Add("P0388", "Crankshaft Position Sensor B Circuit High Input");
            dtcDictionary.Add("P0389", "Crankshaft Position Sensor B Circuit Intermittent");

            // Related to: Auxilliary Emission Controls. EGR, EVAP etc.
            dtcDictionary.Add("P0400", "Exhaust Gas Recirculation Flow Malfunction");
            dtcDictionary.Add("P0401", "Exhaust Gas Recirculation Flow Insufficient Detected");
            dtcDictionary.Add("P0402", "Exhaust Gas Recirculation Flow Excessive Detected");
            dtcDictionary.Add("P0403", "Exhaust Gas Recirculation Circuit Malfunction");
            dtcDictionary.Add("P0404", "Exhaust Gas Recirculation Circuit Range/Performance");
            dtcDictionary.Add("P0405", "Exhaust Gas Recirculation Sensor A Circuit Low");
            dtcDictionary.Add("P0406", "Exhaust Gas Recirculation Sensor A Circuit High");
            dtcDictionary.Add("P0407", "Exhaust Gas Recirculation Sensor B Circuit Low");
            dtcDictionary.Add("P0408", "Exhaust Gas Recirculation Sensor B Circuit High");
            dtcDictionary.Add("P0410", "Secondary Air Injection System Malfunction");
            dtcDictionary.Add("P0411", "Secondary Air Injection System Incorrect Flow Detected");
            dtcDictionary.Add("P0412", "Secondary Air Injection System Switching Valve A Circuit Malfunction");
            dtcDictionary.Add("P0413", "Secondary Air Injection System Switching Valve A Circuit Open");
            dtcDictionary.Add("P0414", "Secondary Air Injection System Switching Valve A Circuit Shorted");
            dtcDictionary.Add("P0415", "Secondary Air Injection System Switching Valve B Circuit Malfunction");
            dtcDictionary.Add("P0416", "Secondary Air Injection System Switching Valve B Circuit Open");
            dtcDictionary.Add("P0417", "Secondary Air Injection System Switching Valve B Circuit Shorted");
            dtcDictionary.Add("P0418", "Secondary Air Injection System Relay A circuit Malfunction");
            dtcDictionary.Add("P0419", "Secondary Air Injection System Relay B Circuit Malfunction");
            dtcDictionary.Add("P0420", "Catalyst System Efficiency Below Threshold (Bank 1)");
            dtcDictionary.Add("P0421", "Warm Up Catalyst Efficiency Below Threshold (Bank 1");
            dtcDictionary.Add("P0422", "Main Catalyst Efficiency Below Threshold (Bank 1");
            dtcDictionary.Add("P0423", "Heated Catalyst Efficiency Below Threshold (Bank l)");
            dtcDictionary.Add("P0424", "Heated Catalyst Temperature Below Threshold (Bank 1)");
            dtcDictionary.Add("P0430", "Catalyst System Efficiency Below Threshold (Bank 2)");
            dtcDictionary.Add("P0431", "Warm Up Catalyst Efficiency Below Threshold (Bank 2)");
            dtcDictionary.Add("P0432", "Main Catalyst Efficiency Below Threshold (Bank 2)");
            dtcDictionary.Add("P0433", "Heated Catalyst Efficiency Below Threshold (Bank 2)");
            dtcDictionary.Add("P0434", "Heated Catalyst Temperature Below Threshold (Bank 2)");
            dtcDictionary.Add("P0440", "Evaporative Emission Control System Malfunction");
            dtcDictionary.Add("P0441", "Evaporative Emission Control System Incorrect Purge flow");
            dtcDictionary.Add("P0442", "Evaporative Emission Control System leak Detected (small leak)");
            dtcDictionary.Add("P0443", "Evaporative Emission Control System Purge Control Valve circuit Malfunction");
            dtcDictionary.Add("P0444", "Evaporative Emission Control System Purge Control Valve Circuit Open");
            dtcDictionary.Add("P0445", "Evaporative Emission Control System Purge Control Valve Circuit Shorted");
            dtcDictionary.Add("P0446", "Evaporative Emission Control System Vent Control Circuit Malfunction");
            dtcDictionary.Add("P0447", "Evaporative Emission Control System Vent Control Circuit Open");
            dtcDictionary.Add("P0448", "Evaporative Emission Control System Vent Control Circuit Shorted");
            dtcDictionary.Add("P0449", "Evaporative Emission Control System Vent Valve/Solenoid Circuit Malfunction");
            dtcDictionary.Add("P0450", "Evaporative Emission Control System Pressure Sensor Malfunction");
            dtcDictionary.Add("P0451", "Evaporative Emission Control System Pressure Sensor Range/Performance");
            dtcDictionary.Add("P0452", "Evaporative Emission Control System Pressure Sensor Low Input");
            dtcDictionary.Add("P0453", "Evaporative Emission Control System Pressure Sensor High Input");
            dtcDictionary.Add("P0454", "Evaporative Emission Control System Pressure Sensor Intermittent");
            dtcDictionary.Add("P0455", "Evaporative Emission Control System Tank Detected (gross leak)");
            dtcDictionary.Add("P0460", "Fuel Level Sensor Circuit Malfunction");
            dtcDictionary.Add("P0461", "Fuel Level Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0462", "Fuel level Sensor Circuit Low Input");
            dtcDictionary.Add("P0463", "Fuel level Sensor Circuit High Input");
            dtcDictionary.Add("P0464", "Fuel level Sensor Circuit Intermittent");
            dtcDictionary.Add("P0465", "Purge Flow Sensor Circuit Malfunction");
            dtcDictionary.Add("P0466", "Purge flow Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0467", "Purge Flow Sensor Circuit Low Input");
            dtcDictionary.Add("P0468", "Purge flow Sensor Circuit High Input");
            dtcDictionary.Add("P0469", "Purge flow Sensor Circuit Intermittent");
            dtcDictionary.Add("P0470", "Exhaust Pressure Sensor Malfunction");
            dtcDictionary.Add("P0471", "Exhaust Pressure Sensor Range/Performance");
            dtcDictionary.Add("P0472", "Exhaust Pressure Sensor Low");
            dtcDictionary.Add("P0473", "Exhaust Pressure Sensor High");
            dtcDictionary.Add("P0474", "Exhaust Pressure Sensor Intermittent");
            dtcDictionary.Add("P0475", "Exhaust Pressure Control Valve Malfunction");
            dtcDictionary.Add("P0476", "Exhaust Pressure Control Valve Range/Performance");
            dtcDictionary.Add("P0477", "Exhaust Pressure Control Valve Low");
            dtcDictionary.Add("P0478", "Exhaust Pressure Control Valve High");
            dtcDictionary.Add("P0479", "Exhaust Pressure Control Valve Intermittent");
            dtcDictionary.Add("P0480", "Cooling Fan 1 Control Circuit Malfunction");
            dtcDictionary.Add("P0481", "Cooling Fan 2 Control Circuit Malfunction");
            dtcDictionary.Add("P0482", "Cooling Fan 3 Control Circuit Malfunction");
            dtcDictionary.Add("P0483", "Cooling Fan Rationality Check Malfunction");
            dtcDictionary.Add("P0484", "Cooling Fan Circuit Over Current");
            dtcDictionary.Add("P0485", "Cooling Fan Power/Ground Circuit Malfunction");

            // Related to: Vehicle Speed, Idle Control, and Auxiliary Inputs
            dtcDictionary.Add("P0500", "Vehicle Speed Sensor Malfunction");
            dtcDictionary.Add("P0501", "Vehicle Speed Sensor Range/Performance");
            dtcDictionary.Add("P0502", "Vehicle Speed Sensor Circuit Low Input");
            dtcDictionary.Add("P0503", "Vehicle Speed Sensor Intermittent/Erratic/High");
            dtcDictionary.Add("P0505", "Idle Control System Malfunction");
            dtcDictionary.Add("P0506", "Idle Control System RPM lower Than Expected");
            dtcDictionary.Add("P0507", "Idle Control System RPM higher Than Expected");
            dtcDictionary.Add("P0510", "Closed Throttle Position Switch Malfunction");
            dtcDictionary.Add("P0520", "Engine Oil Pressure Sensor/Switch Circuit Malfunction");
            dtcDictionary.Add("P0521", "Engine Oil Pressure Sensor/Switch Range/Performance");
            dtcDictionary.Add("P0522", "Engine Oil Pressure Sensor/Switch Low Voltage");
            dtcDictionary.Add("P0523", "Engine Oil Pressure Sensor/Switch High Voltage");
            dtcDictionary.Add("P0530", "A/C Refrigerant Pressure Sensor Circuit Malfunction");
            dtcDictionary.Add("P0531", "A/C Refrigerant Pressure Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0532", "A/C Refrigerant Pressure Sensor Circuit Low Input");
            dtcDictionary.Add("P0533", "A/C Refrigerant pressure Sensor Circuit High Input");
            dtcDictionary.Add("P0534", "Air Conditioner Refrigerant Charge Loss");
            dtcDictionary.Add("P0550", "Power Steering Pressure Sensor Circuit Malfunction");
            dtcDictionary.Add("P0551", "Power Steering Pressure Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0552", "Power Steering Pressure Sensor Circuit Low Input");
            dtcDictionary.Add("P0553", "Power Steering Pressure Sensor Circuit High Input");
            dtcDictionary.Add("P0554", "Power Steering Pressure sensor Circuit Intermittent");
            dtcDictionary.Add("P0560", "System Voltage Malfunction");
            dtcDictionary.Add("P0561", "System Voltage Unstable");
            dtcDictionary.Add("P0562", "System Voltage Low");
            dtcDictionary.Add("P0563", "System Voltage High");
            dtcDictionary.Add("P0565", "Cruise Control On Signal Malfunction");
            dtcDictionary.Add("P0566", "Cruise Control Off Signal Malfunction");
            dtcDictionary.Add("P0567", "Cruise Control Resume Signal Malfunction");
            dtcDictionary.Add("P0568", "Cruise Control Set Signal Malfunction");
            dtcDictionary.Add("P0569", "Cruise Control Coast Signal Malfunction");
            dtcDictionary.Add("P0570", "Cruise Control Accel Signal Malfunction");
            dtcDictionary.Add("P0571", "Cruise Control/Brake Switch A Circuit Malfunction");
            dtcDictionary.Add("P0572", "Cruise Control/Brake Switch A Circuit Low");
            dtcDictionary.Add("P0573", "Cruise Control/Brake Switch A Circuit High");
            dtcDictionary.Add("P0574", "through P0580 Reserved for Cruise Codes");

            // Related to: Computer and Auxiliary Outputs
            dtcDictionary.Add("P0600", "Serial Communication Link Malfunction");
            dtcDictionary.Add("P0601", "Internal Control Module Memory Check Sum Error");
            dtcDictionary.Add("P0602", "Control Module Programming Error");
            dtcDictionary.Add("P0603", "Internal Control Module Keep Alive Memory (KAM) Error");
            dtcDictionary.Add("P0604", "Internal Control Module RAM Error");
            dtcDictionary.Add("P0605", "Internal Control Module ROM Error (Module Identification Defined by SAE J1979)");
            dtcDictionary.Add("P0606", "PCM Processor Fault");
            dtcDictionary.Add("P0608", "Control Module VSS Output 'A' Malfunction");
            dtcDictionary.Add("P0609", "Control Module VSS Output 'B' Malfunction");
            dtcDictionary.Add("P0620", "Generator Control Circuit Malfunction");
            dtcDictionary.Add("P0621", "Generator Lamp 'L' Control Circuit Malfunction");
            dtcDictionary.Add("P0622", "Generator Field 'F' Control Circuit Malfunction");
            dtcDictionary.Add("P0650", "Malfunction Indicator Lamp (MIL) Control Circuit Malfunction");
            dtcDictionary.Add("P0654", "Engine RPM Output Circuit Malfunction");
            dtcDictionary.Add("P0655", "Engine Hot Lamp Output Control Circuit Malfunction");
            dtcDictionary.Add("P0656", "Fuel Level Output Circuit Malfunction");

            // Other
            dtcDictionary.Add("P0700", "Transmission Control System Malfunction");
            dtcDictionary.Add("P0701", "Transmission Control System Range/Performance");
            dtcDictionary.Add("P0702", "Transmission Control System Electrical");
            dtcDictionary.Add("P0703", "Torque Converter/Brake Switch B Circuit Malfunction");
            dtcDictionary.Add("P0704", "Clutch Switch Input Circuit Malfunction");
            dtcDictionary.Add("P0705", "Transmission Range Sensor Circuit malfunction (PRNDL Input)");
            dtcDictionary.Add("P0706", "Transmission Range Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0707", "Transmission Range Sensor Circuit Low Input");
            dtcDictionary.Add("P0708", "Transmission Range Sensor Circuit High Input");
            dtcDictionary.Add("P0709", "Transmission Range Sensor Circuit Intermittent");
            dtcDictionary.Add("P0710", "Transmission Fluid Temperature Sensor Circuit Malfunction");
            dtcDictionary.Add("P0711", "Transmission Fluid Temperature Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0712", "Transmission Fluid Temperature Sensor Circuit Low Input");
            dtcDictionary.Add("P0713", "Transmission Fluid Temperature Sensor Circuit High Input");
            dtcDictionary.Add("P0714", "Transmission Fluid Temperature Sensor Circuit Intermittent");
            dtcDictionary.Add("P0715", "Input/Turbine Speed Sensor Circuit Malfunction");
            dtcDictionary.Add("P0716", "Input/Turbine Speed Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0717", "Input/Turbine Speed Sensor Circuit No Signal");
            dtcDictionary.Add("P0718", "Input/Turbine Speed Sensor Circuit Intermittent");
            dtcDictionary.Add("P0719", "Torque Converter/Brake Switch B Circuit Low");
            dtcDictionary.Add("P0720", "Output Speed Sensor Circuit Malfunction");
            dtcDictionary.Add("P0721", "Output Speed Sensor Range/Performance");
            dtcDictionary.Add("P0722", "Output Speed Sensor No Signal");
            dtcDictionary.Add("P0723", "Output Speed Sensor Intermittent");
            dtcDictionary.Add("P0724", "Torque Converter/Brake Switch B Circuit High");
            dtcDictionary.Add("P0725", "Engine Speed input Circuit Malfunction");
            dtcDictionary.Add("P0726", "Engine Speed Input Circuit Range/Performance");
            dtcDictionary.Add("P0727", "Engine Speed Input Circuit No Signal");
            dtcDictionary.Add("P0728", "Engine Speed Input Circuit Intermittent");
            dtcDictionary.Add("P0730", "Incorrect Gear Ratio");
            dtcDictionary.Add("P0731", "Gear 1 Incorrect ratio");
            dtcDictionary.Add("P0732", "Gear 2 Incorrect ratio");
            dtcDictionary.Add("P0733", "Gear 3 Incorrect ratio");
            dtcDictionary.Add("P0734", "Gear 4 Incorrect ratio");
            dtcDictionary.Add("P0735", "Gear 5 Incorrect ratio");
            dtcDictionary.Add("P0736", "Reverse incorrect gear ratio");
            dtcDictionary.Add("P0740", "Torque Converter Clutch Circuit Malfuction");
            dtcDictionary.Add("P0741", "Torque Converter Clutch Circuit Performance or Stuck Off");
            dtcDictionary.Add("P0742", "Torque Converter Clutch Circuit Stuck On");
            dtcDictionary.Add("P0743", "Torque Converter Clutch Circuit Electrical");
            dtcDictionary.Add("P0744", "Torque Converter Clutch Circuit Intermittent");
            dtcDictionary.Add("P0745", "Pressure Control Solenoid Malfunction");
            dtcDictionary.Add("P0746", "Pressure Control Solenoid Performance or Stuck Off");
            dtcDictionary.Add("P0747", "Pressure Control Solenoid Stuck On");
            dtcDictionary.Add("P0748", "Pressure Control Solenoid Electrical");
            dtcDictionary.Add("P0749", "Pressure Control Solenoid Intermittent");
            dtcDictionary.Add("P0750", "Shift Solenoid A Malfunction");
            dtcDictionary.Add("P0751", "Shift Solenoid A Performance or Stuck Off/1-2 Shift Solenoid Valve Performance");
            dtcDictionary.Add("P0752", "Shift Solenoid A Stuck On");
            dtcDictionary.Add("P0753", "Shift Solenoid A Electrical/1-2 Shift Solenoid Circuit Electrical");
            dtcDictionary.Add("P0754", "Shift Solenoid A Intermittent");
            dtcDictionary.Add("P0755", "Shift Solenoid B Malfunction");
            dtcDictionary.Add("P0756", "Shift Solenoid B Performance or Stuck Off/2-3 Shift Solenoid Valve Performance");
            dtcDictionary.Add("P0757", "Shift Solenoid B Stuck On");
            dtcDictionary.Add("P0758", "Shift Solenoid B Electrical/2-3 Shift Solenoid Circuit Electrical");
            dtcDictionary.Add("P0759", "Shift Solenoid B Intermittent");
            dtcDictionary.Add("P0760", "Shift Solenoid C Malfunction");
            dtcDictionary.Add("P0761", "Shift Solenoid C Performance or Stuck Off");
            dtcDictionary.Add("P0762", "Shift Solenoid C Stuck On");
            dtcDictionary.Add("P0763", "Shift Solenoid C Electrical");
            dtcDictionary.Add("P0764", "Shift Solenoid C Intermittent");
            dtcDictionary.Add("P0765", "Shift Solenoid D Malfunction");
            dtcDictionary.Add("P0766", "Shift Solenoid D Performance or Stuck Off");
            dtcDictionary.Add("P0767", "Shift Solenoid D Stuck On");
            dtcDictionary.Add("P0768", "Shift Solenoid D Electrical");
            dtcDictionary.Add("P0769", "Shift Solenoid D Intermittent");
            dtcDictionary.Add("P0770", "Shift Solenoid E Malfunction");
            dtcDictionary.Add("P0771", "Shift Solenoid E Performance or Stuck Off");
            dtcDictionary.Add("P0772", "Shift Solenoid E Stuck On");
            dtcDictionary.Add("P0773", "Shift Solenoid E Electrical");
            dtcDictionary.Add("P0774", "Shift Solenoid E Intermittent");
            dtcDictionary.Add("P0775", "Pressure Control Solenoid B Malfunction");
            dtcDictionary.Add("P0776", "Pressure Control Solenoid B Performance");
            dtcDictionary.Add("P0777", "Pressure Control Solenoid B Stuck On");
            dtcDictionary.Add("P0778", "Pressure Control Solenoid B Electrical");
            dtcDictionary.Add("P0779", "Pressure Control Solenoid B Intermittent");
            dtcDictionary.Add("P0780", "Shift Malfunction");
            dtcDictionary.Add("P0781", "1-2 Shift Malfunction");
            dtcDictionary.Add("P0782", "2-3 Shift Malfunction");
            dtcDictionary.Add("P0783", "3-4 Shift Malfunction");
            dtcDictionary.Add("P0784", "4-5 Shift Malfunction");
            dtcDictionary.Add("P0785", "Shift/Timing Solenoid Malfunction/ 3-2 Shift Solenoid Circuit Electrical");
            dtcDictionary.Add("P0786", "Shift/Timing Solenoid Range/Performance");
            dtcDictionary.Add("P0787", "Shift/Timing Solenoid Low");
            dtcDictionary.Add("P0788", "Shift/Timing Solenoid High");
            dtcDictionary.Add("P0789", "Shift/Timing Solenoid Intermittent");
            dtcDictionary.Add("P0790", "Normal/Performance Switch Circuit Malfunction");
            dtcDictionary.Add("P0791", "Intermediate Shaft Speed Sensor Circuit");
            dtcDictionary.Add("P0792", "Intermediate Shaft Speed Sensor Circuit Range/Performance");
            dtcDictionary.Add("P0793", "Intermediate Shaft Speed Sensor Circuit No signal");
            dtcDictionary.Add("P0794", "Intermediate Shaft Speed Sensor Circuit Intermittent");
            dtcDictionary.Add("P0795", "Pressure Control Solenoid C");
            dtcDictionary.Add("P0796", "Pressure Control Solenoid C Performance or Stuck off");
            dtcDictionary.Add("P0797", "Pressure Control Solenoid C Stuck On");
            dtcDictionary.Add("P0798", "Pressure Control Solenoid C Electrical");
            dtcDictionary.Add("P0799", "Pressure Control Solenoid C Intermittent");
            dtcDictionary.Add("P0801", "Reverse Inhibit Control Circuit Malfunction");
            dtcDictionary.Add("P0803", "1-4 Upshift (Skip Shift) Solenoid Control Circuit Malfunction");
            dtcDictionary.Add("P0804", "1-4 Upshift (Skip Shift) Lamp Control Circuit Malfunction");
            dtcDictionary.Add("P0805", "Clutch Position Sensor Circuit Malfunction");
            dtcDictionary.Add("P0806", "Clutch Position Sensor Circuit Range/Performance Malfunction");
            dtcDictionary.Add("P0807", "Clutch Position Sensor Circuit Low Malfunction");
            dtcDictionary.Add("P0808", "Clutch Position Sensor Circuit High Malfunction");
            dtcDictionary.Add("P0809", "Clutch Position Sensor Circuit Intermittent Malfunction");
            dtcDictionary.Add("P0810", "Clutch Position Control Error");
            dtcDictionary.Add("P0811", "Excessive Clutch Slippage");
            dtcDictionary.Add("P0812", "Reverse Input Circuit");
            dtcDictionary.Add("P0813", "Reverse Output Circuit");
            dtcDictionary.Add("P0814", "Transmission Range Display Circuit");
            dtcDictionary.Add("P0815", "Upshift Switch Circuit");
            dtcDictionary.Add("P0816", "Downshift Switch Circuit");
            dtcDictionary.Add("P0817", "Starter Disable Circuit");
            dtcDictionary.Add("P0818", "Driveline Disconnect Switch Input Circuit");
            dtcDictionary.Add("P0819", "Up and Down Shift Switch to Transmission Range Correlation");
            dtcDictionary.Add("P0820", "Gear Lever X - Y Position Sensor Circuit");
            dtcDictionary.Add("P0821", "Gear Lever X Position Circuit");
            dtcDictionary.Add("P0822", "Gear Lever Y Position Circuit");
            dtcDictionary.Add("P0823", "Gear Lever X Position Circuit Intermittent");
            dtcDictionary.Add("P0824", "Gear Lever Y Position Circuit Intermittent");
            dtcDictionary.Add("P0825", "Gear Lever Push - Pull Switch (Shift Anticipate)");
            dtcDictionary.Add("P0826", "Up and Down Switch Input Circuit");
            dtcDictionary.Add("P0827", "Up and Down Switch Input Circuit Low");
            dtcDictionary.Add("P0828", "Up and Down Switch Input Circuit High");
            dtcDictionary.Add("P0829", "5-6 Shift");
            dtcDictionary.Add("P0830", "Clutch Pedal Switch A Circuit");
            dtcDictionary.Add("P0831", "Clutch Pedal Switch A Circuit Low");
            dtcDictionary.Add("P0832", "Clutch Pedal Switch A Circuit High");
            dtcDictionary.Add("P0833", "Clutch Pedal Switch B Circuit");
            dtcDictionary.Add("P0834", "Clutch Pedal Switch B Circuit Low");
            dtcDictionary.Add("P0835", "Clutch Pedal Switch B Circuit High");
            dtcDictionary.Add("P0836", "Four Wheel Drive (4WD) Switch Circuit");
            dtcDictionary.Add("P0837", "Four Wheel Drive (4WD) Switch Circuit Range/Performance");
            dtcDictionary.Add("P0838", "Four Wheel Drive (4WD) Switch Circuit Low");
            dtcDictionary.Add("P0839", "Four Wheel Drive (4WD) Switch Circuit High");
            dtcDictionary.Add("P0840", "Transmission Fluid Pressure Sensor/Switch A Circuit");
            dtcDictionary.Add("P0841", "Transmission Fluid Pressure Sensor/Switch A Circuit Range/Performance");
            dtcDictionary.Add("P0842", "Transmission Fluid Pressure Sensor/Switch A Circuit Low");
            dtcDictionary.Add("P0843", "Transmission Fluid Pressure Sensor/Switch A Circuit High");
            dtcDictionary.Add("P0844", "Transmission Fluid Pressure Sensor/Switch A Circuit Intermittent");
            dtcDictionary.Add("P0845", "Transmission Fluid Pressure Sensor/Switch B Circuit");
            dtcDictionary.Add("P0846", "Transmission Fluid Pressure Sensor/Switch B Circuit Range/Performance");
            dtcDictionary.Add("P0847", "Transmission Fluid Pressure Sensor/Switch B Circuit Low");
            dtcDictionary.Add("P0848", "Transmission Fluid Pressure Sensor/Switch B Circuit High");
            dtcDictionary.Add("P0849", "Transmission Fluid Pressure Sensor/Switch B Circuit Intermittent");
            dtcDictionary.Add("P0850", "Park / Neutral Switch Input Circuit");
            dtcDictionary.Add("P0851", "Park / Neutral Switch Input Circuit Low");
            dtcDictionary.Add("P0852", "Park / Neutral Switch Input Circuit High");
            dtcDictionary.Add("P0853", "Drive Switch Input Circuit");
            dtcDictionary.Add("P0854", "Drive Switch Input Circuit Low");
            dtcDictionary.Add("P0855", "Drive Switch Input Circuit High");
            dtcDictionary.Add("P0856", "Traction Control Input Signal");
            dtcDictionary.Add("P0857", "Traction Control Input Signal Range/Performance");
            dtcDictionary.Add("P0858", "Traction Control Input Signal Low");
            dtcDictionary.Add("P0859", "Traction Control Input Signal High");
            dtcDictionary.Add("P0860", "Gear Shift Module Communication Circuit");
            dtcDictionary.Add("P0861", "Gear Shift Module Communication Circuit Low");
            dtcDictionary.Add("P0862", "Gear Shift Module Communication Circuit High");
            dtcDictionary.Add("P0863", "TCM Communication Circuit");
            dtcDictionary.Add("P0864", "TCM Communication Circuit Range/Performance");
            dtcDictionary.Add("P0865", "TCM Communication Circuit Low");
            dtcDictionary.Add("P0866", "TCM Communication Circuit High");
            dtcDictionary.Add("P0867", "Transmission Fluid Pressure");
            dtcDictionary.Add("P0868", "Transmission Fluid Pressure Low");
            dtcDictionary.Add("P0869", "Transmission Fluid Pressure High");
            dtcDictionary.Add("P0870", "Transmission Fluid Pressure Sensor/Switch C Circuit");
            dtcDictionary.Add("P0871", "Transmission Fluid Pressure Sensor/Switch C Circuit Range/Performance");
            dtcDictionary.Add("P0872", "Transmission Fluid Pressure Sensor/Switch C Circuit Low");
            dtcDictionary.Add("P0873", "Transmission Fluid Pressure Sensor/Switch C Circuit High");
            dtcDictionary.Add("P0874", "Transmission Fluid Pressure Sensor/Switch C Circuit Intermittent");
            dtcDictionary.Add("P0875", "Transmission Fluid Pressure Sensor/Switch D Circuit");
            dtcDictionary.Add("P0876", "Transmission Fluid Pressure Sensor/Switch D Circuit Range/Performance");
            dtcDictionary.Add("P0877", "Transmission Fluid Pressure Sensor/Switch D Circuit Low");
            dtcDictionary.Add("P0878", "Transmission Fluid Pressure Sensor/Switch D Circuit High");
            dtcDictionary.Add("P0879", "Transmission Fluid Pressure Sensor/Switch D Circuit Intermittent");
            dtcDictionary.Add("P0880", "TCM Power Input Signal");
            dtcDictionary.Add("P0881", "TCM Power Input Signal Range/Performance");
            dtcDictionary.Add("P0882", "TCM Power Input Signal Low");
            dtcDictionary.Add("P0883", "TCM Power Input Signal High");
            dtcDictionary.Add("P0884", "TCM Power Input Signal Intermittent");
            dtcDictionary.Add("P0885", "TCM Power Relay Control Circuit /Open");
            dtcDictionary.Add("P0886", "TCM Power Relay Control Circuit Low");
            dtcDictionary.Add("P0887", "TCM Power Relay Control Circuit High");
            dtcDictionary.Add("P0888", "TCM Power Relay Sense Circuit");
            dtcDictionary.Add("P0889", "TCM Power Relay Sense Circuit Range/Performance");
            dtcDictionary.Add("P0890", "TCM Power Relay Sense Circuit Low");
            dtcDictionary.Add("P0891", "TCM Power Relay Sense Circuit High");
            dtcDictionary.Add("P0892", "TCM Power Relay Sense Circuit Intermittent");
            dtcDictionary.Add("P0893", "Multiple Gears Engaged");
            dtcDictionary.Add("P0894", "Transmission Component Slipping");
            dtcDictionary.Add("P0895", "Shift Time Too Short");
            dtcDictionary.Add("P0896", "Shift Time Too Long");
            dtcDictionary.Add("P0897", "Transmission Fluid Deteriorated");
            dtcDictionary.Add("P0898", "Transmission Control System MIL Request Circuit Low");
            dtcDictionary.Add("P0899", "Transmission Control System MIL Request Circuit High");
            dtcDictionary.Add("P0900", "Clutch Actuator Circuit / Open");
            dtcDictionary.Add("P0901", "Clutch Actuator Circuit Range/Performance");
            dtcDictionary.Add("P0902", "Clutch Actuator Circuit Low");
            dtcDictionary.Add("P0903", "Clutch Actuator Circuit High");
            dtcDictionary.Add("P0904", "Gate Select Position Circuit [senses left / right position]");
            dtcDictionary.Add("P0905", "Gate Select Position Circuit Range/Performance");
            dtcDictionary.Add("P0906", "Gate Select Position Circuit Low");
            dtcDictionary.Add("P0907", "Gate Select Position Circuit High");
            dtcDictionary.Add("P0908", "Gate Select Position Circuit Intermittent");
            dtcDictionary.Add("P0909", "Gate Select Control Error");
            dtcDictionary.Add("P0910", "Gate Select Actuator Circuit / Open [left / right motion]");
            dtcDictionary.Add("P0911", "Gate Select Actuator Circuit Range/Performance");
            dtcDictionary.Add("P0912", "Gate Select Actuator Circuit Low");
            dtcDictionary.Add("P0913", "Gate Select Actuator Circuit High");
            dtcDictionary.Add(
                "P0914",
                "Gear Shift Position Circuit [senses forward / rearward position, odd / even gears]");
            dtcDictionary.Add("P0915", "Gear Shift Position Circuit Range/Performance");
            dtcDictionary.Add("P0916", "Gear Shift Position Circuit Low");
            dtcDictionary.Add("P0917", "Gear Shift Position Circuit High");
            dtcDictionary.Add("P0918", "Gear Shift Position Circuit Intermittent");
            dtcDictionary.Add("P0919", "Gear Shift Position Control Error");
            dtcDictionary.Add("P0920", "Gear Shift Forward Actuator Circuit / Open [forward motion, odd gears, 1,3,5]");
            dtcDictionary.Add("P0921", "Gear Shift Foward Actuator Circuit Range/Performance");
            dtcDictionary.Add("P0922", "Gear Shift Foward Actuator Circuit Low");
            dtcDictionary.Add("P0923", "Gear Shift Foward Actuator Circuit High");
            dtcDictionary.Add(
                "P0924",
                "Gear Shift Reverse Actuator Circuit / Open [rearward motion, even gears, 2,4,6]");
            dtcDictionary.Add("P0925", "Gear Shift Reverse Actuator Circuit Range/Performance");
            dtcDictionary.Add("P0926", "Gear Shift Reverse Actuator Circuit Low");
            dtcDictionary.Add("P0927", "Gear Shift Reverse Actuator Circuit High");
            dtcDictionary.Add("P0928", "Gear Shift Lock Solenoid Circuit / Open");
            dtcDictionary.Add("P0929", "Gear Shift Lock Solenoid Circuit Range/Performance");
            dtcDictionary.Add("P0930", "Gear Shift Lock Solenoid Circuit Low");
            dtcDictionary.Add("P0931", "Gear Shift Lock Solenoid Circuit High");
            dtcDictionary.Add("P0932", "Hydraulic Pressure Sensor Circuit");
            dtcDictionary.Add("P0933", "Hydraulic Pressure Sensor Range/Performance");
            dtcDictionary.Add("P0934", "Hydraulic Pressure Sensor Circuit Low Input");
            dtcDictionary.Add("P0935", "Hydraulic Pressure Sensor Circuit High Input");
            dtcDictionary.Add("P0936", "Hydraulic Pressure Sensor Circuit Intermittent");
            dtcDictionary.Add("P0937", "Hydraulic Oil Temperature Sensor Circuit");
            dtcDictionary.Add("P0938", "Hydraulic Oil Temperature Sensor Range/Performance");
            dtcDictionary.Add("P0939", "Hydraulic Oil Temperature Sensor Circuit Low Input");
            dtcDictionary.Add("P0940", "Hydraulic Oil Temperature Sensor Circuit High Input");
            dtcDictionary.Add("P0941", "Hydraulic Oil Temperature Sensor Circuit Intermittent");
            dtcDictionary.Add("P0942", "Hydraulic Pressure Unit");
            dtcDictionary.Add("P0943", "Hydraulic Pressure Unit Cycling Period Too Short");
            dtcDictionary.Add("P0944", "Hydraulic Pressure Unit Loss of Pressure");
            dtcDictionary.Add("P0945", "Hydraulic Pump Relay Circuit / Open");
            dtcDictionary.Add("P0946", "Hydraulic Pump Relay Circuit Range/Performance");
            dtcDictionary.Add("P0947", "Hydraulic Pump Relay Circuit Low");
            dtcDictionary.Add("P0948", "Hydraulic Pump Relay Circuit High");
            dtcDictionary.Add("P0949", "ASM Adaptive Learning Not Done");
            dtcDictionary.Add("P0950", "ASM Control Circuit [Up / Down / Auto / etc]");
            dtcDictionary.Add("P0951", "ASM Control Circuit Range/Performance");
            dtcDictionary.Add("P0952", "ASM Control Circuit Low");
            dtcDictionary.Add("P0953", "ASM Control Circuit High");
            dtcDictionary.Add("P0954", "ASM Control Circuit Intermittent");
            dtcDictionary.Add("P0955", "ASM Mode Circuit [Perf / Winter / Sport / etc]");
            dtcDictionary.Add("P0956", "ASM Mode Circuit Range/Performance");
            dtcDictionary.Add("P0957", "ASM Mode Circuit Low");
            dtcDictionary.Add("P0958", "ASM Mode Circuit High");
            dtcDictionary.Add("P0959", "ASM Mode Circuit Intermittent");
            dtcDictionary.Add("P0960", "Pressure Control Solenoid A Control Circuit / Open");
            dtcDictionary.Add("P0961", "Pressure Control Solenoid A Control Circuit Range/Performance");
            dtcDictionary.Add("P0962", "Pressure Control Solenoid A Control Circuit Low");
            dtcDictionary.Add("P0963", "Pressure Control Solenoid A Control Circuit High");
            dtcDictionary.Add("P0964", "Pressure Control Solenoid B Control Circuit / Open");
            dtcDictionary.Add("P0965", "Pressure Control Solenoid B Control Circuit Range/Performance");
            dtcDictionary.Add("P0966", "Pressure Control Solenoid B Control Circuit Low");
            dtcDictionary.Add("P0967", "Pressure Control Solenoid B Control Circuit High");
            dtcDictionary.Add("P0968", "Pressure Control Solenoid C Control Circuit / Open");
            dtcDictionary.Add("P0969", "Pressure Control Solenoid C Control Circuit Range/Performance");
            dtcDictionary.Add("P0970", "Pressure Control Solenoid C Control Circuit Low");
            dtcDictionary.Add("P0971", "Pressure Control Solenoid C Control Circuit High");
            dtcDictionary.Add("P0972", "Shift Solenoid A Control Circuit Range/Performance");
            dtcDictionary.Add("P0973", "Shift Solenoid A Control Circuit Low");
            dtcDictionary.Add("P0974", "Shift Solenoid A Control Circuit High");
            dtcDictionary.Add("P0975", "Shift Solenoid B Control Circuit Range/Performance");
            dtcDictionary.Add("P0976", "Shift Solenoid B Control Circuit Low");
            dtcDictionary.Add("P0977", "Shift Solenoid B Control Circuit High");
            dtcDictionary.Add("P0978", "Shift Solenoid C Control Circuit Range/Performance");
            dtcDictionary.Add("P0979", "Shift Solenoid C Control Circuit Low");
            dtcDictionary.Add("P0980", "Shift Solenoid C Control Circuit High");
            dtcDictionary.Add("P0981", "Shift Solenoid D Control Circuit Range/Performance");
            dtcDictionary.Add("P0982", "Shift Solenoid D Control Circuit Low");
            dtcDictionary.Add("P0983", "Shift Solenoid D Control Circuit High");
            dtcDictionary.Add("P0984", "Shift Solenoid E Control Circuit Range/Performance");
            dtcDictionary.Add("P0985", "Shift Solenoid E Control Circuit Low");
            dtcDictionary.Add("P0986", "Shift Solenoid E Control Circuit High");
            dtcDictionary.Add("P0987", "Transmission Fluid Pressure Sensor/Switch E Circuit");
            dtcDictionary.Add("P0988", "Transmission Fluid Pressure Sensor/Switch E Circuit Range/Performance");
            dtcDictionary.Add("P0989", "Transmission Fluid Pressure Sensor/Switch E Circuit Low");
            dtcDictionary.Add("P0990", "Transmission Fluid Pressure Sensor/Switch E Circuit High");
            dtcDictionary.Add("P0991", "Transmission Fluid Pressure Sensor/Switch E Circuit Intermittent");
            dtcDictionary.Add("P0992", "Transmission Fluid Pressure Sensor/Switch F Circuit");
            dtcDictionary.Add("P0993", "Transmission Fluid Pressure Sensor/Switch F Circuit Range/Performance");
            dtcDictionary.Add("P0994", "Transmission Fluid Pressure Sensor/Switch F Circuit Low");
            dtcDictionary.Add("P0995", "Transmission Fluid Pressure Sensor/Switch F Circuit High");
            dtcDictionary.Add("P0996", "Transmission Fluid Pressure Sensor/Switch F Circuit Intermittent");
            dtcDictionary.Add("P0997", "Shift Solenoid F Control Circuit Range/Performance");
            dtcDictionary.Add("P0998", "Shift Solenoid F Control Circuit Low");
            dtcDictionary.Add("P0999", "Shift Solenoid F Control Circuit High");

            dtcDictionary.Add("P1100", "MAF Sensor Intermittent/ Check of all OBDII Systems Not Complete");
            dtcDictionary.Add("P1101", "MAF Sensor Out Of Self Test Range./KOER Not Able To Complete KOER Aborted");
            dtcDictionary.Add("P1102", "MAF Sensor In Range But Lower Than Expected");
            dtcDictionary.Add("P1103", "MAF Sensor In Range But Higher Than Expected");
            dtcDictionary.Add("P1104", "MAF Ground Malfunction");
            dtcDictionary.Add("P1105", "Dual Alternator Upper Fault");
            dtcDictionary.Add(
                "P1106",
                "Dual Alternator Lower Fault/ Manifold Absolute Pressure (MAP) Sensor Circuit Intermittent High Voltage");
            dtcDictionary.Add(
                "P1107",
                "Dual Alternator Lower Circuit Malfunction/ Manifold Absolute Pressure (MAP) Sensor Circuit Intermittent Low Voltage");
            dtcDictionary.Add("P1108", "Dual Alternator Battery Lamp Circuit Malfunction");
            dtcDictionary.Add("P1109", "IAT - B Sensor Intermittent");
            dtcDictionary.Add("P1110", "IAT Sensor (D/C) Open/Short ");
            dtcDictionary.Add("P1111", "Intake Air Temperature (IAT) Sensor Circuit Intermittent High Voltage ");
            dtcDictionary.Add("P1112", "Intake Air Temperature (IAT) Sensor Circuit Intermittent Low Voltage ");
            dtcDictionary.Add("P1113", "IAT Sensor Open/Short ");
            dtcDictionary.Add(
                "P1114",
                "Engine Coolant Temperature (ECT) Sensor Circuit Intermittent Low Voltage/IAT - B Circuit Low Input ");
            dtcDictionary.Add(
                "P1115",
                "Engine Coolant Temperature (ECT) Sensor Circuit Intermittent High Voltage/IAT - B Circuit High Input ");
            dtcDictionary.Add("P1116", "Engine Coolant sensor out of range/ECT Sensor Out Of Self Test Range ");
            dtcDictionary.Add("P1117", "Engine Coolant Sensor intermittent/ECT Sensor Intermittent");
            dtcDictionary.Add("P1118", "Manifold Absolute Temperature Circuit Low Input");
            dtcDictionary.Add("P1119", "Manifold Absolute Temperature Circuit High Input");
            dtcDictionary.Add("P1120", "Throttle position sensor out of range");
            dtcDictionary.Add("P1121", "Throttle Position (TP) Sensor Circuit Intermittent High Voltage");
            dtcDictionary.Add("P1122", "Throttle Position (TP) Sensor Circuit Intermittent Low Voltage");
            dtcDictionary.Add("P1123", "Throttle Position Sensor In Range But Higher Than Expected");
            dtcDictionary.Add("P1124", "Throttle Position Sensor Out Of Self Test Range");
            dtcDictionary.Add("P1125", "Throttle position sensor intermittent");
            dtcDictionary.Add("P1126", "Throttle Position (Narrow Range) Sensor Circuit Malfunction");
            dtcDictionary.Add("P1127", "Exhaust Not Warm, Downstream O2 Sensor");
            dtcDictionary.Add("P1128", "Upstream Heated O2 Sensors Swapped");
            dtcDictionary.Add("P1129", "Downstream Heated O2 Sensors Swapped");
            dtcDictionary.Add("P1130", "Lack Of HO2S Switch - Adaptive Fuel At Limit");
            dtcDictionary.Add("P1131", "Lack Of HO2S Switch - Sensor Indicates Lean");
            dtcDictionary.Add("P1132", "Lack Of HO2S Switch - Sensor Indicates Rich");
            dtcDictionary.Add("P1133", "HO2S Insufficient Switching Sensor 1");
            dtcDictionary.Add("P1134", "HO2S Transition Time Ratio Sensor 1");
            dtcDictionary.Add("P1135", "Pedal Position Sensor A Circuit Intermittent");
            dtcDictionary.Add("P1136", "Fan Control Circuit Malfunction");
            dtcDictionary.Add("P1137", "Lack Of HO2S Switch - Sensor Indicates Lean");
            dtcDictionary.Add("P1138", "Lack Of HO2S12 Switch - Sensor Indicates Rich");
            dtcDictionary.Add("P1139", "Water In Fuel Indicator Circuit Malfunction");
            dtcDictionary.Add("P1140", "Water In Fuel Condition");
            dtcDictionary.Add("P1141", "Fuel Restriction Indicator Circuit Malfunction");
            dtcDictionary.Add("P1142", "Fuel Restriction Condition");
            dtcDictionary.Add("P1143", "Air Assist Control Valve Range/Performance");
            dtcDictionary.Add("P1144", "Air Assist Control Valve Circuit Malfunction");
            dtcDictionary.Add("P1150", "Lack Of HO2S21 Switch - Adaptive Fuel At Limit");
            dtcDictionary.Add("P1151", "Lack Of HO2S21 Switch - Sensor Indicates Lean");
            dtcDictionary.Add("P1152", "Lack Of HO2S21 Switch - Sensor Indicates Rich");
            dtcDictionary.Add("P1153", "Bank 2 Fuel Control Shifted Lean");
            dtcDictionary.Add("P1154", "Bank 2 Fuel Control Shifted Rich");
            dtcDictionary.Add("P1155", "Alternative Fuel Controller");
            dtcDictionary.Add("P1156", "Fuel Select Switch Malfunction");
            dtcDictionary.Add("P1157", "Lack Of HO2S22 Switch - Sensor Indicates Lean");
            dtcDictionary.Add("P1158", "Lack Of HO2S22 Switch - Sensor Indicates Rich");
            dtcDictionary.Add("P1159", "Fuel Stepper Motor Malfunction");
            dtcDictionary.Add("P1167", "Invalid Test,throttle not depressed");
            dtcDictionary.Add("P1168", "Fuel Rail Sensor In-Range Low Failure");
            dtcDictionary.Add("P1169", "Fuel Rail Sensor In-Range High Failure");
            dtcDictionary.Add("P1170", "ESO - Engine Shut Off Solenoid Fault");
            dtcDictionary.Add("P1171", "Rotor Sensor Fault");
            dtcDictionary.Add("P1172", "Rotor Control Fault");
            dtcDictionary.Add("P1173", "Rotor Calibration Fault");
            dtcDictionary.Add("P1174", "Cam Sensor Fault");
            dtcDictionary.Add("P1175", "Cam Control Fault");
            dtcDictionary.Add("P1176", "Cam Calibration Fault");
            dtcDictionary.Add("P1177", "Synchronization Fault");
            dtcDictionary.Add("P1178", "( open )");
            dtcDictionary.Add("P1180", "Fuel Delivery System Malfunction - Low");
            dtcDictionary.Add("P1181", "Fuel Delivery System Malfunction - High");
            dtcDictionary.Add("P1182", "Fuel Shut Off Solenoid Malfunction");
            dtcDictionary.Add("P1183", "Engine Oil Temperature Circuit Malfunction");
            dtcDictionary.Add("P1184", "Engine Oil Temperature Out Of Self Test Range");
            dtcDictionary.Add("P1185", "FTS High - Fuel Pump Temperature Sensor High");
            dtcDictionary.Add("P1186", "FTS Low - Fuel Pump Temperature Sensor Low");
            dtcDictionary.Add("P1187", "Variant Selection");
            dtcDictionary.Add("P1188", "Calibration Memory Fault");
            dtcDictionary.Add("P1189", "Pump Speed Signal Fault");
            dtcDictionary.Add("P1190", "Calibration Resistor Out Of Range");
            dtcDictionary.Add("P1191", "Key Line Voltage");
            dtcDictionary.Add("P1192", "Voltage External");
            dtcDictionary.Add("P1193", "EGR Drive Overcurrent");
            dtcDictionary.Add("P1194", "ECU A/D Converter");
            dtcDictionary.Add("P1195", "SCP HBCC Failed To Initialize");
            dtcDictionary.Add("P1196", "Key Off Voltage High");
            dtcDictionary.Add("P1197", "Key Off Voltage Low");
            dtcDictionary.Add("P1198", "Pump Rotor Control Underfueling");
            dtcDictionary.Add("P1199", "Fuel Level Input Circuit Low");
            dtcDictionary.Add("P1200", "Injector Control Circuit");
            dtcDictionary.Add("P1201", "Injector Circuit Open / Shorted - Cylinder #1");
            dtcDictionary.Add("P1202", "Injector Circuit Open / Shorted - Cylinder #2");
            dtcDictionary.Add("P1203", "Injector Circuit Open / Shorted - Cylinder #3");
            dtcDictionary.Add("P1204", "Injector Circuit Open / Shorted - Cylinder #4");
            dtcDictionary.Add("P1205", "Injector Circuit Open / Shorted - Cylinder #5");
            dtcDictionary.Add("P1206", "Injector Circuit Open / Shorted - Cylinder #6");
            dtcDictionary.Add("P1209", "Injector Control Pressure System Fault");
            dtcDictionary.Add("P1210", "Injector Control Pressure Above Expected Level");
            dtcDictionary.Add("P1211", "Injector Control Pressure Sensor Above / Below Desired");
            dtcDictionary.Add("P1212", "Injector Control Pressure Not Detected During Crank");
            dtcDictionary.Add("P1213", "Start Injector Circuit Malfunction");
            dtcDictionary.Add("P1214", "Pedal Position Sensor B Circuit Intermittent");
            dtcDictionary.Add("P1215", "Pedal Position Sensor C Circuit Low Input");
            dtcDictionary.Add("P1216", "Pedal Position Sensor C Circuit High Input");
            dtcDictionary.Add("P1217", "Pedal Position Sensor C Circuit Intermittent");
            dtcDictionary.Add("P1218", "CID High");
            dtcDictionary.Add("P1219", "CID Low");
            dtcDictionary.Add("P1220", "Series Throttle Control System Malfunction");
            dtcDictionary.Add("P1221", "Traction Control System Malfunction");
            dtcDictionary.Add("P1222", "Traction Control Output Circuit Malfunction");
            dtcDictionary.Add("P1223", "Pedal Demand Sensor B Circuit High Input");
            dtcDictionary.Add("P1224", "Throttle Position Sensor B Out Of Self Test Range");
            dtcDictionary.Add("P1225", "Needle Lift Sensor Malfunction");
            dtcDictionary.Add("P1226", "Control Sleeve Sensor Malfunction");
            dtcDictionary.Add("P1227", "Wastegate Failed Closed (Over Pressure)");
            dtcDictionary.Add("P1228", "Wastegate Failed Open (Under Pressure)");
            dtcDictionary.Add("P1229", "Intercooler Pump Driver Fault");
            dtcDictionary.Add("P1230", "Fuel Pump Low Speed Malfunction");
            dtcDictionary.Add("P1231", "Fuel Pump Secondary Circuit Low, High Speed");
            dtcDictionary.Add("P1232", "Fuel Pump Speed Primary Circuit Malfunction");
            dtcDictionary.Add("P1233", "Fuel Pump Driver Module Off Line");
            dtcDictionary.Add("P1234", "Fuel Pump Driver Module Off Line");
            dtcDictionary.Add("P1235", "Fuel Pump Control Out Of Range");
            dtcDictionary.Add("P1236", "Fuel Pump Control Out Of Range");
            dtcDictionary.Add("P1237", "Fuel Pump Secondary Circuit Malfunction");
            dtcDictionary.Add("P1238", "Fuel Pump Secondary Circuit Malfunction");
            dtcDictionary.Add("P1239", "Speed Fuel Pump Positive Feed Fault");
            dtcDictionary.Add("P1240", "Sensor Power Supply Malfunction");
            dtcDictionary.Add("P1241", "Sensor Power Supply Low Input");
            dtcDictionary.Add("P1242", "Sensor Power Supply High Input");
            dtcDictionary.Add("P1243", "Second Fuel Pump Faulty or Ground Fault");
            dtcDictionary.Add("P1244", "Alternator Load Input Failed High");
            dtcDictionary.Add("P1245", "Alternator Load Input Failed Low");
            dtcDictionary.Add("P1246", "Alternator Load Input Failed");
            dtcDictionary.Add("P1247", "Turbo Boost Pressure Low");
            dtcDictionary.Add("P1248", "Turbo Boost Pressure Not Detected");
            dtcDictionary.Add("P1249", "Wastegate Control Valve Performance");
            dtcDictionary.Add("P1250", "PRC Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1251", "Air Mixture Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1252", "Pedal Correlation PDS1 and LPDS High");
            dtcDictionary.Add("P1253", "Pedal Correlation PDS1 and LPDS Low");
            dtcDictionary.Add("P1254", "Pedal Correlation PDS2 and LPDS High");
            dtcDictionary.Add("P1255", "Pedal Correlation PDS2 and LPDS Low");
            dtcDictionary.Add("P1256", "Pedal Correlation PDS1 and HPDS");
            dtcDictionary.Add("P1257", "Pedal Correlation PDS2 and HPDS");
            dtcDictionary.Add("P1258", "Pedal Correlation PDS1 and PDS2");
            dtcDictionary.Add("P1259", "Immobilizer to PCM Signal Error");
            dtcDictionary.Add("P1260", "THEFT Detected, Vehicle Immobilzed");
            dtcDictionary.Add("P1261", "Cylinder #1 High To Low Side Short");
            dtcDictionary.Add("P1262", "Cylinder #2 High To Low Side Short");
            dtcDictionary.Add("P1263", "Cylinder #3 High To Low Side Short");
            dtcDictionary.Add("P1264", "Cylinder #4 High To Low Side Short");
            dtcDictionary.Add("P1265", "Cylinder #5 High To Low Side Short");
            dtcDictionary.Add("P1266", "Cylinder #6 High To Low Side Short");
            dtcDictionary.Add("P1267", "Cylinder #7 High To Low Side Short");
            dtcDictionary.Add("P1268", "Cylinder #8 High To Low Side Short");
            dtcDictionary.Add("P1269", "Immobilizer Code Not Programmed");
            dtcDictionary.Add("P1270", "Engine RPM Or Speed Limiter Reached");
            dtcDictionary.Add("P1271", "Cylinder #1 High To Low Side Open");
            dtcDictionary.Add("P1272", "Cylinder #2 High To Low Side Open");
            dtcDictionary.Add("P1273", "Cylinder #3 High To Low Side Open");
            dtcDictionary.Add("P1274", "Cylinder #4 High To Low Side Open");
            dtcDictionary.Add("P1275", "Cylinder #5 High To Low Side Open");
            dtcDictionary.Add("P1276", "Cylinder #6 High To Low Side Open");
            dtcDictionary.Add("P1277", "Cylinder #7 High To Low Side Open");
            dtcDictionary.Add("P1278", "Cylinder #8 High To Low Side Open");
            dtcDictionary.Add("P1280", "Injection Control Pressure Out Of Range Low");
            dtcDictionary.Add("P1281", "Injection Control Pressure Out Of Range High");
            dtcDictionary.Add("P1282", "Excessive Injection Control Pressure");
            dtcDictionary.Add("P1283", "IPR Circuit Failure");
            dtcDictionary.Add("P1284", "Aborted KOER - ICP Failure");
            dtcDictionary.Add("P1285", "Cylinder head over temp sensed");
            dtcDictionary.Add("P1286", "Fuel Pulse In Range But Lower Than Expected");
            dtcDictionary.Add("P1287", "Fuel Pulse In Range But Higher Than Expected");
            dtcDictionary.Add("P1288", "Cylinder Head Temp Sensor Out Of Self Test Range");
            dtcDictionary.Add("P1289", "Cylinder Head Temp Sensor High Input");
            dtcDictionary.Add("P1290", "Cylinder Head Temp Sensor Low Input");
            dtcDictionary.Add("P1291", "Injector High Side Short To GND Or VBATT - Bank 1");
            dtcDictionary.Add("P1292", "Injector High Side Short To GND Or VBATT - Bank 2");
            dtcDictionary.Add("P1293", "Injector High Side Open - Bank 1");
            dtcDictionary.Add("P1294", "Injector High Side Open - Bank 2/Target idle not reached");
            dtcDictionary.Add("P1295", "Multi-faults - Bank 1 - With Low Side Shorts");
            dtcDictionary.Add("P1296", "Multi-faults - Bank 2 - With Low Side Shorts");
            dtcDictionary.Add("P1297", "Injector High Sides Shorted Together");
            dtcDictionary.Add("P1298", "IDM Failure");
            dtcDictionary.Add("P1299", "Cylinder Head Overtemperature Protection Active");
            dtcDictionary.Add("P1300", "Boost Calibration Fault");
            dtcDictionary.Add("P1301", "Boost Calibration High");
            dtcDictionary.Add("P1302", "Boost Calibration Low");
            dtcDictionary.Add("P1303", "EGR Calibration Fault");
            dtcDictionary.Add("P1304", "EGR Calibration High");
            dtcDictionary.Add("P1305", "EGR Calibration Low");
            dtcDictionary.Add("P1306", "Kickdown Relay Pull - In Circuit Fault");
            dtcDictionary.Add("P1307", "Kickdown Relay Hold Circuit Fault");
            dtcDictionary.Add("P1308", "A/C Clutch Circuit Fault");
            dtcDictionary.Add("P1309", "Misfire Monitor AICE Chip Fault");
            dtcDictionary.Add("P1313", "Misfire Rate Catalyst Damage Fault - Bank 1");
            dtcDictionary.Add("P1314", "Misfire Rate Catalyst Damage Fault - Bank 2");
            dtcDictionary.Add("P1315", "Persistent Misfire");
            dtcDictionary.Add("P1316", "Injector Circuit / IDM Codes Detected");
            dtcDictionary.Add("P1317", "Injector Circuit / IDM Codes Not Updated");
            dtcDictionary.Add("P1336", "Crank / Cam Sensor Range / Performance");
            dtcDictionary.Add("P1340", "Camshaft Position Sensor B Circuit Malfunction");
            dtcDictionary.Add("P1341", "Camshaft Position Sensor B Range / Performance");
            dtcDictionary.Add(
                "P1345",
                "SGC (Cam Position) Sensor Circuit Malfunction/ Crankshaft Position - Camshaft Position Correlation");
            dtcDictionary.Add("P1346", "Fuel Level Sensor B Circuit Malfunction");
            dtcDictionary.Add("P1347", "Fuel Level Sensor B Range / Performance");
            dtcDictionary.Add("P1348", "Fuel Level Sensor B Circuit Low");
            dtcDictionary.Add("P1349", "Fuel Level Sensor B Circuit High");
            dtcDictionary.Add("P1350", "Fuel Level Sensor B Intermittent/Bypass Line Monitor");
            dtcDictionary.Add("P1351", "IDM Input Circuit Malfunction/ Ignition Coil Control Circuit High Voltage");
            dtcDictionary.Add("P1352", "Ignition Coil A Primary Circuit Malfunction");
            dtcDictionary.Add("P1353", "Ignition Coil B Primary Circuit Malfunction");
            dtcDictionary.Add("P1354", "Ignition Coil C Primary Circuit Malfunction");
            dtcDictionary.Add("P1355", "Ignition Coil D Primary Circuit Malfunction");
            dtcDictionary.Add("P1360", "Ignition Coil A Secondary Circuit Malfunction");
            dtcDictionary.Add("P1361", "Ignition Control (IC) Circuit Low Voltage");
            dtcDictionary.Add("P1362", "Ignition Coil C Secondary Circuit Malfunction");
            dtcDictionary.Add("P1363", "Ignition Coil D Secondary Circuit Malfunction");
            dtcDictionary.Add("P1364", "Ignition Coil Primary Circuit Failure");
            dtcDictionary.Add("P1365", "Ignition Coil Secondary Circuit Failure");
            dtcDictionary.Add("P1366", "Ignition Spare");
            dtcDictionary.Add("P1367", "Ignition Spare");
            dtcDictionary.Add("P1368", "Ignition Spare");
            dtcDictionary.Add("P1369", "Engine Temperature Light Monitor Failure");
            dtcDictionary.Add("P1370", "Insufficient RMP Increase During Spark Test");
            dtcDictionary.Add("P1371", "Ignition Coil - Cylinder 1 - Early Activation Fault");
            dtcDictionary.Add("P1372", "Ignition Coil - Cylinder 2 - Early Activation Fault");
            dtcDictionary.Add("P1373", "Ignition Coil - Cylinder 3 - Early Activation Fault");
            dtcDictionary.Add("P1374", "Crankshaft Position (CKP)/Ignition Coil - Cylinder 4 - Early Activation Fault");
            dtcDictionary.Add("P1375", "Ignition Coil - Cylinder 5 - Early Activation Fault");
            dtcDictionary.Add("P1376", "Ignition Coil - Cylinder 6 - Early Activation Fault");
            dtcDictionary.Add("P1380", "Misfire Detected - Rough Road Data Not Available");
            dtcDictionary.Add(
                "P1381",
                "Variable Cam Timing Overadvanced (Bank #1)/ Misfire Detected - No Communication with BCM");
            dtcDictionary.Add("P1382", "Variable Cam Timing Solenoid #1 Circuit Malfunction");
            dtcDictionary.Add("P1383", "Variable Cam Timing Overretarded (Bank #1)");
            dtcDictionary.Add("P1384", "VVT Solenoid A Malfunction");
            dtcDictionary.Add("P1385", "Variable Cam Timing Solenoid B Malfunction");
            dtcDictionary.Add("P1386", "Variable Cam Timing Overadvanced (Bank #2)");
            dtcDictionary.Add("P1387", "Variable Cam Timing Solenoid #2 Circuit Malfunction");
            dtcDictionary.Add("P1388", "Variable Cam Timing Overretarded (Bank #2)");
            dtcDictionary.Add("P1389", "Glow Plug Circuit High Side Low Input");
            dtcDictionary.Add("P1390", "Octane Adjust Pin Out Of Self Test Range");
            dtcDictionary.Add("P1391", "Glow Plug Circuit Low Input (Bank #1)");
            dtcDictionary.Add("P1392", "Glow Plug Circuit High Input (Bank #1)");
            dtcDictionary.Add("P1393", "Glow Plug Circuit Low Input (Bank #2)");
            dtcDictionary.Add("P1394", "Glow Plug Circuit High Input (Bank #2)");
            dtcDictionary.Add("P1395", "Glow Plug Monitor Fault (Bank #1)");
            dtcDictionary.Add("P1396", "Glow Plug Monitor Fault (Bank #2)");
            dtcDictionary.Add("P1397", "System Voltage Out Of Self Test Range");
            dtcDictionary.Add("P1398", "VVT Solenoid B Circuit High Input");
            dtcDictionary.Add("P1399", "Glow Plug Circuit High Side, High Input");
            dtcDictionary.Add("P1400", "DPFE Circuit Low Input");
            dtcDictionary.Add("P1401", "DPFE Circuit High Input");
            dtcDictionary.Add("P1402", "EGR Metering Orifice Restricted");
            dtcDictionary.Add("P1403", "DPFE Sensor Hoses Reversed");
            dtcDictionary.Add(
                "P1404",
                "IAT - B Circuit Malfunction/ Exhaust Gas Recirculation Closed Position Performance");
            dtcDictionary.Add("P1405", "DPFE Sensor Upstream Hose Off Or Plugged");
            dtcDictionary.Add("P1406", "Exhaust Gas Recirculation (EGR) Position Sensor Performance");
            dtcDictionary.Add("P1407", "EGR No Flow Detected");
            dtcDictionary.Add("P1408", "EGR Flow Out Of Self Test Range");
            dtcDictionary.Add("P1409", "EVR Control Circuit Malfunction");
            dtcDictionary.Add("P1411", "SAI System Incorrect Downstream Flow Detected");
            dtcDictionary.Add("P1413", "SAI System Monitor Circuit Low Input");
            dtcDictionary.Add("P1414", "SAI System Monitor Circuit High Input");
            dtcDictionary.Add("P1415", "Air Pump Circuit Malfunction/ (AIR) System Bank 1");
            dtcDictionary.Add("P1416", "Port Air Circuit Malfunction/ (AIR) System Bank 2");
            dtcDictionary.Add("P1417", "Port Air Relief Circuit Malfunction");
            dtcDictionary.Add("P1418", "Split Air #1 Circuit Malfunction");
            dtcDictionary.Add("P1419", "Split Air #2 Circuit Malfunction");
            dtcDictionary.Add("P1420", "Catalyst Temperature Sensor Failure");
            dtcDictionary.Add("P1421", "Catalyst Damage");
            dtcDictionary.Add("P1422", "EGI Temperature Sensor Failure");
            dtcDictionary.Add("P1423", "EGI Functionality Test Failed");
            dtcDictionary.Add("P1424", "EGI Glow Plug Primary Failure");
            dtcDictionary.Add("P1425", "EGI Glow Plug Secondary Failure");
            dtcDictionary.Add("P1426", "EGI Mini - MAF Failed Out Of Range");
            dtcDictionary.Add("P1427", "EGI Mini - MAF Failed Short Circuit");
            dtcDictionary.Add("P1428", "EGI Mini - MAF Failed Open Circuit");
            dtcDictionary.Add("P1429", "Electric Air Pump Primary Failure");
            dtcDictionary.Add("P1430", "Electric Air Pump Secondary Failure");
            dtcDictionary.Add("P1433", "A/C Refrigerant Temperature Circuit Low");
            dtcDictionary.Add("P1434", "A/C Refrigerant Temperature Circuit High");
            dtcDictionary.Add("P1435", "A/C Refrigerant Temperature Circuit Range/Performance");
            dtcDictionary.Add("P1436", "A/C Evaporator Air Temperature Circuit Low");
            dtcDictionary.Add("P1437", "A/C Evaporator Air Temperature Circuit High");
            dtcDictionary.Add("P1438", "A/C Evaporator Air Temperature Circuit Range/Performance");
            dtcDictionary.Add("P1439", "Floor Temperature Switch Circuit Malfunction");
            dtcDictionary.Add("P1440", "Purge Valve Stuck Open");
            dtcDictionary.Add("P1441", "Evaporative Emission (EVAP) System Flow During Non-Purge Chevrolet Only");
            dtcDictionary.Add("P1442", "Evaporative Emission Control System Leak Detected");
            dtcDictionary.Add("P1443", "Evaporative Emission Control System Control Valve");
            dtcDictionary.Add("P1444", "Purge Flow Sensor Circuit Low Input");
            dtcDictionary.Add("P1445", "Purge Flow Sensor Circuit High Input");
            dtcDictionary.Add("P1446", "Evaporative Vac Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1447", "ELC System Closure Valve Flow Fault");
            dtcDictionary.Add("P1448", "ELC System 2 Fault");
            dtcDictionary.Add("P1449", "Evaporative Check Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1450", "Unable To Bleed Up Fuel Tank Vacuum");
            dtcDictionary.Add("P1451", "Evap Emission Control Sys Vent Control Valve Circuit");
            dtcDictionary.Add("P1452", "Unable To Bleed - Up Vacuum in Tank");
            dtcDictionary.Add("P1453", "Fuel Tank Pressure Relief Valve Malfunction");
            dtcDictionary.Add("P1454", "Evaporative System Vacuum Test Malfunction");
            dtcDictionary.Add("P1455", "Evap Emission Control Sys Leak Detected (Gross Leak/No Flow)");
            dtcDictionary.Add("P1456", "Fuel Tank Temperature Sensor Circuit Malfunction");
            dtcDictionary.Add("P1457", "Unable To Pull Vacuum In Tank");
            dtcDictionary.Add("P1460", "Wide open throttle A/C cutoff relay circuit");
            dtcDictionary.Add("P1461", "A/C pressure sensor circuit voltage low");
            dtcDictionary.Add("P1462", "A/C pressure sensor circuit voltage high");
            dtcDictionary.Add("P1463", "A/C Pressure Sensor Insufficient Pressure Change");
            dtcDictionary.Add("P1464", "A/C Demand Out of Self Test Range");
            dtcDictionary.Add("P1465", "A/C Relay Circuit Malfunction");
            dtcDictionary.Add("P1466", "A/C Refrigerant Temperature Sensor/Circuit Malfunction");
            dtcDictionary.Add("P1467", "A/C Compressor Temperature Sensor Malfunction");
            dtcDictionary.Add("P1468", "SSPOD Open Circuit or Closed Circuit Fault");
            dtcDictionary.Add("P1469", "Low A/C Cycling Period");
            dtcDictionary.Add("P1470", "A/C Cycling Period Too Short");
            dtcDictionary.Add("P1471", "Electrodrive Fan 1 Operational Failure (Driver Side)");
            dtcDictionary.Add("P1472", "Electrodrive Fan 2 Operational Failure (Passenger Side)");
            dtcDictionary.Add("P1473", "Fan Secondary High With Fan(s) Off");
            dtcDictionary.Add("P1474", "Low Fan Control Primary Circuit Malfunction");
            dtcDictionary.Add("P1475", "Fan Relay (Low) Circuit Malfunction");
            dtcDictionary.Add("P1476", "Fan Relay (High) Circuit Malfunction");
            dtcDictionary.Add("P1477", "Additional Fan Relay Circuit Malfunction");
            dtcDictionary.Add("P1478", "Cooling Fan Driver Fault");
            dtcDictionary.Add("P1479", "High Fan Control Primary Circuit Malfunction");
            dtcDictionary.Add("P1480", "Fan Secondary Low with Low Fan On");
            dtcDictionary.Add("P1481", "Fan Secondary Low With High Fan On");
            dtcDictionary.Add("P1482", "SCP");
            dtcDictionary.Add("P1483", "Power To Fan Circuit Overcurrent");
            dtcDictionary.Add("P1484", "Open Power To Ground VCRM");
            dtcDictionary.Add("P1485", "EGRV Circuit Malfunction");
            dtcDictionary.Add("P1486", "EGRA Circuit Malfunction");
            dtcDictionary.Add("P1487", "EGRCHK Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1490", "Secondary Air Relief Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1491", "Secondary Switch Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1492", "APLSOL Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1493", "RCNT Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1494", "SPCUT Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1495", "TCSPL Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1500", "Vehicle Speed Sensor Intermittent");
            dtcDictionary.Add("P1501", "Vehicle Speed Sensor Out Of Self Test Range");
            dtcDictionary.Add("P1502", "Vehicle Speed Sensor Intermittent Malfunction");
            dtcDictionary.Add("P1503", "Auxillary Speed Sensor Fault");
            dtcDictionary.Add("P1504", "Idle Air Control Circuit Malfunction");
            dtcDictionary.Add("P1505", "Idle Air Control System At Adaptive Clip");
            dtcDictionary.Add("P1506", "Idle Air Control Overspeed Error");
            dtcDictionary.Add("P1507", "Idle Air Control Underspeed Error");
            dtcDictionary.Add("P1508", "Idle Control System Circuit Open");
            dtcDictionary.Add("P1509", "Idle Control System Circuit Shorted");
            dtcDictionary.Add("P1510", "Idle Signal Circuit Malfunction");
            dtcDictionary.Add("P1511", "Idle Switch (Electric Control Throttle) Circuit Malfunction");
            dtcDictionary.Add("P1512", "Intake Manifold Runner Control (Bank 1) Stuck Closed");
            dtcDictionary.Add("P1513", "Intake Manifold Runner Control (Bank 2) Stuck Closed");
            dtcDictionary.Add("P1514", "High Load Neutral/Drive Fault");
            dtcDictionary.Add("P1515", "Electric Current Circuit Malfunction");
            dtcDictionary.Add("P1516", "IMRC Input Error (Bank 1)");
            dtcDictionary.Add("P1517", "IMRC Input Error (Bank 2)");
            dtcDictionary.Add("P1518", "Intake Manifold Runner Control (Stuck Open)");
            dtcDictionary.Add("P1519", "Intake Manifold Runner Control (Stuck Closed)");
            dtcDictionary.Add("P1520", "Intake Manifold Runner Control Circuit Malfunction");
            dtcDictionary.Add("P1521", "Variable Intake Solenoid #1 Circuit Malfunction");
            dtcDictionary.Add("P1522", "Variable Intake Solenoid #2 Circuit Malfunction");
            dtcDictionary.Add("P1523", "IVC Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1524", "Variable Intake Solenoid System");
            dtcDictionary.Add("P1525", "Air Bypass Valve System");
            dtcDictionary.Add("P1526", "Air Bypass System");
            dtcDictionary.Add("P1527", "Accelerate Warmup Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1528", "Subsidiary Throttle Valve Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1529", "SCAIR Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1530", "A/C Clutch Circuit Malfunction");
            dtcDictionary.Add("P1531", "Invalid Test - Accelerator Pedal Movement");
            dtcDictionary.Add("P1532", "IMCC Circuit Malfunction, Bank B");
            dtcDictionary.Add("P1533", "AAI Circuit Malfunction");
            dtcDictionary.Add("P1534", "Inertia Switch Activated");
            dtcDictionary.Add("P1535", "Blower Fan Speed Circuit Range/Performance");
            dtcDictionary.Add("P1536", "Parking Brake Switch Circuit Failure");
            dtcDictionary.Add("P1537", "Intake Manifold Runner Control (Bank 1) Stuck Open");
            dtcDictionary.Add("P1538", "Intake Manifold Runner Control (Bank 2) Stuck Open");
            dtcDictionary.Add("P1539", "Power To A/C Clutch Circuit Overcurrent");
            dtcDictionary.Add("P1540", "Air Bypass Valve Circuit Malfunction");
            dtcDictionary.Add("P1549", "IMCC Circuit Malfunction, Bank B");
            dtcDictionary.Add("P1550", "PSPS Out Of Self Test Range");
            dtcDictionary.Add("P1565", "Speed Control Command Switch Out of Range High");
            dtcDictionary.Add("P1566", "Speed Control Command Switch Out of Range Low");
            dtcDictionary.Add("P1567", "Speed Control Output Circuit Continuity");
            dtcDictionary.Add("P1568", "Speed Control Unable to Hold Speed");
            dtcDictionary.Add("P1571", "Brake Switch Malfunction");
            dtcDictionary.Add("P1572", "Brake Pedal Switch Circuit Malfunction");
            dtcDictionary.Add("P1573", "Throttle Position Not Available");
            dtcDictionary.Add("P1574", "Throttle Position Sensor Disagreement btwn Sensors");
            dtcDictionary.Add("P1575", "Pedal Position Out of Self Test Range");
            dtcDictionary.Add("P1576", "Pedal Position Not Available");
            dtcDictionary.Add("P1577", "Pedal Position Sensor Disagreement btwn Sensors");
            dtcDictionary.Add("P1578", "ETC Power Less Than Demand");
            dtcDictionary.Add("P1579", "ETC In Power Limiting Mode");
            dtcDictionary.Add("P1580", "Electronic Throttle Monitor PCM Override");
            dtcDictionary.Add("P1581", "Electronic Throttle Monitor Malfunction");
            dtcDictionary.Add("P1582", "Electronic Throttle Monitor Data Available");
            dtcDictionary.Add("P1583", "Electronic Throttle Monitor Cruise Disable");
            dtcDictionary.Add("P1584", "TCU Detected IPE Circuit Malfunction");
            dtcDictionary.Add("P1585", "Throttle Control Unit Malfunction");
            dtcDictionary.Add("P1586", "Throttle Control Unit Throttle Position Malfunction");
            dtcDictionary.Add("P1587", "Throttle Control Unit Modulated Command Malfunction");
            dtcDictionary.Add("P1588", "Throttle Control Unit Detected Loss of Return Spring");
            dtcDictionary.Add("P1589", "TCU Unable To Control Desired Throttle Angle");
            dtcDictionary.Add("P1600", "Loss of KAM Power; Open Circuit");
            dtcDictionary.Add("P1601", "ECM/TCM Serial Communication Error");
            dtcDictionary.Add("P1602", "Immobilizer/ECM Communication Error");
            dtcDictionary.Add("P1603", "EEPROM Malfunction");
            dtcDictionary.Add("P1604", "Code Word Unregestered");
            dtcDictionary.Add("P1605", "Keep Alive Memory Test Failure");
            dtcDictionary.Add("P1606", "ECM Control Relay O/P Circuit Malfunction");
            dtcDictionary.Add("P1607", "MIL O/P Circuit Malfunction");
            dtcDictionary.Add("P1608", "Internal ECM Malfunction");
            dtcDictionary.Add("P1609", "Diagnostic Lamp Driver Fault");
            dtcDictionary.Add("P1610", "SBDS Interactive Codes");
            dtcDictionary.Add("P1611", "SBDS Interactive Codes");
            dtcDictionary.Add("P1612", "SBDS Interactive Codes");
            dtcDictionary.Add("P1613", "SBDS Interactive Codes");
            dtcDictionary.Add("P1614", "SBDS Interactive Codes");
            dtcDictionary.Add("P1615", "SBDS Interactive Codes");
            dtcDictionary.Add("P1616", "SBDS Interactive Codes");
            dtcDictionary.Add("P1617", "SBDS Interactive Codes");
            dtcDictionary.Add("P1618", "SBDS Interactive Codes");
            dtcDictionary.Add("P1619", "SBDS Interactive Codes");
            dtcDictionary.Add("P1620", "SBDS Interactive Codes");
            dtcDictionary.Add(
                "P1621",
                "Control Module Long Term Memory Performance/ Immobilizer Code Words Do Not Match");
            dtcDictionary.Add("P1622", "Immobilizer ID Does Not Match");
            dtcDictionary.Add("P1623", "Immobilizer Code Word/ID Number Write Failure");
            dtcDictionary.Add("P1624", "Anti Theft System");
            dtcDictionary.Add("P1625", "B+ Supply To VCRM Fan Circuit Malfunction");
            dtcDictionary.Add(
                "P1626",
                "Theft Deterrent Fuel Enable Signal Not Received/ B+ Supply To VCRM A/C Circuit Malfunction");
            dtcDictionary.Add("P1627", "Module Supply Voltage Out Of Range");
            dtcDictionary.Add("P1628", "Module Ignition Supply Input Malfunction");
            dtcDictionary.Add("P1629", "Internal Voltage Regulator Malfunction");
            dtcDictionary.Add("P1630", "Internal Vref Malfunction");
            dtcDictionary.Add(
                "P1631",
                "Theft Deterrent Start Enable Signal Not Correct/ Main Relay Malfunction (Power Hold)");
            dtcDictionary.Add("P1632", "Smart Alternator Faults Sensor/Circuit Malfunction");
            dtcDictionary.Add("P1633", "KAM Voltage Too Low");
            dtcDictionary.Add("P1634", "Data Output Link Circuit Failure");
            dtcDictionary.Add("P1635", "Tire / Axle Ratio Out of Acceptable Range");
            dtcDictionary.Add("P1636", "Inductive Signature Chip Communication Error");
            dtcDictionary.Add("P1637", "Can Link ECM/ABSCM Circuit / Network Malfunction");
            dtcDictionary.Add("P1638", "Can Link ECM/INSTM Circuit / Network Malfunction");
            dtcDictionary.Add("P1639", "Vehicle ID Block Corrupted or Not Programmed");
            dtcDictionary.Add("P1640", "Powertrain DTCs Available in Another Module");
            dtcDictionary.Add("P1641", "Fuel Pump Primary Circuit Failure");
            dtcDictionary.Add("P1642", "Fuel Pump Monitor Circuit High Input");
            dtcDictionary.Add("P1643", "Fuel Pump Monitor Circuit Low Input");
            dtcDictionary.Add("P1644", "Fuel Pump Speed Control Circuit Malfunction");
            dtcDictionary.Add("P1645", "Fuel Pump Resistor Switch Circuit Malfunction");
            dtcDictionary.Add("P1650", "PSP Switch Out of Self Test Range");
            dtcDictionary.Add("P1651", "PSP Switch Input Malfunction");
            dtcDictionary.Add("P1652", "IAC Monitor Disabled by PSP Switch Failed On");
            dtcDictionary.Add("P1653", "Power Steering Output Circuit Malfunction");
            dtcDictionary.Add("P1654", "Recirculation Override Circuit Malfunction");
            dtcDictionary.Add("P1655", "Starter Disable Circuit Malfunction");
            dtcDictionary.Add("P1660", "Output Circuit Check Signal High");
            dtcDictionary.Add("P1661", "Output Circuit Check Signal Low");
            dtcDictionary.Add("P1662", "IDM_EN Circuit Failure");
            dtcDictionary.Add("P1663", "Fuel Demand Command Signal Circuit Malfunction");
            dtcDictionary.Add("P1667", "CI Circuit Malfunction");
            dtcDictionary.Add("P1668", "PCM - IDM Communications Error");
            dtcDictionary.Add("P1670", "Electronic Feedback Signal Not Detected");
            dtcDictionary.Add("P1680", "Metering Oil Pump Malfunction");
            dtcDictionary.Add("P1681", "Metering Oil Pump Malfunction");
            dtcDictionary.Add("P1682", "Metering Oil Pump Malfunction");
            dtcDictionary.Add("P1683", "Metering Oil Pump Temperature Sensor Circuit Malfunction");
            dtcDictionary.Add("P1684", "Metering Oil Pump Position Sensor Circuit Malfunction");
            dtcDictionary.Add("P1685", "Metering Oil Pump Stepping Motor Circuit Malfunction");
            dtcDictionary.Add("P1686", "Metering Oil Pump Stepping Motor Circuit Malfunction");
            dtcDictionary.Add("P1687", "Metering Oil Pump Stepping Motor Circuit Malfunction");
            dtcDictionary.Add("P1688", "Metering Oil Pump Stepping Motor Circuit Malfunction");
            dtcDictionary.Add("P1689", "Oil Pressure Control Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1690", "Wastegate Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1691", "Turbo Pressure Control Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1692", "Turbo Control Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1693", "Turbo Charge Control Circuit Malfunction");
            dtcDictionary.Add("P1694", "Turbo Charge Relief Circuit Malfunction");
            dtcDictionary.Add("P1700", "Transmission Indeterminate Failure (Failed to Neutral)");
            dtcDictionary.Add("P1701", "Reverse Engagement Error");
            dtcDictionary.Add("P1702", "TRS Circuit Intermittent Malfunction");
            dtcDictionary.Add("P1703", "Brake Switch Out Of Self Test Range");
            dtcDictionary.Add("P1704", "Digital TRS Failed to Transition States in KOEO / KOER");
            dtcDictionary.Add("P1705", "Not in P or N During KOEO / KOER");
            dtcDictionary.Add("P1706", "High Vehicle Speed Observed in Park");
            dtcDictionary.Add("P1707", "Transfer Case Neutral Indicator Hard Fault Present");
            dtcDictionary.Add("P1708", "Clutch Switch Circuit Malfunction");
            dtcDictionary.Add("P1709", "PNP Switch Out Of Self Test Range");
            dtcDictionary.Add("P1711", "TFT Sensor Out Of Self Test Range");
            dtcDictionary.Add("P1712", "Trans Torque Reduction Request Signal Malfunction");
            dtcDictionary.Add("P1713", "TFT Sensor In Range Failure Low Value");
            dtcDictionary.Add("P1714", "SSA Inductive Signature Malfunction");
            dtcDictionary.Add("P1715", "SSB Inductive Signature Malfunction");
            dtcDictionary.Add("P1716", "SSC Inductive Signature Malfunction");
            dtcDictionary.Add("P1717", "SSD Inductive Signature Malfunction");
            dtcDictionary.Add("P1718", "TFT Sensor In Range Failure High");
            dtcDictionary.Add("P1720", "Vehicle Speed (Meter) Circuit Malfunction");
            dtcDictionary.Add("P1721", "Gear 1 Incorrect Ratio");
            dtcDictionary.Add("P1722", "Gear 2 Incorrect Ratio");
            dtcDictionary.Add("P1723", "Gear 3 incorrect Ratio");
            dtcDictionary.Add("P1724", "Gear 4 Incorrect Ratio");
            dtcDictionary.Add("P1725", "Insufficient Engine Speed Increase During Self Test");
            dtcDictionary.Add("P1726", "Insufficient Engine Speed Decrease During Self Test");
            dtcDictionary.Add("P1727", "Coast Clutch Solenoid Inductive Signature Malfunction");
            dtcDictionary.Add("P1728", "Transmission Slip Error");
            dtcDictionary.Add("P1729", "4x4 Low Switch Error");
            dtcDictionary.Add("P1730", "Gear Control Malfunction 2,3,5");
            dtcDictionary.Add("P1731", "1-2 Shift Malfunction");
            dtcDictionary.Add("P1732", "2-3 Shift Malfunction");
            dtcDictionary.Add("P1733", "3-4 Shift Malfunction");
            dtcDictionary.Add("P1734", "Gear Control Malfunction");
            dtcDictionary.Add("P1735", "First Gear Switch Circuit Malfunction");
            dtcDictionary.Add("P1736", "Second Gear Switch Circuit Malfunction");
            dtcDictionary.Add("P1737", "Lockup Solenoid System");
            dtcDictionary.Add("P1738", "Shift Time Error");
            dtcDictionary.Add("P1739", "Slip Solenoid System");
            dtcDictionary.Add("P1740", "Torque Converter Clutch Inductive Signature Malfunction");
            dtcDictionary.Add("P1741", "Torque Converter Clutch Control Error");
            dtcDictionary.Add("P1742", "Torque Converter Clutch Solenoid Failed On");
            dtcDictionary.Add("P1743", "Torque Converter Clutch Solenoid Failied On");
            dtcDictionary.Add("P1744", "Torque Converter Clutch System Performance");
            dtcDictionary.Add("P1745", "Line Pressure Solenoid System");
            dtcDictionary.Add("P1746", "Pressure Control Solenoid 'A' Open Circuit");
            dtcDictionary.Add("P1747", "Pressure Control Solenoid 'A' Short Circuit");
            dtcDictionary.Add("P1748", "EPC Malfunction");
            dtcDictionary.Add("P1749", "Pressure Control Solenoid Failed Low");
            dtcDictionary.Add("P1751", "Shift Solenoid A Performance");
            dtcDictionary.Add("P1754", "Coast Clutch Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1755", "Intermediate Speed Sensor (ISS) Malfunction");
            dtcDictionary.Add("P1756", "Shift Solenoid B Performance");
            dtcDictionary.Add("P1760", "Pressure Control Solenoid 'A' Short Circuit");
            dtcDictionary.Add("P1761", "Shift Solenoid C Performance");
            dtcDictionary.Add("P1762", "Overdrive Band Failed Off");
            dtcDictionary.Add("P1765", "Timing Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1767", "Torque Converter Clutch Circuit Malfunction");
            dtcDictionary.Add("P1768", "Performance / Normal / Winter Mode Input Malfunction");
            dtcDictionary.Add("P1769", "AG4 Transmission Torque Modulation Fault");
            dtcDictionary.Add("P1770", "Clutch Solenoid Circuit Malfunction");
            dtcDictionary.Add("P1775", "Transmission System MIL Fault");
            dtcDictionary.Add("P1776", "Ignition Retard Request Duration Fault");
            dtcDictionary.Add("P1777", "Ignition Retard Request Circuit Fault");
            dtcDictionary.Add("P1778", "Transmission Reverse I/P Circuit Malfunction");
            dtcDictionary.Add("P1779", "TCIL Circuit Malfunction");
            dtcDictionary.Add("P1780", "Trans Control Switch (O/D Cancel) Out of Self Test Range");
            dtcDictionary.Add("P1781", "4X4 Switch Out of Self Test Range");
            dtcDictionary.Add("P1782", "P/ES Circuit Out Of Self Test Range");
            dtcDictionary.Add("P1783", "Transmission Overtemperature Condition");
            dtcDictionary.Add("P1784", "Transmission Mechanical Failure - First And Reverse");
            dtcDictionary.Add("P1785", "Transmission Mechanical Failure - First And Second");
            dtcDictionary.Add("P1786", "3-2 Downshift Error");
            dtcDictionary.Add("P1787", "2-1 Downshift Error");
            dtcDictionary.Add("P1788", "Pressure Control Solenoid 'B' Open Circuit");
            dtcDictionary.Add("P1789", "Pressure Control Solenoid 'B' Short Circuit");
            dtcDictionary.Add("P1790", "TP (Mechanical) Circuit Malfunction");
            dtcDictionary.Add("P1791", "TP (Electric) Circuit Malfunction");
            dtcDictionary.Add("P1792", "Barometer Pressure Circuit Malfunction");
            dtcDictionary.Add("P1793", "Intake Air Volume Circuit Malfunction");
            dtcDictionary.Add("P1794", "Battery Voltage Circuit Malfunction");
            dtcDictionary.Add("P1795", "Idle Switch Circuit Malfunction");
            dtcDictionary.Add("P1796", "Kick Down Switch Circuit Malfunction");
            dtcDictionary.Add("P1797", "Neutral Switch Circuit Malfunction");
            dtcDictionary.Add("P1798", "Coolant Temperature Circuit Malfunction");
            dtcDictionary.Add("P1799", "Hold Switch Circuit Malfunction");
            dtcDictionary.Add("P1800", "Transmission Clutch Interlock Safety Switch Circuit Failure");
            dtcDictionary.Add("P1801", "Transmission Clutch Interlock Safety Switch Open Circuit");
            dtcDictionary.Add("P1802", "Transmission Clutch Interlock Safety Switch Short Circuit To Battery");
            dtcDictionary.Add("P1803", "Transmission Clutch Interlock Safety Switch Short Circuit To Ground");
            dtcDictionary.Add("P1804", "Transmission 4-Wheel Drive High Indicator Circuit Failure");
            dtcDictionary.Add("P1805", "Transmission 4-Wheel Drive High Indicator Open Circuit");
            dtcDictionary.Add("P1806", "Transmission 4-Wheel Drive High Indicator Short Circuit To Battery");
            dtcDictionary.Add("P1807", "Transmission 4-Wheel Drive High Indicator Short Circuit To Ground");
            dtcDictionary.Add("P1808", "Transmission 4-Wheel Drive Low Indicator Circuit Failure");
            dtcDictionary.Add("P1809", "Transmission 4-Wheel Drive Low Indicator Open Circuit");
            dtcDictionary.Add(
                "P1810",
                "TFP Valve Position Switch Circuit/ Transmission 4-Wheel Drive Low Indicator Short Circuit To Battery");
            dtcDictionary.Add("P1811", "Transmission 4-Wheel Drive Low Indicator Short Circuit To Ground");
            dtcDictionary.Add("P1812", "Transmission 4-Wheel Drive Mode Select Circuit Failure");
            dtcDictionary.Add("P1813", "Transmission 4-Wheel Drive Mode Select Open Circuit");
            dtcDictionary.Add("P1814", "Transmission 4-Wheel Drive Mode Select Short Circuit To Battery");
            dtcDictionary.Add("P1815", "Transmission 4-Wheel Drive Mode Select Short Circuit To Ground");
            dtcDictionary.Add("P1816", "Transmission Neutral Safety Switch Circuit Failure");
            dtcDictionary.Add("P1817", "Transmission Neutral Safety Switch Open Circuit");
            dtcDictionary.Add("P1818", "Transmission Neutral Safety Switch Short Circuit To Battery");
            dtcDictionary.Add("P1819", "Transmission Neutral Safety Switch Short Circuit To Ground");
            dtcDictionary.Add("P1820", "Transmission Transfer Case Clockwise Shift Relay Coil Circuit Failure");
            dtcDictionary.Add("P1821", "Transmission Transfer Case Clockwise Shift Relay Coil Open Circuit");
            dtcDictionary.Add("P1822", "Transmission Transfer Case Clockwise Shift Relay Coil Short Circuit To Battery");
            dtcDictionary.Add("P1823", "Transmission Transfer Case Clockwise Shift Relay Coil Short Circuit To Ground");
            dtcDictionary.Add("P1824", "Transmission 4-Wheel Drive Clutch Relay Circuit Failure");
            dtcDictionary.Add("P1825", "Transmission 4-Wheel Drive Clutch Relay Open Circuit");
            dtcDictionary.Add("P1826", "Transmission 4-Wheel Drive Low Clutch Relay Circuit To Battery");
            dtcDictionary.Add("P1827", "Transmission 4-Wheel Drive Low Clutch Relay Circuit To Ground");
            dtcDictionary.Add("P1828", "Transmission Transfer Case Counter Clockwise Shift Relay Coil Circuit Failure");
            dtcDictionary.Add("P1829", "Transmission Transfer Case Counter Clockwise Shift Relay Coil Open Circuit");
            dtcDictionary.Add(
                "P1830",
                "Transmission Transfer Case Counter Clockwise Shift Relay Coil Short Circuit To Battery");
            dtcDictionary.Add(
                "P1831",
                "Transmission Transfer Case Counter Clockwise Shift Relay Coil Short Circuit To Ground");
            dtcDictionary.Add("P1832", "Transmission Transfer Case Differential Lock-Up Solenoid Circuit Failure");
            dtcDictionary.Add("P1833", "Transmission Transfer Case Differential Lock-Up Solenoid Open Circuit");
            dtcDictionary.Add(
                "P1834",
                "Transmission Transfer Case Differential Lock-Up Solenoid Short Circuit To Battery");
            dtcDictionary.Add(
                "P1835",
                "Transmission Transfer Case Differential Lock-Up Solenoid Short Circuit To Ground");
            dtcDictionary.Add("P1836", "Transmission Transfer Case Front Shaft Speed Sensor Circuit Failure");
            dtcDictionary.Add("P1837", "Transmission Transfer Case Rear Shaft Speed Sensor Circuit Failure");
            dtcDictionary.Add("P1838", "Transmission Transfer Case Shift Motor Circuit Failure");
            dtcDictionary.Add("P1839", "Transmission Transfer Case Shift Motor Open Circuit");
            dtcDictionary.Add("P1840", "Transmission Transfer Case Shift Motor Short Circuit To Battery");
            dtcDictionary.Add("P1841", "Transmission Transfer Case Shift Motor Short Circuit To Ground");
            dtcDictionary.Add(
                "P1842",
                "Transmission Transfer Case Differential Lock-Up Feedback Switch Circuit Failure");
            dtcDictionary.Add("P1843", "Transmission Transfer Case Differential Lock-Up Feedback Switch Open Circuit");
            dtcDictionary.Add(
                "P1844",
                "Transmission Transfer Case Differential Lock-Up Feedback Switch Short Circuit To Battery");
            dtcDictionary.Add(
                "P1845",
                "Transmission Transfer Case Differential Lock-Up Feedback Switch Short Circuit To Ground");
            dtcDictionary.Add("P1846", "Transmission Transfer Case Contact Plate 'A' Circuit Failure");
            dtcDictionary.Add("P1847", "Transmission Transfer Case Contact Plate 'A' Open Circuit");
            dtcDictionary.Add("P1848", "Transmission Transfer Case Contact Plate 'A' Short Circuit To Battery");
            dtcDictionary.Add("P1849", "Transmission Transfer Case Contact Plate 'A' Short Circuit To Ground");
            dtcDictionary.Add("P1850", "Transmission Transfer Case Contact Plate 'B' Circuit Failure");
            dtcDictionary.Add("P1851", "Transmission Transfer Case Contact Plate 'B' Open Circuit");
            dtcDictionary.Add("P1852", "Transmission Transfer Case Contact Plate 'B' Short Circuit To Battery");
            dtcDictionary.Add("P1853", "Transmission Transfer Case Contact Plate 'B' Short Circuit To Ground");
            dtcDictionary.Add("P1854", "Transmission Transfer Case Contact Plate 'C' Circuit Failure");
            dtcDictionary.Add("P1855", "Transmission Transfer Case Contact Plate 'C' Open Circuit");
            dtcDictionary.Add("P1856", "Transmission Transfer Case Contact Plate 'C' Short Circuit To Battery");
            dtcDictionary.Add("P1857", "Transmission Transfer Case Contact Plate 'C' Short Circuit To Ground");
            dtcDictionary.Add("P1858", "Transmission Transfer Case Contact Plate 'D' Circuit Failure");
            dtcDictionary.Add("P1859", "Transmission Transfer Case Contact Plate 'D' Open Circuit");
            dtcDictionary.Add(
                "P1860",
                "TCC PWM Solenoid Circuit Electrical/ Transmission Transfer Case Contact Plate 'D' Short Circuit To Battery");
            dtcDictionary.Add("P1861", "Transmission Transfer Case Contact Plate 'D' Short Circuit To Ground");
            dtcDictionary.Add("P1862", "Transmission Transfer Case Contact Plate Power Circuit Failure");
            dtcDictionary.Add("P1863", "Transmission Transfer Case Contact Plate Power Open Circuit");
            dtcDictionary.Add("P1864", "Transmission Transfer Case Contact Plate Power Short To Battery");
            dtcDictionary.Add("P1865", "Transmission Transfer Case Contact Plate Power Short To Ground");
            dtcDictionary.Add("P1866", "Transmission Transfer Case System Concern - Servicing Required");
            dtcDictionary.Add("P1867", "Transmission Transfer Case Contact Plate General Circuit Failure");
            dtcDictionary.Add("P1868", "Transmission Automatic 4-Wheel Drive Indicator (Lamp) Circuit Failure");
            dtcDictionary.Add("P1869", "Transmission Automatic 4-Wheel Drive Indicator (Lamp) Circuit Short To Battery");
            dtcDictionary.Add(
                "P1870",
                "Transmission Component Slipping/ Transmission Mechanical Transfer Case 4x4 Switch Circuit Failure");
            dtcDictionary.Add("P1871", "Transmission Mechanical Transfer Case 4x4 Switch Circuit Short To Battery");
            dtcDictionary.Add("P1872", "Transmission Mechanical 4-Wheel Drive Axle Lock Lamp Circuit Failure");
            dtcDictionary.Add("P1873", "Transmission Mechanical 4-Wheel Drive Axle Lock Lamp Circuit Short To Battery");
            dtcDictionary.Add("P1874", "Transmission Automatic Hall Effect Sensor Power Circuit Failure");
            dtcDictionary.Add(
                "P1875",
                "Transmission Automatic Hall Effect Sensor Power Circuit Short To Battery / 4WD Low Switch Circuit Electrical");
            dtcDictionary.Add("P1876", "Transmission Transfer Case 2-Wheel Drive Solenoid Circuit Failure");
            dtcDictionary.Add("P1877", "Transmission Transfer Case 2-Wheel Drive Solenoid Circuit Short To Battery");
            dtcDictionary.Add("P1878", "Transmission Transfer Case Disengaged Solenoid Circuit Failure");
            dtcDictionary.Add("P1879", "Transmission Transfer Case Disengaged Solenoid Open Circuit");
            dtcDictionary.Add("P1880", "Transmission Transfer Case Disengaged Solenoid Short to Battery");
            dtcDictionary.Add("P1881", "Engine Coolant Level Switch Circuit Failure, GEM");
            dtcDictionary.Add("P1882", "Engine Coolant Level Switch Circuit Short to Ground");
            dtcDictionary.Add("P1883", "Engine Coolant Level Switch Circuit Failure, GEM");
            dtcDictionary.Add("P1884", "Engine Coolant Level Lamp Circuit Short to Ground");
            dtcDictionary.Add("P1885", "Transmission Transfer Case Disengaged Solenoid Short to Ground");
            dtcDictionary.Add("P1886", "4X4 Initialization Failure");
            dtcDictionary.Add("P1890", "Transmission 4WD Mode Select Return Input Circuit Failure");
            dtcDictionary.Add("P1891", "Transmission Transfer Case Contact Plate Ground Return Open Circuit");
            dtcDictionary.Add("P1900", "OSS Circuit Intermittent Malfunction");
            dtcDictionary.Add("P1901", "TSS Circuit Intermittent Malfunction");
            dtcDictionary.Add("P1902", "Pressure Control Solenoid 'B' Intermittent Short");
            dtcDictionary.Add("P1903", "Pressure Control Solenoid 'C' Short Circuit");
            dtcDictionary.Add("P1904", "Pressure Control Solenoid 'C' Open Circuit");
            dtcDictionary.Add("P1905", "Pressure Control Solenoid 'C' Intermittent Short");
            dtcDictionary.Add("P1906", "Kickdown Pull Relay Open or Short Circuit to Ground");
            dtcDictionary.Add("P1907", "Kickdown Hold Relay Open or Short Circuit to Ground");
            dtcDictionary.Add("P1908", "Transmission Pressure Circuit Solenoid Open or Short to Ground");
            dtcDictionary.Add("P1909", "Trans Temp Sensor Circuit Open or Shorted to Pwr or Gnd");
            dtcDictionary.Add("P1910", "VFS A Pressure Output Failed Low");
            dtcDictionary.Add("P1911", "VFS B Pressure Output Failed Low");
            dtcDictionary.Add("P1912", "VFS C Pressure Output Failed Low");
            dtcDictionary.Add("P1913", "Pressure Switch A Circuit Malfunction");
            dtcDictionary.Add("P1914", "Manually Shifted Automatic (MSA) Sw Circuit Malf");
            dtcDictionary.Add("P1915", "Reverse Switch Circuit Malfunction");
            dtcDictionary.Add("P1916", "High Clutch Drum Speed Sensor Malfunction");
            dtcDictionary.Add("P1917", "High Clutch Drum Speed Sensor Intermittent");
            dtcDictionary.Add("P1918", "Transmission Range Display Circuit Malfunction");
            dtcDictionary.Add("C1091", "Speed Wheel Sensor All Coherency Failure");
            dtcDictionary.Add("C1095", "ABS Hydraulic Pump Motor Circuit Failure");
            dtcDictionary.Add("C1096", "ABS Hydraulic Pump Motor Circuit Open");
            dtcDictionary.Add("C1097", "ABS Hydraulic Pump Motor Circuit Short To Ground");
            dtcDictionary.Add("C1098", "ABS Hydraulic Pump Motor Circuit Short To Battery");
            dtcDictionary.Add("C1100", "ABS Pump Switch Circuit Failure");
            dtcDictionary.Add("C1101", "ABS Hydraulic Valve Circuit Failure");
            dtcDictionary.Add("C1102", "ABS Acceleration Switch Circuit Failure");
            dtcDictionary.Add("C1103", "ABS Hydraulic Brake Switch Circuit Failure");
            dtcDictionary.Add("C1104", "Traction Control Active Lamp - Circuit Short to Battery");
            dtcDictionary.Add("C1105", "Traction Control Disable Lamp - Circuit Short to Battery");
            dtcDictionary.Add("C1106", "Traction Control Disable Switch Circuit Short to Battery");
            dtcDictionary.Add("C1107", "ABS Function Enabled Input Circuit Failure");
            dtcDictionary.Add("C1109", "Speed Control Actuator Assembly Cable Release Failure");
            dtcDictionary.Add("C1110", "ABS Power Relay Coil Circuit Failure");
            dtcDictionary.Add("C1111", "ABS Power Relay Coil Open Circuit");
            dtcDictionary.Add("C1112", "ABS Power Relay Coil Short Circuit To Ground");
            dtcDictionary.Add("C1113", "ABS Power Relay Coil Short Circuit To Battery");
            dtcDictionary.Add("C1114", "ABS Power Relay Output Short Circuit To Ground");
            dtcDictionary.Add("C1115", "ABS Power Relay Output Short Circuit To Battery");
            dtcDictionary.Add("C1116", "Starter Motor Circuit Failure");
            dtcDictionary.Add("C1117", "RPM Input Circuit Failure");
            dtcDictionary.Add("C1123", "Vehicle Speed Sensor Input Short to Battery");
            dtcDictionary.Add("C1124", "Input shaft speed signal missing/faulted");
            dtcDictionary.Add("C1125", "Brake Fluid Level Sensor Input Circuit Failure");
            dtcDictionary.Add("C1126", "Cruise Control Command Switch Assembly Circuit Failure");
            dtcDictionary.Add("C1127", "Cruise Control Deactivator Brake Switch Circuit Failure");
            dtcDictionary.Add("C1132", "Clutch position ckt short to ground");
            dtcDictionary.Add("C1133", "Clutch position ckt short to battery");
            dtcDictionary.Add("C1134", "Gear Shift position short to ground");
            dtcDictionary.Add("C1135", "Gear Shift position short to battery");
            dtcDictionary.Add("C1136", "Gear Select position short to ground");
            dtcDictionary.Add("C1137", "ECU is Defective");
            dtcDictionary.Add("C1138", "Gear Select position short to Battery");
            dtcDictionary.Add("C1139", "Wheel Speed Sensor Center Tone Ring Missing Tooth Fault");
            dtcDictionary.Add("C1140", "Hydraulic Base Brake Failure");
            dtcDictionary.Add("C1141", "Wheel Speed Sensor LF Tone Ring Tooth Missing Fault");
            dtcDictionary.Add("C1142", "Wheel Speed Sensor RF Tone Ring Tooth Missing Fault");
            dtcDictionary.Add("C1143", "Wheel Speed Sensor LR Tone Ring Tooth Missing Fault");
            dtcDictionary.Add("C1144", "Wheel Speed Sensor RR Tone Ring Tooth Missing Fault");
            dtcDictionary.Add("C1145", "Speed Wheel Sensor RF Input Circuit Failure");
            dtcDictionary.Add("C1146", "Speed Wheel Sensor RF Circuit Open");
            dtcDictionary.Add("C1148", "Speed Wheel Sensor RF Coherency Fault");
            dtcDictionary.Add("C1149", "Hydraulic Fluid Pressure/ Flow Circuit Failure");
            dtcDictionary.Add("C1150", "Two Speed Rear Axle Input Switch Circuit Short to Vbatt");
            dtcDictionary.Add("C1155", "Speed Wheel Sensor LF Input Circuit Failure");
            dtcDictionary.Add("C1156", "Speed Wheel Sensor LF Circuit Open");
            dtcDictionary.Add("C1157", "Park Brake Actuator Assembly Switch Applied Circuit Failure");
            dtcDictionary.Add("C1158", "Speed Wheel Sensor LF Coherency Fault");
            dtcDictionary.Add("C1159", "Hydraulic Fluid Pressure/ Flow Circuit Short to Gnd");
            dtcDictionary.Add("C1161", "Air Pressure Low Circuit Short to Gnd");
            dtcDictionary.Add("C1162", "Park Brake Switch # 2 Released Circuit Failure");
            dtcDictionary.Add("C1163", "Park Brake Switch # 2 Applied Circuit Short to Gnd");
            dtcDictionary.Add("C1164", "Park Brake Actuator Assembly Switch Circuit Failure");
            dtcDictionary.Add("C1165", "Speed Wheel Sensor RR Input Circuit Failure");
            dtcDictionary.Add("C1166", "Speed Wheel Sensor RR Input Open Circuit");
            dtcDictionary.Add("C1167", "Park Brake Actuator Assembly Switch Released Circuit Short to Gnd");
            dtcDictionary.Add("C1168", "Speed Wheel Sensor RR Coherency Fault");
            dtcDictionary.Add("C1169", "ABS Fluid Dumping Exceeds Maximum Timing");
            dtcDictionary.Add("C1170", "PRNDL Switch Circuit Failure");
            dtcDictionary.Add("C1172", "Park Brake Switch # 1 Applied Circuit Failure");
            dtcDictionary.Add("C1173", "Park Brake Switch # 1 Released Circuit Short to Gnd");
            dtcDictionary.Add("C1174", "Park Brake Switch # 2 Applied Circuit Failure");
            dtcDictionary.Add("C1175", "Speed Wheel Sensor LR Input Circuit Failure");
            dtcDictionary.Add("C1176", "Speed Wheel Sensor LR Circuit Open");
            dtcDictionary.Add("C1177", "Park Brake Actuator Assembly Switch Released Circuit Failure");
            dtcDictionary.Add("C1178", "Speed Wheel Sensor LR Coherency Fault");
            dtcDictionary.Add("C1179", "Speed Control Actuator Assembly Cable Slack Failure");
            dtcDictionary.Add("C1180", "Park Brake Valve Solenoid #1 Sense Input Circuit Failure");
            dtcDictionary.Add("C1181", "Park Brake Valve Solenoid #1 Sense Input Circuit Short to Gound");
            dtcDictionary.Add("C1182", "Park Lamp Flash Relay Circuit Failure");
            dtcDictionary.Add("C1183", "Park Lamp Flash Relay Circuit Short to Battery");
            dtcDictionary.Add("C1184", "ABS System Is Not Operational");
            dtcDictionary.Add("C1185", "ABS Power Relay Output Circuit Failure");
            dtcDictionary.Add("C1186", "ABS Power Relay Output Open Circuit");
            dtcDictionary.Add("C1187", "Brake Fluid Level Sensor Input Open Circuit");
            dtcDictionary.Add("C1188", "Brake Fluid Level Sensor Input Short Circuit To Battery");
            dtcDictionary.Add("C1189", "Brake Fluid Level Sensor Input Short Circuit To Ground");
            dtcDictionary.Add("C1190", "Speed Wheel Sensor LF Input Short Circuit To Battery");
            dtcDictionary.Add("C1191", "Speed Wheel Sensor LF Input Short Circuit To Ground");
            dtcDictionary.Add("C1192", "Speed Wheel Sensor RF Input Short Circuit To Battery");
            dtcDictionary.Add("C1193", "Speed Wheel Sensor RF Input Short Circuit To Ground");
            dtcDictionary.Add("C1194", "ABS Outlet Valve Coil LF Circuit Failure");
            dtcDictionary.Add("C1195", "ABS Outlet Valve Coil LF Circuit Open");
            dtcDictionary.Add("C1196", "ABS Outlet Valve Coil LF Circuit Short To Battery");
            dtcDictionary.Add("C1197", "ABS Outlet Valve Coil LF Circuit Short To Ground");
            dtcDictionary.Add("C1198", "ABS Inlet Valve Coil LF Circuit Failure");
            dtcDictionary.Add("C1199", "ABS Inlet Valve Coil LF Circuit Open");
            dtcDictionary.Add("C1200", "ABS Inlet Valve Coil LF Circuit Short To Battery");
            dtcDictionary.Add("C1201", "ABS Inlet Valve Coil LF Circuit Short To Ground");
            dtcDictionary.Add("C1202", "ABS Outlet Valve Coil Rear Circuit Failure");
            dtcDictionary.Add("C1203", "ABS Outlet Valve Coil Rear Circuit Open");
            dtcDictionary.Add("C1204", "ABS Outlet Valve Coil Rear Circuit Short To Battery");
            dtcDictionary.Add("C1205", "ABS Outlet Valve Coil Rear Circuit Short To Ground");
            dtcDictionary.Add("C1206", "ABS Inlet Valve Coil Rear Circuit Failure");
            dtcDictionary.Add("C1207", "ABS Inlet Valve Coil Rear Circuit Open");
            dtcDictionary.Add("C1208", "ABS Inlet Valve Coil Rear Circuit Short To Battery");
            dtcDictionary.Add("C1209", "ABS Inlet Valve Coil Rear Circuit Short To Ground");
            dtcDictionary.Add("C1210", "ABS Outlet Valve Coil RF Circuit Failure");
            dtcDictionary.Add("C1211", "ABS Outlet Valve Coil RF Circuit Open");
            dtcDictionary.Add("C1212", "ABS Outlet Valve Coil RF Circuit Short To Battery");
            dtcDictionary.Add("C1213", "ABS Outlet Valve Coil RF Circuit Short To Ground");
            dtcDictionary.Add("C1214", "ABS Inlet Valve Coil RF Circuit Failure");
            dtcDictionary.Add("C1215", "ABS Inlet Valve Coil RF Circuit Open");
            dtcDictionary.Add("C1216", "ABS Inlet Valve Coil RF Circuit Short To Battery");
            dtcDictionary.Add("C1217", "ABS Inlet Valve Coil RF Circuit Short To Ground");
            dtcDictionary.Add("C1218", "Lamp ABS Warning Output Circuit Failure");
            dtcDictionary.Add("C1219", "Lamp ABS Warning Output Circuit Open");
            dtcDictionary.Add("C1220", "Lamp ABS Warning Output Circuit Short To Battery");
            dtcDictionary.Add("C1221", "Lamp ABS Warning Output Circuit Short To Ground");
            dtcDictionary.Add("C1222", "Speed Wheel Mismatch");
            dtcDictionary.Add("C1223", "Lamp Brake Warning Output Circuit Failure");
            dtcDictionary.Add("C1224", "Lamp Brake Warning Output Circuit Open");
            dtcDictionary.Add("C1225", "Lamp Brake Warning Output Circuit Short To Battery");
            dtcDictionary.Add("C1226", "Lamp Brake Warning Output Circuit Short To Ground");
            dtcDictionary.Add("C1227", "Speed Wheel Sensor LR Input Short Circuit To Battery");
            dtcDictionary.Add("C1228", "Speed Wheel Sensor LR Input Short Circuit To Ground");
            dtcDictionary.Add("C1229", "Speed Wheel Sensor Rear Center Coherency Fault");
            dtcDictionary.Add("C1230", "Speed Wheel Sensor Rear Center Input Circuit Failure");
            dtcDictionary.Add("C1231", "Speed Wheel Sensor Rear Center Circuit Open");
            dtcDictionary.Add("C1232", "Speed Wheel Sensor Rear Center Input Short Circuit To Battery");
            dtcDictionary.Add("C1233", "Speed Wheel LF Input Signal Missing");
            dtcDictionary.Add("C1234", "Speed Wheel RF Input Signal Missing");
            dtcDictionary.Add("C1235", "Speed Wheel RR Input Signal Missing");
            dtcDictionary.Add("C1236", "Speed Wheel LR Input Signal Missing");
            dtcDictionary.Add("C1237", "Speed Wheel Rear Input Signal Missing");
            dtcDictionary.Add("C1238", "ABS Hydraulic Pressure Differential Switch Input Circuit Failure");
            dtcDictionary.Add("C1239", "ABS Hydraulic Pressure Differential Switch Input Open Circuit");
            dtcDictionary.Add("C1240", "ABS Hydraulic Pressure Differential Switch Input Short Circuit To Battery");
            dtcDictionary.Add("C1241", "ABS Hydraulic Pressure Differential Switch Input Short Circuit To Ground");
            dtcDictionary.Add("C1242", "ABS Outlet Valve Coil LR Circuit Failure");
            dtcDictionary.Add("C1243", "ABS Outlet Valve Coil LR Circuit Open");
            dtcDictionary.Add("C1244", "ABS Outlet Valve Coil LR Circuit Short To Battery");
            dtcDictionary.Add("C1245", "ABS Outlet Valve Coil LR Circuit Short To Ground");
            dtcDictionary.Add("C1246", "ABS Outlet Valve Coil RR Circuit Failure");
            dtcDictionary.Add("C1247", "ABS Outlet Valve Coil RR Circuit Open");
            dtcDictionary.Add("C1248", "ABS Outlet Valve Coil RR Circuit Short To Battery");
            dtcDictionary.Add("C1249", "ABS Outlet Valve Coil RR Circuit Short To Ground");
            dtcDictionary.Add("C1250", "ABS Inlet Valve Coil LR Circuit Failure");
            dtcDictionary.Add("C1251", "ABS Inlet Valve Coil LR Circuit Open");
            dtcDictionary.Add("C1252", "ABS Inlet Valve Coil LR Circuit Short To Battery");
            dtcDictionary.Add("C1253", "ABS Inlet Valve Coil LR Circuit Short To Ground");
            dtcDictionary.Add("C1254", "ABS Inlet Valve Coil RR Circuit Failure");
            dtcDictionary.Add("C1255", "ABS Inlet Valve Coil RR Circuit Open");
            dtcDictionary.Add("C1256", "ABS Inlet Valve Coil RR Circuit Short To Battery");
            dtcDictionary.Add("C1257", "ABS Inlet Valve Coil RR Circuit Short To Ground");
            dtcDictionary.Add("C1258", "Speed Wheel LF Comparison Failure");
            dtcDictionary.Add("C1259", "Speed Wheel RF Comparison Failure");
            dtcDictionary.Add("C1260", "Speed Wheel RR Comparison Failure");
            dtcDictionary.Add("C1261", "Speed Wheel LR Comparison Failure");
            dtcDictionary.Add("C1262", "Lamp Warning Relay Circuit Failure");
            dtcDictionary.Add("C1263", "Lamp Warning Relay Circuit Open");
            dtcDictionary.Add("C1264", "Lamp Warning Relay Circuit Short To Battery");
            dtcDictionary.Add("C1265", "Lamp Warning Relay Circuit Short To Ground");
            dtcDictionary.Add("C1266", "ABS Valve Power Relay Circuit Failure");
            dtcDictionary.Add("C1267", "ABS Functions Temporarily Disabled");
            dtcDictionary.Add("C1268", "Motor Relay # 1 Circuit Failure");
            dtcDictionary.Add("C1269", "Motor Relay # 1 Circuit Short to Battery");
            dtcDictionary.Add("C1270", "Motor # 1 Input Circuit Failure");
            dtcDictionary.Add("C1271", "Motor # 1 Input Circuit Short to Gnd");
            dtcDictionary.Add("C1272", "Motor # 2 Input Circuit Failure");
            dtcDictionary.Add("C1273", "Motor # 2 Input Circuit Short to Vbat");
            dtcDictionary.Add("C1274", "Solenoid Relay # 1 Circuit Failure");
            dtcDictionary.Add("C1275", "Solenoid Relay # 1 Circuit Short to Vbatt");
            dtcDictionary.Add("C1276", "Park Brake Actuator Assembly Switch Applied Circuit Short to Gnd");
            dtcDictionary.Add("C1277", "STEERING Wheel Angle 1and 2 Circuit Failure");
            dtcDictionary.Add("C1278", "STEERING Wheel Angle 1and 2 Signal Faulted");
            dtcDictionary.Add("C1279", "Yaw Rate Sensor circuit Failure");
            dtcDictionary.Add("C1280", "Yaw Rate Sensor Signal Fault");
            dtcDictionary.Add("C1281", "Lateral Accelerometer circuit Failure");
            dtcDictionary.Add("C1282", "Lateral Accelerometer Signal Fault");
            dtcDictionary.Add("C1283", "Switch Test Signal Failure");
            dtcDictionary.Add("C1284", "Oil Pressure Switch Failure");
            dtcDictionary.Add("C1285", "Booster Solenoid circuit Failure");
            dtcDictionary.Add("C1286", "Booster Mechanical Failure");
            dtcDictionary.Add("C1287", "Booster Pedal Force switch circuit Failure");
            dtcDictionary.Add("C1288", "Pressure Transducer Main / Primary Input Circuit Failure");
            dtcDictionary.Add("C1289", "Pressure Transducer Redundant / Secondary Input Circuit Failure");
            dtcDictionary.Add("C1400", "Traction Control Valve RF Circuit Failure");
            dtcDictionary.Add("C1401", "Traction Control Valve RF Circuit Open");
            dtcDictionary.Add("C1402", "Traction Control Valve RF Circuit Short To Ground");
            dtcDictionary.Add("C1403", "Traction Control Valve RF Circuit Short To Battery");
            dtcDictionary.Add("C1404", "Traction Control Valve Rear Circuit Failure");
            dtcDictionary.Add("C1405", "Traction Control Valve Rear Circuit Open");
            dtcDictionary.Add("C1406", "Traction Control Valve Rear Circuit Short To Ground");
            dtcDictionary.Add("C1407", "Traction Control Valve Rear Circuit Short To Battery");
            dtcDictionary.Add("C1410", "Traction Control Valve LF Circuit Failure");
            dtcDictionary.Add("C1411", "Traction Control Valve LF Circuit Open");
            dtcDictionary.Add("C1412", "Traction Control Valve LF Circuit Short To Ground");
            dtcDictionary.Add("C1413", "Traction Control Valve LF Circuit Short To Battery");
            dtcDictionary.Add("C1414", "Incorrect Module Design Level");
            dtcDictionary.Add("C1415", "Incorrect Module Configuration");
            dtcDictionary.Add("C1416", "Damper RF Circuit Short To Battery");
            dtcDictionary.Add("C1417", "Damper RF Circuit Short to Ground");
            dtcDictionary.Add("C1418", "Damper RF Circuit Failure");
            dtcDictionary.Add("C1419", "Damper RF Circuit Open");
            dtcDictionary.Add("C1420", "Hydraulic Fluid Pressure/ Flow Circuit Short to Vbatt");
            dtcDictionary.Add("C1421", "Damper LF Circuit Short To Battery");
            dtcDictionary.Add("C1422", "Damper LF Circuit Short to Ground");
            dtcDictionary.Add("C1423", "Damper LF Circuit Failure");
            dtcDictionary.Add("C1424", "Damper LF Circuit Open");
            dtcDictionary.Add("C1425", "Damper RR Circuit Short to Ground");
            dtcDictionary.Add("C1426", "Damper RR Circuit Short To Battery");
            dtcDictionary.Add("C1427", "Damper RR Circuit Open");
            dtcDictionary.Add("C1428", "Damper RR Circuit Failure");
            dtcDictionary.Add("C1429", "Input-shaft-speed input circuit failure");
            dtcDictionary.Add("C1430", "Damper LR Circuit Open");
            dtcDictionary.Add("C1431", "Damper LR Circuit Short To Battery");
            dtcDictionary.Add("C1432", "Damper LR Circuit Short to Ground");
            dtcDictionary.Add("C1433", "Damper LR Circuit Failure");
            dtcDictionary.Add("C1435", "Accelerometer Rear Circuit Failure");
            dtcDictionary.Add("C1436", "Accelerometer Rear Circuit Signal Is Not Sensed");
            dtcDictionary.Add("C1437", "Accelerometer Rear Circuit Short To Ground");
            dtcDictionary.Add("C1438", "Accelerometer Rear Circuit Short To Battery");
            dtcDictionary.Add("C1439", "Vehicle Acceleration EEC-IV Circuit Failure");
            dtcDictionary.Add("C1440", "Pressure Transducer Main / Primary signal Faulted");
            dtcDictionary.Add("C1441", "Steering Phase A Circuit Signal Is Not Sensed");
            dtcDictionary.Add("C1442", "Steering Phase B Circuit Signal Is Not Sensed");
            dtcDictionary.Add("C1443", "Steering Phase A Circuit Short To Ground");
            dtcDictionary.Add("C1444", "Steering Phase B Circuit Short To Ground");
            dtcDictionary.Add("C1445", "Speed Vehicle Signal Circuit Failure");
            dtcDictionary.Add("C1446", "Brake Switch Circuit Failure");
            dtcDictionary.Add("C1447", "Traction Control Module Request Circuit Failure");
            dtcDictionary.Add("C1448", "Lamp Adaptive Damping Warning Circuit Failure");
            dtcDictionary.Add("C1449", "Traction Control Motor Coherency Fault");
            dtcDictionary.Add("C1450", "Traction Control Motor Circuit Failure");
            dtcDictionary.Add("C1451", "Traction Control Motor Circuit Open");
            dtcDictionary.Add("C1452", "Traction Control Motor Circuit Short to Battery");
            dtcDictionary.Add("C1453", "Traction Control Motor Circuit Short to Ground");
            dtcDictionary.Add("C1454", "Front Lateral Accelerometer Circuit Failure");
            dtcDictionary.Add("C1455", "Accelerometer Front Circuit Failure");
            dtcDictionary.Add("C1456", "Accelerometer Front Circuit Is Not Sensed");
            dtcDictionary.Add("C1457", "Accelerometer Front Circuit Short To Ground");
            dtcDictionary.Add("C1458", "Accelerometer Front Circuit Short To Battery");
            dtcDictionary.Add("C1459", "Adaptive Mode Switch Circuit Failure");
            dtcDictionary.Add("C1460", "Vehicle Accelerometer Power Circuit Failure");
            dtcDictionary.Add("C1461", "Vehicle Accelerometer Power Circuit Short to Battery");
            dtcDictionary.Add("C1462", "Left Front Vertical Accelerometer Circuit Failure");
            dtcDictionary.Add("C1463", "Right Front Vertical Accelerometer Circuit Failure");
            dtcDictionary.Add("C1464", "Pressure Transducer Redundant / Secondary Signal Faulted");
            dtcDictionary.Add("C1465", "Damper High Side Front Circuit Short To Battery");
            dtcDictionary.Add("C1466", "Damper Circuit Failure");
            dtcDictionary.Add("C1467", "Damper High Side Rear Circuit Short To Battery");
            dtcDictionary.Add("C1468", "Damper Low Side Front Circuit Failure");
            dtcDictionary.Add("C1469", "Damper Low Side Rear Circuit Failure");
            dtcDictionary.Add("C1495", "Traction Control Motor Potentiometer Circuit Failure");
            dtcDictionary.Add("C1496", "Traction Control Motor Potentiometer Circuit Open");
            dtcDictionary.Add("C1497", "Traction Control Motor Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("C1498", "Traction Control Motor Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("C1499", "Transfer Case Contact Plate 'A' Encoder Circuit Failure");
            dtcDictionary.Add("C1500", "Transfer Case Contact Plate 'B' Encoder Circuit Failure");
            dtcDictionary.Add("C1501", "Transfer Case Contact Plate 'C' Encoder Circuit Failure");
            dtcDictionary.Add("C1502", "Transfer Case Contact Plate 'D' Encoder Circuit Failure");
            dtcDictionary.Add("C1503", "Dynamic Stability Control Left Front Valve Malfunction");
            dtcDictionary.Add("C1504", "Dynamic Stability Control Right Front Valve Malfunction");
            dtcDictionary.Add("C1505", "Dynamic Stability Control Left Rear Valve Malfunction");
            dtcDictionary.Add("C1506", "Dynamic Stability Control Right Rear Valve Malfunction");
            dtcDictionary.Add("C1507", "Traction Control of Brake Exceeds Time-Out");
            dtcDictionary.Add("C1508", "Traction Control of Engine Exceeds Time-Out");
            dtcDictionary.Add("C1510", "Right Front Wheel Pressure Reduction Performance Problem");
            dtcDictionary.Add("C1511", "Left Front Wheel Pressure Reduction Performance Problem");
            dtcDictionary.Add("C1512", "Right Rear Wheel Pressure Reduction Performance Problem");
            dtcDictionary.Add("C1513", "Left rear Wheel Pressure Reduction Performance Problem");
            dtcDictionary.Add("C1699", "Left Rear Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1700", "Left Rear Sensor Circuit Failure");
            dtcDictionary.Add("C1701", "Left Rear Sensor Circuit Fault");
            dtcDictionary.Add("C1702", "RightRear Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1703", "Right Rear Sensor Circuit Failure");
            dtcDictionary.Add("C1704", "Right Rear Sensor Circuit Fault");
            dtcDictionary.Add("C1705", "Left Rear Center Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1706", "Left Rear Center Sensor Circuit Failure");
            dtcDictionary.Add("C1707", "Left Rear Center Sensor Circuit Fault");
            dtcDictionary.Add("C1708", "Right Rear Center Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1709", "Right Rear Center Sensor Circuit Failure");
            dtcDictionary.Add("C1710", "Right Rear Center Sensor Circuit Fault");
            dtcDictionary.Add("C1711", "Left Front Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1712", "Left Front Sensor Circuit Failure");
            dtcDictionary.Add("C1713", "Left Front Sensor Circuit Fault");
            dtcDictionary.Add("C1714", "Right Front Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1715", "Right Front Sensor Circuit Failure");
            dtcDictionary.Add("C1716", "Right Front Sensor Circuit Fault");
            dtcDictionary.Add("C1717", "Left Front Center Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1718", "Left Front Center Sensor Circuit Failure");
            dtcDictionary.Add("C1719", "Left Front Center Sensor Circuit Fault");
            dtcDictionary.Add("C1721", "Air Suspension Height Sensor Power Circuit Open");
            dtcDictionary.Add("C1722", "Air Suspension Height Sensor Power Circuit Short To Battery");
            dtcDictionary.Add("C1723", "Air Suspension Height Sensor Power Circuit Short To Ground");
            dtcDictionary.Add("C1724", "Air Suspension Height Sensor Power Circuit Failure");
            dtcDictionary.Add("C1725", "Air Suspension Front Pneumatic Failure");
            dtcDictionary.Add("C1726", "Air Suspension Rear Pneumatic Failure");
            dtcDictionary.Add("C1727", "Air Suspension Reservoir Pneumatic Failure");
            dtcDictionary.Add("C1728", "Transfer Case unable to transition between 2H and 4H");
            dtcDictionary.Add("C1729", "Transfer Case unable to transition between 4H and 4L");
            dtcDictionary.Add("C1730", "Reference Voltage Out of Range (+5 v)");
            dtcDictionary.Add("C1731", "Air Suspension LF Corner Up Timeout");
            dtcDictionary.Add("C1732", "Air Suspension LF Corner Down Timeout");
            dtcDictionary.Add("C1733", "Air Suspension RF Corner Up Timeout");
            dtcDictionary.Add("C1734", "Air Suspension RF Corner Down Timeout");
            dtcDictionary.Add("C1735", "Air Suspension LR Corner Up Timeout");
            dtcDictionary.Add("C1736", "Air Suspension LR Corner Down Timeout");
            dtcDictionary.Add("C1737", "Air Suspension RR Corner Up Timeout");
            dtcDictionary.Add("C1738", "Air Suspension RR Corner Down Timeout");
            dtcDictionary.Add("C1739", "Right Front Center Sensor Circuit Short to Vbat");
            dtcDictionary.Add("C1740", "Right Front Center Sensor Circuit Failure");
            dtcDictionary.Add("C1741", "Right Front Center Sensor Circuit Fault");
            dtcDictionary.Add("C1742", "Rear Sounder Circuit Failure");
            dtcDictionary.Add("C1743", "Rear Sounder Circuit Short to Vbatt");
            dtcDictionary.Add("C1744", "Front Sounder Circuit Failure");
            dtcDictionary.Add("C1745", "Front Sounder Circuit Short to Vbatt");
            dtcDictionary.Add("C1748", "Switch input Circuit Short to Gnd");
            dtcDictionary.Add("C1749", "Trailer Input Circuit Failure");
            dtcDictionary.Add("C1750", "Accelerator Position Sensor Out of Range");
            dtcDictionary.Add("C1751", "Vehicle Speed Sensor # 1 Output Circuit Short to Vbatt");
            dtcDictionary.Add("C1752", "Vehicle Speed Sensor # 1 Output Circuit Short to Gnd");
            dtcDictionary.Add("C1753", "Hydraulic Clutch Actuator Valve Signal Fault");
            dtcDictionary.Add("C1754", "Hydraulic Clutch Actuator Valve Circuit Failure");
            dtcDictionary.Add("C1755", "Power Limit Shutdown Fault");
            dtcDictionary.Add("C1756", "Air Suspension Front Height Sensor High (SE) Signal Circuit Failure");
            dtcDictionary.Add("C1757", "Air Suspension Front Height Sensor High (SE) Signal Circuit Open");
            dtcDictionary.Add("C1758", "Air Suspension Front Height Sensor High (SE) Signal Circuit Short To Battery");
            dtcDictionary.Add("C1759", "Air Suspension Front Height Sensor High (SE) Signal Circuit Short To Ground");
            dtcDictionary.Add("C1760", "Air Suspension Rear Height Sensor High (SE) Signal Circuit Failure");
            dtcDictionary.Add("C1761", "Air Suspension Rear Height Sensor High (SE) Signal Circuit Open");
            dtcDictionary.Add("C1762", "Air Suspension Rear Height Sensor High (SE) Signal Circuit Short To Battery");
            dtcDictionary.Add("C1763", "Air Suspension Rear Height Sensor High (SE) Signal Circuit Short To Ground");
            dtcDictionary.Add("C1765", "Air Suspension Rear Height Sensor Low Signal Circuit Failure");
            dtcDictionary.Add("C1766", "Air Suspension Rear Height Sensor Low Signal Circuit Open");
            dtcDictionary.Add("C1767", "Air Suspension Rear Height Sensor Low Signal Circuit Short To Battery");
            dtcDictionary.Add("C1768", "Air Suspension Rear Height Sensor Low Signal Circuit Short To Ground");
            dtcDictionary.Add("C1770", "Air Suspension Vent Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1771", "Air Suspension Vent Solenoid Output Circuit Open");
            dtcDictionary.Add("C1772", "Air Suspension Vent Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1773", "Air Suspension Vent Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1774", "Coolant Temp Out of Range");
            dtcDictionary.Add("C1775", "DC-DC Converter Failure");
            dtcDictionary.Add("C1776", "Heater System Failure");
            dtcDictionary.Add("C1777", "Vacuum Pressure Circuit Failure");
            dtcDictionary.Add("C1778", "Power Steering Failure");
            dtcDictionary.Add("C1779", "Blower Switch Failure");
            dtcDictionary.Add("C1780", "Temperature Select Failure");
            dtcDictionary.Add("C1781", "Engine Coolant Temperature Signal Missing/Fault");
            dtcDictionary.Add("C1790", "Air Suspension LR Air Spring/Shock Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1791", "Air Suspension LR Air Spring/Shock Solenoid Output Circuit Open");
            dtcDictionary.Add("C1792", "Air Suspension LR Air Spring/Shock Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1793", "Air Suspension LR Air Spring/Shock Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1795", "Air Suspension RR Air Spring/Shock Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1796", "Air Suspension RR Air Spring/Shock Solenoid Output Circuit Open");
            dtcDictionary.Add("C1797", "Air Suspension RR Air Spring/Shock Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1798", "Air Suspension RR Air Spring/Shock Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1800", "Air Suspension Reservoir Solenoid Circuit Failure");
            dtcDictionary.Add("C1805", "Mismatched PCM and/or ABS-TC Module");
            dtcDictionary.Add("C1813", "Air Suspension LR Vent Request Exceeded Max Timing");
            dtcDictionary.Add("C1814", "Air Suspension RR Vent Request Exceeded Max Timing");
            dtcDictionary.Add("C1818", "Air Suspension LR Air Compress Request Exceeded Max Timing");
            dtcDictionary.Add("C1819", "Air Suspension RR Air Compress Request Exceeded Max Timing");
            dtcDictionary.Add("C1820", "Air Suspension RF Air Compress Request Exceeded Max Timing");
            dtcDictionary.Add("C1830", "Air Suspension Compressor Relay Circuit Failure");
            dtcDictionary.Add("C1831", "Air Suspension Compressor Relay Circuit Open");
            dtcDictionary.Add("C1832", "Air Suspension Compressor Relay Circuit Short To Battery");
            dtcDictionary.Add("C1833", "Air Suspension Compressor Relay Circuit Short To Ground");
            dtcDictionary.Add("C1834", "Gauge Drive Current Fault");
            dtcDictionary.Add("C1835", "Current Sense Circuit Failure");
            dtcDictionary.Add("C1836", "Battery Temp out of Range");
            dtcDictionary.Add("C1837", "Battery Heater Circuit Failure");
            dtcDictionary.Add("C1838", "Charging System Fault");
            dtcDictionary.Add("C1839", "Leakage Fault");
            dtcDictionary.Add("C1840", "Air Suspension Disable Switch Circuit Failure");
            dtcDictionary.Add("C1841", "Air Suspension Disable Switch Circuit Open");
            dtcDictionary.Add("C1842", "Air Suspension Disable Switch Circuit Short To Battery");
            dtcDictionary.Add("C1843", "Air Suspension Disable Switch Circuit Short To Ground");
            dtcDictionary.Add("C1844", "Air Suspension Secondary Front Inflator Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1845", "Air Suspension Front Inflator Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1846", "Air Suspension Front Inflator Solenoid Output Circuit Open");
            dtcDictionary.Add("C1847", "Air Suspension Front Inflator Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1848", "Air Suspension Front Inflator Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1849", "Master Cylinder Pressure Out of Range");
            dtcDictionary.Add("C1850", "Air Suspension Warning Lamp Circuit Failure");
            dtcDictionary.Add("C1851", "Air Suspension Warning Lamp Circuit Open");
            dtcDictionary.Add("C1852", "Air Suspension Warning Lamp Circuit Short To Battery");
            dtcDictionary.Add("C1853", "Air Suspension Warning Lamp Circuit Short To Ground");
            dtcDictionary.Add("C1854", "Motor Temperature Out of Range");
            dtcDictionary.Add("C1855", "Acceleration Position Sensor Conflict");
            dtcDictionary.Add("C1856", "Traction Motor Encoder circuit Failure");
            dtcDictionary.Add("C1859", "PRNDL Input #2 Circuit Failure");
            dtcDictionary.Add("C1860", "PRNDL Input #3 Circuit Failure");
            dtcDictionary.Add("C1861", "PRNDL Input #4 Circuit Failure");
            dtcDictionary.Add("C1862", "Contactor Circuit Failure");
            dtcDictionary.Add("C1863", "External Charging Fault");
            dtcDictionary.Add("C1864", "Battery Module Fault");
            dtcDictionary.Add("C1865", "Air Suspension Rear Inflator Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1866", "Air Suspension Rear Inflator Solenoid Output Circuit Open");
            dtcDictionary.Add("C1867", "Air Suspension Rear Inflator Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1868", "Air Suspension Rear Inflator Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1869", "Air Suspension Gate Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1870", "Air Suspension Gate Solenoid Output Circuit Open");
            dtcDictionary.Add("C1871", "Air Suspension Gate Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1872", "Air Suspension Gate Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1873", "Air Suspension RF Air Spring Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1874", "Air Suspension RF Air Spring Solenoid Output Circuit Open");
            dtcDictionary.Add("C1875", "Air Suspension RF Air Spring Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1876", "Air Suspension RF Air Spring Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1877", "Air Suspension LF Air Spring Solenoid Output Circuit Failure");
            dtcDictionary.Add("C1878", "Air Suspension LF Air Spring Solenoid Output Circuit Open");
            dtcDictionary.Add("C1879", "Air Suspension LF Air Spring Solenoid Output Circuit Short To Battery");
            dtcDictionary.Add("C1880", "Air Suspension LF Air Spring Solenoid Output Circuit Short To Ground");
            dtcDictionary.Add("C1881", "Air Suspension RF Height Sensor Circuit Failure");
            dtcDictionary.Add("C1882", "Air Suspension RF Height Sensor Circuit Open");
            dtcDictionary.Add("C1883", "Air Suspension RF Height Sensor Circuit Short To Battery");
            dtcDictionary.Add("C1884", "Air Suspension RF Height Sensor Circuit Short To Ground");
            dtcDictionary.Add("C1885", "Air Suspension RR Height Sensor Circuit Failure");
            dtcDictionary.Add("C1886", "Air Suspension RR Height Sensor Circuit Open");
            dtcDictionary.Add("C1887", "Air Suspension RR Height Sensor Circuit Short To Battery");
            dtcDictionary.Add("C1888", "Air Suspension RR Height Sensor Circuit Short To Ground");
            dtcDictionary.Add("C1889", "Air Suspension LF Height Sensor Circuit Failure");
            dtcDictionary.Add("C1890", "Air Suspension LF Height Sensor Circuit Open");
            dtcDictionary.Add("C1891", "Air Suspension LF Height Sensor Circuit Short To Battery");
            dtcDictionary.Add("C1892", "Air Suspension LF Height Sensor Circuit Short To Ground");
            dtcDictionary.Add("C1893", "Air Suspension LR Height Sensor Circuit Failure");
            dtcDictionary.Add("C1894", "Air Suspension LR Height Sensor Circuit Open");
            dtcDictionary.Add("C1895", "Air Suspension LR Height Sensor Circuit Short To Battery");
            dtcDictionary.Add("C1896", "Air Suspension LR Height Sensor Circuit Short To Ground");
            dtcDictionary.Add("C1897", "Steering VAPS II Circuit Loop Failure");
            dtcDictionary.Add("C1898", "Steering VAPS II Circuit Loop Open");
            dtcDictionary.Add("C1899", "Steering VAPS II Circuit Loop Short To Battery");
            dtcDictionary.Add("C1900", "Steering VAPS II Circuit Loop Short To Ground");
            dtcDictionary.Add("C1901", "Ride Control RR Shock Actuator Circuit Failure");
            dtcDictionary.Add("C1902", "Ride Control RR Shock Actuator Circuit Open");
            dtcDictionary.Add("C1903", "Ride Control RR Shock Actuator Circuit Short To Battery");
            dtcDictionary.Add("C1904", "Ride Control RR Shock Actuator Circuit Short To Ground");
            dtcDictionary.Add("C1905", "Ride Control LR Shock Actuator Circuit Failure");
            dtcDictionary.Add("C1906", "Ride Control LR Shock Actuator Circuit Open");
            dtcDictionary.Add("C1907", "Ride Control LR Shock Actuator Circuit Short To Battery");
            dtcDictionary.Add("C1908", "Ride Control LR Shock Actuator Circuit Short To Ground");
            dtcDictionary.Add("C1909", "Ride Control RF Shock Actuator Circuit Failure");
            dtcDictionary.Add("C1910", "Ride Control RF Shock Actuator Circuit Open");
            dtcDictionary.Add("C1911", "Ride Control RF Shock Actuator Circuit Short To Battery");
            dtcDictionary.Add("C1912", "Ride Control RF Shock Actuator Circuit Short To Ground");
            dtcDictionary.Add("C1913", "Ride Control LF Shock Actuator Circuit Failure");
            dtcDictionary.Add("C1914", "Ride Control LF Shock Actuator Circuit Open");
            dtcDictionary.Add("C1915", "Ride Control LF Shock Actuator Circuit Short To Battery");
            dtcDictionary.Add("C1916", "Ride Control LF Shock Actuator Circuit Short To Ground");
            dtcDictionary.Add("C1917", "Steering EVO Out-of-Range Fault");
            dtcDictionary.Add("C1918", "Air Suspension Ride Height Select Switch Circuit Failure");
            dtcDictionary.Add("C1920", "Led #1 Circuit Failure");
            dtcDictionary.Add("C1921", "VAPS Solenoid Actuator Output Circuit Failure");
            dtcDictionary.Add("C1922", "VAPS Solenoid Actuator Output Circuit Open");
            dtcDictionary.Add("C1923", "VAPS Solenoid Actuator Output Circuit Short To Battery");
            dtcDictionary.Add("C1924", "VAPS Solenoid Actuator Output Circuit Short To Ground");
            dtcDictionary.Add("C1925", "VAPS Solenoid Actuator Return Circuit Failure");
            dtcDictionary.Add("C1926", "VAPS Solenoid Actuator Return Circuit Open");
            dtcDictionary.Add("C1927", "VAPS Solenoid Actuator Return Circuit Short To Battery");
            dtcDictionary.Add("C1928", "VAPS Solenoid Actuator Return Circuit Short To Ground");
            dtcDictionary.Add("C1929", "Air Suspension Front Compressor Relay Circuit Failure");
            dtcDictionary.Add("C1930", "Air Suspension Front Compressor Relay Circuit Open");
            dtcDictionary.Add("C1931", "Air Suspension Front Compressor Relay Circuit Short To Battery");
            dtcDictionary.Add("C1932", "Air Suspension Front Compressor Relay Circuit Short To Ground");
            dtcDictionary.Add("C1933", "Solenoid Current Out Of Range");
            dtcDictionary.Add("C1934", "HPU (Hydraulic Pump Unit) Pressurisation Failure");
            dtcDictionary.Add("C1935", "Chime Circuit Failure");
            dtcDictionary.Add("C1936", "Hydraulic Pump Relay Circuit Failure");
            dtcDictionary.Add("C1937", "Steering Wheel Angle Sensor Offset Failure");
            dtcDictionary.Add("C1938", "Invalid Steering Wheel Angle Sensor ID");
            dtcDictionary.Add("C1939", "Brake Pressure Switch Input Circuit Failure");
            dtcDictionary.Add("C1940", "Brake Pressure Switch Mechanical Failure");
            dtcDictionary.Add("C1942", "Unrecognized Powertrain Configuration");
            dtcDictionary.Add("C1943", "Airbag Deployment Indication Input Fault");
            dtcDictionary.Add("C1944", "Gauge Driver Circuit Fault");
            dtcDictionary.Add("C1945", "Park Switch Indicates Park with Vehicle Moving");
            dtcDictionary.Add("C1946", "Seat Track Position Switch Circuit Open");
            dtcDictionary.Add("C1947", "Seat Track Position Switch Circuit Short to Ground");
            dtcDictionary.Add("C1948", "Seat Track Position Switch Circuit Resistance Out of Range");
            dtcDictionary.Add("C1949", "Accelerometer Sensor Circuit Open");
            dtcDictionary.Add("C1950", "Accelerometer Sensor Circuit Failure");
            dtcDictionary.Add("C1951", "Lateral Accelerometer Sensor Circuit Open");
            dtcDictionary.Add("C1952", "Yaw Rate Sensor Circuit Open");
            dtcDictionary.Add("C1953", "Master Cylinder Pressure Sensor Circuit Open");
            dtcDictionary.Add("C1954", "Master Cylinder Pressure Sensor Circuit Failure");
            dtcDictionary.Add("C1955", "Steering Angle Sensor Circuit Open");
            dtcDictionary.Add("C1956", "Steering Angle Sensor Circuit Failure");
            dtcDictionary.Add("C1957", "Dynamic Stability Control Valve RF Circuit Failure");
            dtcDictionary.Add("C1958", "Dynamic Stability Control Valve LF Circuit Failure");
            dtcDictionary.Add("C1959", "Lateral Accelerometer Sensor Circuit Failure");
            dtcDictionary.Add("C1960", "Driver Brake Apply Circuit Fault");
            dtcDictionary.Add("C1961", "Park Lamp Relay Coil Circuit Failure");
            dtcDictionary.Add("C1962", "Park Lamp Relay Coil Short to Battery");
            dtcDictionary.Add("C1963", "Stability Control Inhibit Warning");
            dtcDictionary.Add("B1200", "Climate Control Pushbutton Circuit Failure");
            dtcDictionary.Add("B1201", "Fuel Sender Circuit Failure");
            dtcDictionary.Add("B1202", "Fuel Sender Circuit Open");
            dtcDictionary.Add("B1203", "Fuel Sender Circuit Short To Battery");
            dtcDictionary.Add("B1204", "Fuel Sender Circuit Short To Ground");
            dtcDictionary.Add("B1205", "EIC Switch-1 Assembly Circuit Failure");
            dtcDictionary.Add("B1206", "EIC Switch-1 Assembly Circuit Open");
            dtcDictionary.Add("B1207", "EIC Switch-1 Assembly Circuit Short To Battery");
            dtcDictionary.Add("B1208", "EIC Switch-1 Assembly Circuit Short To Ground");
            dtcDictionary.Add("B1209", "EIC Switch-2 Assembly Circuit Failure");
            dtcDictionary.Add("B1210", "EIC Switch-2 Assembly Circuit Open");
            dtcDictionary.Add("B1211", "EIC Switch-2 Assembly Circuit Short To Battery");
            dtcDictionary.Add("B1212", "EIC Switch-2 Assembly Circuit Short To Ground");
            dtcDictionary.Add("B1213", "Anti-Theft Number of Programmed Keys Is Below Minimum");
            dtcDictionary.Add("B1214", "Running Board Lamp Circuit Failure");
            dtcDictionary.Add("B1215", "Running Board Lamp Circuit Short to Battery");
            dtcDictionary.Add("B1216", "Emergency & Road Side Assistance Switch Circuit Short to Ground");
            dtcDictionary.Add("B1217", "Horn Relay Coil Circuit Failure");
            dtcDictionary.Add("B1218", "Horn Relay Coil Circuit Short to Vbatt");
            dtcDictionary.Add("B1219", "Fuel Tank Pressure Sensor Circuit Failure");
            dtcDictionary.Add("B1220", "Fuel Tank Pressure Sensor Circuit Open");
            dtcDictionary.Add("B1222", "Fuel Temperature Sensor #1 Circuit Failure");
            dtcDictionary.Add("B1223", "Fuel Temperature Sensor #1 Circuit Open");
            dtcDictionary.Add("B1224", "Fuel Temperature Sensor #1 Circuit Short to Battery");
            dtcDictionary.Add("B1225", "Fuel Temperature Sensor #1 Circuit Short to Ground");
            dtcDictionary.Add("B1226", "Fuel Temperature Sensor #2 Circuit Failure");
            dtcDictionary.Add("B1227", "Fuel Temperature Sensor #2 Circuit Open");
            dtcDictionary.Add("B1228", "Fuel Temperature Sensor #2 Circuit Short to Battery");
            dtcDictionary.Add("B1229", "Fuel Temperature Sensor #2 Circuit Short to Ground");
            dtcDictionary.Add("B1231", "Longitudinal Acceleration Threshold Exceeded");
            dtcDictionary.Add("B1232", "See Manufacturer");
            dtcDictionary.Add("B1233", "Glass Break Sensor Failure");
            dtcDictionary.Add("B1234", "Mirror Switch Invalid Code");
            dtcDictionary.Add("B1235", "Window Feedback Failure");
            dtcDictionary.Add("B1236", "Window Feedback Loss of Signal");
            dtcDictionary.Add("B1237", "Window Feedback Out of Range");
            dtcDictionary.Add("B1238", "Over Temperature Fault");
            dtcDictionary.Add("B1239", "Air Flow Blend Door Driver Circuit Failure");
            dtcDictionary.Add("B1240", "Wiper Washer Rear Pump Relay Circuit Failure");
            dtcDictionary.Add("B1241", "Wiper Washer Rear Pump Relay Circuit Short to Battery");
            dtcDictionary.Add("B1242", "Air Flow Recirculation Door Driver Circuit Failure");
            dtcDictionary.Add("B1243", "Express Window Down Switch Circuit Short to Battery");
            dtcDictionary.Add("B1244", "Wiper Rear Motor Run Relay Circuit Failure");
            dtcDictionary.Add("B1245", "Wiper Rear Motor Run Relay Circuit Short to Battery");
            dtcDictionary.Add("B1246", "Dim Panel Potentiometer Switch Circuit Failure");
            dtcDictionary.Add("B1247", "Panel Dim Switch Circuit Open");
            dtcDictionary.Add("B1248", "Passenger's Seatback Autoglide Rearward Switch Circuit Short to Ground");
            dtcDictionary.Add("B1249", "Blend Door Failure");
            dtcDictionary.Add("B1250", "Air Temperature Internal Sensor Circuit Failure");
            dtcDictionary.Add("B1251", "Air Temperature Internal Sensor Circuit Open");
            dtcDictionary.Add("B1252", "Air Temperature Internal Sensor Circuit Short To Battery");
            dtcDictionary.Add("B1253", "Air Temperature Internal Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1254", "Air Temperature External Sensor Circuit Failure");
            dtcDictionary.Add("B1255", "Air Temperature External Sensor Circuit Open");
            dtcDictionary.Add("B1256", "Air Temperature External Sensor Circuit Short To Battery");
            dtcDictionary.Add("B1257", "Air Temperature External Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1258", "Solar Radiation Sensor Circuit Failure");
            dtcDictionary.Add("B1259", "Solar Radiation Sensor Circuit Open");
            dtcDictionary.Add("B1260", "Solar Radiation Sensor Circuit Short To Battery");
            dtcDictionary.Add("B1261", "Solar Radiation Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1262", "Servo Motor Defrost Circuit Failure");
            dtcDictionary.Add("B1263", "Servo Motor Vent Circuit Failure");
            dtcDictionary.Add("B1264", "Servo Motor Foot Circuit Failure");
            dtcDictionary.Add("B1265", "Servo Motor Coolair Bypass Circuit Failure");
            dtcDictionary.Add("B1266", "Servo Motor Airintake Left Circuit Failure");
            dtcDictionary.Add("B1267", "Servo Motor Airintake Right Circuit Failure");
            dtcDictionary.Add("B1268", "Servo Motor Potentiometer Defrost Circuit Failure");
            dtcDictionary.Add("B1269", "Servo Motor Potentiometer Defrost Circuit Open");
            dtcDictionary.Add("B1270", "Servo Motor Potentiometer Defrost Circuit Short To Battery");
            dtcDictionary.Add("B1271", "Servo Motor Potentiometer Defrost Circuit Short To Ground");
            dtcDictionary.Add("B1272", "Servo Motor Potentiometer Vent Circuit Failure");
            dtcDictionary.Add("B1273", "Servo Motor Potentiometer Vent Circuit Open");
            dtcDictionary.Add("B1274", "Servo Motor Potentiometer Vent Circuit Short To Battery");
            dtcDictionary.Add("B1275", "Servo Motor Potentiometer Vent Circuit Short To Ground");
            dtcDictionary.Add("B1276", "Servo Motor Potentiometer Foot Circuit Failure");
            dtcDictionary.Add("B1277", "Servo Motor Potentiometer Foot Circuit Open");
            dtcDictionary.Add("B1278", "Servo Motor Potentiometer Foot Circuit Short To Battery");
            dtcDictionary.Add("B1279", "Servo Motor Potentiometer Foot Circuit Short To Ground");
            dtcDictionary.Add("B1280", "Servo Motor Potentiometer Coolair Circuit Failure");
            dtcDictionary.Add("B1281", "Servo Motor Potentiometer Coolair Circuit Open");
            dtcDictionary.Add("B1282", "Servo Motor Potentiometer Coolair Circuit Short To Battery");
            dtcDictionary.Add("B1283", "Servo Motor Potentiometer Coolair Circuit Short To Ground");
            dtcDictionary.Add("B1284", "Servo Motor Potentiometer Airintake Left Circuit Failure");
            dtcDictionary.Add("B1285", "Servo Motor Potentiometer Airintake Left Circuit Open");
            dtcDictionary.Add("B1286", "Servo Motor Potentiometer Airintake Left Circuit Short To Battery");
            dtcDictionary.Add("B1287", "Servo Motor Potentiometer Airintake Left Circuit Short To Ground");
            dtcDictionary.Add("B1288", "Servo Motor Potentiometer Airintake Right Circuit Failure");
            dtcDictionary.Add("B1289", "Servo Motor Potentiometer Airintake Right Circuit Open");
            dtcDictionary.Add("B1290", "Servo Motor Potentiometer Airintake Right Circuit Short To Battery");
            dtcDictionary.Add("B1291", "Servo Motor Potentiometer Airintake Right Circuit Short To Ground");
            dtcDictionary.Add("B1292", "Battery Power Relay Circuit Failure");
            dtcDictionary.Add("B1293", "Battery Power Relay Circuit Open");
            dtcDictionary.Add("B1294", "Battery Power Relay Circuit Short To Battery");
            dtcDictionary.Add("B1295", "Battery Power Relay Circuit Short To Ground");
            dtcDictionary.Add("B1296", "Power Supply Sensor Circuit Failure");
            dtcDictionary.Add("B1297", "Power Supply Sensor Circuit Open");
            dtcDictionary.Add("B1298", "Power Supply Sensor Circuit Short To Battery");
            dtcDictionary.Add("B1299", "Power Supply Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1300", "Power Door Lock Circuit Failure");
            dtcDictionary.Add("B1301", "Power Door Lock Circuit Open");
            dtcDictionary.Add("B1302", "Accessory Delay Relay Coil Circuit Failure");
            dtcDictionary.Add("B1303", "Accessory Delay Relay Coil Circuit Open");
            dtcDictionary.Add("B1304", "Accessory Delay Relay Coil Circuit Short To Battery");
            dtcDictionary.Add("B1305", "Accessory Delay Relay Coil Circuit Short To Ground");
            dtcDictionary.Add("B1306", "Oil Level Switch Circuit Open");
            dtcDictionary.Add("B1307", "Oil Level Switch Circuit Short To Battery");
            dtcDictionary.Add("B1308", "Oil Level Switch Circuit Short To Ground");
            dtcDictionary.Add("B1309", "Power Door Lock Circuit Short To Ground");
            dtcDictionary.Add("B1310", "Power Door Unlock Circuit Failure");
            dtcDictionary.Add("B1311", "Power Door Unlock Circuit Open");
            dtcDictionary.Add("B1312", "Lamp Headlamp Input Circuit Short To Battery");
            dtcDictionary.Add("B1313", "Battery Saver Relay Coil Circuit Failure");
            dtcDictionary.Add("B1314", "Battery Saver Relay Coil Circuit Open");
            dtcDictionary.Add("B1315", "Battery Saver Relay Coil Circuit Short To Battery");
            dtcDictionary.Add("B1316", "Battery Saver Relay Coil Circuit Short To Ground");
            dtcDictionary.Add("B1317", "Battery Voltage High");
            dtcDictionary.Add("B1318", "Battery Voltage Low");
            dtcDictionary.Add("B1319", "Driver Door Ajar Circuit Failure");
            dtcDictionary.Add("B1320", "Driver Door Ajar Circuit Open");
            dtcDictionary.Add("B1321", "Driver Door Ajar Circuit Short To Battery");
            dtcDictionary.Add("B1322", "Driver Door Ajar Circuit Short To Ground");
            dtcDictionary.Add("B1323", "Door Ajar Lamp Circuit Failure");
            dtcDictionary.Add("B1324", "Door Ajar Lamp Circuit Open");
            dtcDictionary.Add("B1325", "Door Ajar Lamp Circuit Short To Battery");
            dtcDictionary.Add("B1326", "Door Ajar Lamp Circuit Short To Ground");
            dtcDictionary.Add("B1327", "Passenger Door Ajar Circuit Failure");
            dtcDictionary.Add("B1328", "Passenger Door Ajar Circuit Open");
            dtcDictionary.Add("B1329", "Passenger Door Ajar Circuit Short To Battery");
            dtcDictionary.Add("B1330", "Passenger Door Ajar Circuit Short To Ground");
            dtcDictionary.Add("B1331", "Decklid Ajar Rear Door Circuit Failure");
            dtcDictionary.Add("B1332", "Decklid Ajar Rear Door Circuit Open");
            dtcDictionary.Add("B1333", "Decklid Ajar Rear Door Circuit Short To Battery");
            dtcDictionary.Add("B1334", "Decklid Ajar Rear Door Circuit Short To Ground");
            dtcDictionary.Add("B1335", "Door Ajar RR Circuit Failure");
            dtcDictionary.Add("B1336", "Door Ajar RR Circuit Open");
            dtcDictionary.Add("B1337", "Door Ajar RR Circuit Short To Battery");
            dtcDictionary.Add("B1338", "Door Ajar RR Circuit Short To Ground");
            dtcDictionary.Add("B1339", "Chime Input Request Circuit Short To Battery");
            dtcDictionary.Add("B1340", "Chime Input Request Circuit Short To Ground");
            dtcDictionary.Add("B1341", "Power Door Unlock Circuit Short To Ground");
            dtcDictionary.Add("B1342", "ECU Is Defective");
            dtcDictionary.Add("B1343", "Heated Backlite Input Circuit Failure");
            dtcDictionary.Add("B1344", "Heated Backlite Input Circuit Open");
            dtcDictionary.Add("B1345", "Heated Backlite Input Circuit Short To Ground");
            dtcDictionary.Add("B1346", "Heated Backlite Input Circuit Short To Battery");
            dtcDictionary.Add("B1347", "Heated Backlite Relay Circuit Failure");
            dtcDictionary.Add("B1348", "Heated Backlite Relay Circuit Open");
            dtcDictionary.Add("B1349", "Heated Backlite Relay Short To Battery");
            dtcDictionary.Add("B1350", "Heated Backlite Relay Short To Ground");
            dtcDictionary.Add("B1351", "Ignition Key-In Circuit Short To Battery");
            dtcDictionary.Add("B1352", "Ignition Key-In Circuit Failure");
            dtcDictionary.Add("B1353", "Ignition Key-In Circuit Open");
            dtcDictionary.Add("B1354", "Ignition Key-In Circuit Short To Ground");
            dtcDictionary.Add("B1355", "Ignition Run Circuit Failure");
            dtcDictionary.Add("B1356", "Ignition Run Circuit Open");
            dtcDictionary.Add("B1357", "Ignition Run Circuit Short To Battery");
            dtcDictionary.Add("B1358", "Ignition Run Circuit Short To Ground");
            dtcDictionary.Add("B1359", "Ignition Run/Acc Circuit Failure");
            dtcDictionary.Add("B1360", "Ignition Run/Acc Circuit Open");
            dtcDictionary.Add("B1361", "Ignition Run/Acc Circuit Short To Battery");
            dtcDictionary.Add("B1362", "Ignition Run/Acc Circuit Short To Ground");
            dtcDictionary.Add("B1363", "Ignition Start Circuit Failure");
            dtcDictionary.Add("B1364", "Ignition Start Circuit Open");
            dtcDictionary.Add("B1365", "Ignition Start Circuit Short To Battery");
            dtcDictionary.Add("B1366", "Ignition Start Circuit Short To Ground");
            dtcDictionary.Add("B1367", "Ignition Tach Circuit Failure");
            dtcDictionary.Add("B1368", "Ignition Tach Circuit Open");
            dtcDictionary.Add("B1369", "Ignition Tach Circuit Short To Battery");
            dtcDictionary.Add("B1370", "Ignition Tach Circuit Short To Ground");
            dtcDictionary.Add("B1371", "Illuminated Entry Relay Circuit Failure");
            dtcDictionary.Add("B1372", "Illuminated Entry Relay Circuit Open");
            dtcDictionary.Add("B1373", "Illuminated Entry Relay Short To Battery");
            dtcDictionary.Add("B1374", "Illuminated Entry Relay Short To Ground");
            dtcDictionary.Add("B1375", "Oil Change Lamp Circuit Open");
            dtcDictionary.Add("B1376", "Oil Change Lamp Circuit Short To Battery");
            dtcDictionary.Add("B1377", "Oil Change Lamp Circuit Failure");
            dtcDictionary.Add("B1378", "Oil Change Lamp Circuit Short To Ground");
            dtcDictionary.Add("B1379", "Oil Change Reset Button Circuit Short To Ground");
            dtcDictionary.Add("B1380", "Oil Change Reset Button Circuit Failure");
            dtcDictionary.Add("B1381", "Oil Change Reset Button Circuit Open");
            dtcDictionary.Add("B1382", "Oil Change Reset Button Circuit Short To Battery");
            dtcDictionary.Add("B1383", "Oil Level Lamp Circuit Short To Battery");
            dtcDictionary.Add("B1384", "Oil Level Lamp Circuit Failure");
            dtcDictionary.Add("B1385", "Oil Level Lamp Circuit Open");
            dtcDictionary.Add("B1386", "Oil Level Lamp Circuit Short To Ground");
            dtcDictionary.Add("B1387", "Oil Temperature Sensor Circuit Open");
            dtcDictionary.Add("B1388", "Oil Temperature Sensor Circuit Short To Battery");
            dtcDictionary.Add("B1389", "Oil Temperature Sensor Circuit Failure");
            dtcDictionary.Add("B1390", "Oil Temperature Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1391", "Oil Level Switch Circuit Failure");
            dtcDictionary.Add("B1392", "Power Door Memory Lock Relay Circuit Failure");
            dtcDictionary.Add("B1393", "Power Door Memory Lock Relay Circuit Open");
            dtcDictionary.Add("B1394", "Power Door Memory Lock Relay Circuit Short To Battery");
            dtcDictionary.Add("B1395", "Power Door Memory Lock Relay Circuit Short To Ground");
            dtcDictionary.Add("B1396", "Power Door Lock Circuit Short To Battery");
            dtcDictionary.Add("B1397", "Power Door Unlock Circuit Short To Battery");
            dtcDictionary.Add("B1398", "Driver Power Window One Touch Window Relay Circuit Failure");
            dtcDictionary.Add("B1399", "Driver Power Window One Touch Window Relay Circuit Open");
            dtcDictionary.Add("B1400", "Driver Power Window One Touch Window Relay Circuit Short To Battery");
            dtcDictionary.Add("B1401", "Driver Power Window One Touch Window Relay Circuit Short To Ground");
            dtcDictionary.Add("B1402", "Driver Power Window Down Switch Circuit Failure");
            dtcDictionary.Add("B1403", "Driver Power Window Up Switch Circuit Failure");
            dtcDictionary.Add("B1404", "Driver Power Window Down Circuit Open");
            dtcDictionary.Add("B1405", "Driver Power Window Down Circuit Short To Battery");
            dtcDictionary.Add("B1406", "Driver Power Window Down Circuit Short To Ground");
            dtcDictionary.Add("B1407", "Driver Power Window Up Circuit Open");
            dtcDictionary.Add("B1408", "Driver Power Window Up Circuit Short To Battery");
            dtcDictionary.Add("B1409", "Driver Power Window Up Circuit Short To Ground");
            dtcDictionary.Add("B1410", "Driver Power Window Motor Circuit Failure");
            dtcDictionary.Add("B1411", "Driver Power Window Motor Circuit Open");
            dtcDictionary.Add("B1412", "Driver Power Window Motor Circuit Short To Battery");
            dtcDictionary.Add("B1413", "Driver Power Window Motor Circuit Short To Ground");
            dtcDictionary.Add("B1414", "Power Window LR Motor Circuit Failure");
            dtcDictionary.Add("B1415", "Power Window LR Motor Circuit Open");
            dtcDictionary.Add("B1416", "Power Window LR Motor Circuit Short To Battery");
            dtcDictionary.Add("B1417", "Power Window LR Motor Circuit Short To Ground");
            dtcDictionary.Add("B1418", "Passenger Power Window Motor Circuit Failure");
            dtcDictionary.Add("B1419", "Passenger Power Window Motor Circuit Open");
            dtcDictionary.Add("B1420", "Passenger Power Window Motor Circuit Short To Battery");
            dtcDictionary.Add("B1421", "Passenger Power Window Motor Circuit Short To Ground");
            dtcDictionary.Add("B1422", "Power Window RR Motor Circuit Failure");
            dtcDictionary.Add("B1423", "Power Window RR Motor Circuit Open");
            dtcDictionary.Add("B1424", "Power Window RR Motor Circuit Short To Battery");
            dtcDictionary.Add("B1425", "Power Window RR Motor Circuit Short To Ground");
            dtcDictionary.Add("B1426", "Lamp Seat Belt Circuit Short To Battery");
            dtcDictionary.Add("B1427", "Lamp Seat Belt Circuit Open");
            dtcDictionary.Add("B1428", "Lamp Seat Belt Circuit Failure");
            dtcDictionary.Add("B1429", "Lamp Seat Belt Circuit Short To Ground");
            dtcDictionary.Add("B1430", "Seat Belt Switch Circuit Short To Ground");
            dtcDictionary.Add("B1431", "Wiper Brake/Run Relay Circuit Failure");
            dtcDictionary.Add("B1432", "Wiper Brake/Run Relay Circuit Short To Battery");
            dtcDictionary.Add("B1433", "Wiper Brake/Run Relay Circuit Short To Ground");
            dtcDictionary.Add("B1434", "Wiper Hi/Low Speed Relay Coil Circuit Failure");
            dtcDictionary.Add("B1435", "Wiper Hi/Low Speed Relay Coil Circuit Open");
            dtcDictionary.Add("B1436", "Wiper Hi/Low Speed Relay Coil Circuit Short To Battery");
            dtcDictionary.Add("B1437", "Wiper Hi/Low Speed Relay Coil Circuit Short To Ground");
            dtcDictionary.Add("B1438", "Wiper Mode Select Switch Circuit Failure");
            dtcDictionary.Add("B1439", "Wiper Mode Select Switch Circuit Open");
            dtcDictionary.Add("B1440", "Wiper Mode Select Switch Circuit Short To Battery");
            dtcDictionary.Add("B1441", "Wiper Mode Select Switch Circuit Short To Ground");
            dtcDictionary.Add("B1442", "Door Handle Switch Circuit Failure");
            dtcDictionary.Add("B1443", "Door Handle Switch Circuit Open");
            dtcDictionary.Add("B1444", "Door Handle Switch Circuit Short To Battery");
            dtcDictionary.Add("B1445", "Door Handle Switch Circuit Short To Ground");
            dtcDictionary.Add("B1446", "Wiper Park Sense Circuit Failure");
            dtcDictionary.Add("B1447", "Wiper Park Sense Circuit Open");
            dtcDictionary.Add("B1448", "Wiper Park Sense Circuit Short To Battery");
            dtcDictionary.Add("B1449", "Wiper Park Sense Circuit Short To Ground");
            dtcDictionary.Add("B1450", "Wiper Wash/Delay Switch Circuit Failure");
            dtcDictionary.Add("B1451", "Wiper Wash/Delay Switch Circuit Open");
            dtcDictionary.Add("B1452", "Wiper Wash/Delay Switch Circuit Short To Battery");
            dtcDictionary.Add("B1453", "Wiper Wash/Delay Switch Circuit Short To Ground");
            dtcDictionary.Add("B1454", "Wiper Washer Fluid Lamp Circuit Failure");
            dtcDictionary.Add("B1455", "Wiper Washer Fluid Lamp Circuit Open");
            dtcDictionary.Add("B1456", "Wiper Washer Fluid Lamp Circuit Short To Battery");
            dtcDictionary.Add("B1457", "Wiper Washer Fluid Lamp Circuit Short To Ground");
            dtcDictionary.Add("B1458", "Wiper Washer Pump Motor Relay Circuit Failure");
            dtcDictionary.Add("B1459", "Wiper Washer Pump Motor Relay Coil Circuit Open");
            dtcDictionary.Add("B1460", "Wiper Washer Pump Motor Relay Coil Circuit Short To Battery");
            dtcDictionary.Add("B1461", "Wiper Washer Pump Motor Relay Coil Circuit Short To Ground");
            dtcDictionary.Add("B1462", "Seat Belt Switch Circuit Failure");
            dtcDictionary.Add("B1463", "Seat Belt Switch Circuit Open");
            dtcDictionary.Add("B1464", "Seat Belt Switch Circuit Short To Battery");
            dtcDictionary.Add("B1465", "Wiper Brake/Run Relay Circuit Open");
            dtcDictionary.Add("B1466", "Wiper Hi/Low Speed Not Switching");
            dtcDictionary.Add("B1467", "Wiper Hi/Low Speed Circuit Motor Short To Battery");
            dtcDictionary.Add("B1468", "Chime Input Request Circuit Failure");
            dtcDictionary.Add("B1469", "Chime Input Request Circuit Open");
            dtcDictionary.Add("B1470", "Lamp Headlamp Input Circuit Failure");
            dtcDictionary.Add("B1471", "Lamp Headlamp Input Circuit Open");
            dtcDictionary.Add("B1472", "Lamp Headlamp Input Circuit Short To Ground");
            dtcDictionary.Add("B1473", "Wiper Low Speed Circuit Motor Failure");
            dtcDictionary.Add("B1474", "Battery Saver Power Relay Circuit Short To Battery");
            dtcDictionary.Add("B1475", "Accessory Delay Relay Contact Short To Battery");
            dtcDictionary.Add("B1476", "Wiper High Speed Circuit Motor Failure");
            dtcDictionary.Add("B1477", "Wiper Hi/Low Circuit Motor Short To Ground");
            dtcDictionary.Add("B1478", "Power Window One Touch Up/Down Activated Simultaneously");
            dtcDictionary.Add("B1479", "Wiper Washer Fluid Level Sensor Circuit Failure");
            dtcDictionary.Add("B1480", "Wiper Washer Fluid Level Sensor Circuit Open");
            dtcDictionary.Add("B1481", "Wiper Washer Fluid Level Sensor Circuit Short To Battery");
            dtcDictionary.Add("B1482", "Wiper Washer Fluid Level Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1483", "Brake Pedal Input Circuit Failure");
            dtcDictionary.Add("B1484", "Brake Pedal Input Open Circuit");
            dtcDictionary.Add("B1485", "Brake Pedal Input Circuit Battery Short");
            dtcDictionary.Add("B1486", "Brake Pedal Input Circuit Ground Short");
            dtcDictionary.Add("B1487", "Door Handle Right Front Circuit Failure");
            dtcDictionary.Add("B1488", "Door Handle Right Front Circuit Open");
            dtcDictionary.Add("B1489", "Door Handle Right Front Short To Battery");
            dtcDictionary.Add("B1490", "Door Handle Right Front Short To Ground");
            dtcDictionary.Add("B1491", "Ignition Cylinder Sensor Circuit Failure");
            dtcDictionary.Add("B1492", "Ignition Cylinder Sensor Open Circuit");
            dtcDictionary.Add("B1493", "Ignition Cylinder Sensor Battery Short");
            dtcDictionary.Add("B1494", "Ignition Cylinder Sensor Ground Short");
            dtcDictionary.Add("B1495", "Decklid Punch-Out Sensor Circuit Failure");
            dtcDictionary.Add("B1496", "Decklid Punch-Out Sensor Open Circuit");
            dtcDictionary.Add("B1497", "Decklid Punch-Out Sensor Battery Short");
            dtcDictionary.Add("B1498", "Decklid Punch-Out Sensor Ground Short");
            dtcDictionary.Add("B1499", "Lamp Turn Signal Left Circuit Failure");
            dtcDictionary.Add("B1500", "Lamp Turn Signal Left Circuit Open");
            dtcDictionary.Add("B1501", "Lamp Turn Signal Left Circuit Short To Battery");
            dtcDictionary.Add("B1502", "Lamp Turn Signal Left Circuit Short To Ground");
            dtcDictionary.Add("B1503", "Lamp Turn Signal Right Circuit Failure");
            dtcDictionary.Add("B1504", "Lamp Turn Signal Right Circuit Open");
            dtcDictionary.Add("B1505", "Lamp Turn Signal Right Circuit Short To Battery");
            dtcDictionary.Add("B1506", "Lamp Turn Signal Right Circuit Short To Ground");
            dtcDictionary.Add("B1507", "Flash To Pass Switch Circuit Failure");
            dtcDictionary.Add("B1508", "Flash To Pass Switch Circuit Open");
            dtcDictionary.Add("B1509", "Flash To Pass Switch Circuit Short To Battery");
            dtcDictionary.Add("B1510", "Flash To Pass Switch Circuit Short To Ground");
            dtcDictionary.Add("B1511", "Driver Door Handle Circuit Failure");
            dtcDictionary.Add("B1512", "Driver Door Handle Circuit Open");
            dtcDictionary.Add("B1513", "Driver Door Handle Circuit Short To Battery");
            dtcDictionary.Add("B1514", "Driver Door Handle Circuit Short To Ground");
            dtcDictionary.Add("B1515", "Seat Driver Occupied Switch Circuit Failure");
            dtcDictionary.Add("B1516", "Seat Driver Occupied Switch Circuit Open");
            dtcDictionary.Add("B1517", "Seat Driver Occupied Switch Circuit Short To Battery");
            dtcDictionary.Add("B1518", "Seat Driver Occupied Switch Circuit Short To Ground");
            dtcDictionary.Add("B1519", "Hood Switch Circuit Failure");
            dtcDictionary.Add("B1520", "Hood Switch Circuit Open");
            dtcDictionary.Add("B1521", "Hood Switch Circuit Short To Battery");
            dtcDictionary.Add("B1522", "Hood Switch Circuit Short To Ground");
            dtcDictionary.Add("B1523", "Keyless Entry Circuit Failure");
            dtcDictionary.Add("B1524", "Keyless Entry Circuit Open");
            dtcDictionary.Add("B1525", "Keyless Entry Circuit Short To Battery");
            dtcDictionary.Add("B1526", "Keyless Entry Circuit Short To Ground");
            dtcDictionary.Add("B1527", "Memory Set Switch Circuit Failure");
            dtcDictionary.Add("B1528", "Memory Set Switch Circuit Open");
            dtcDictionary.Add("B1529", "Memory Set Switch Circuit Short To Battery");
            dtcDictionary.Add("B1530", "Memory Set Switch Circuit Short To Ground");
            dtcDictionary.Add("B1531", "Memory 1 Switch Circuit Failure");
            dtcDictionary.Add("B1532", "Memory 1 Switch Circuit Open");
            dtcDictionary.Add("B1533", "Memory 1 Switch Circuit Short To Battery");
            dtcDictionary.Add("B1534", "Memory 1 Switch Circuit Short To Ground");
            dtcDictionary.Add("B1535", "Memory 2 Switch Circuit Failure");
            dtcDictionary.Add("B1536", "Memory 2 Switch Circuit Open");
            dtcDictionary.Add("B1537", "Memory 2 Switch Circuit Short To Battery");
            dtcDictionary.Add("B1538", "Memory 2 Switch Circuit Short To Ground");
            dtcDictionary.Add("B1539", "Mirror Driver Switch Assembly Circuit Failure");
            dtcDictionary.Add("B1540", "Mirror Driver Switch Assembly Circuit Open");
            dtcDictionary.Add("B1541", "Mirror Driver Switch Assembly Circuit Short To Battery");
            dtcDictionary.Add("B1542", "Mirror Driver Switch Assembly Circuit Short To Ground");
            dtcDictionary.Add("B1543", "Seat Direction Switch Assembly Circuit Failure");
            dtcDictionary.Add("B1544", "Seat Direction Switch Assembly Circuit Open");
            dtcDictionary.Add("B1545", "Seat Direction Switch Assembly Circuit Short To Battery");
            dtcDictionary.Add("B1546", "Seat Direction Switch Assembly Circuit Short To Ground");
            dtcDictionary.Add("B1547", "Power Window Master Circuit Failure");
            dtcDictionary.Add("B1548", "Power Window Master Circuit Open");
            dtcDictionary.Add("B1549", "Power Window Master Circuit Short To Battery");
            dtcDictionary.Add("B1550", "Power Window Master Circuit Short To Ground");
            dtcDictionary.Add("B1551", "Decklid Release Circuit Failure");
            dtcDictionary.Add("B1552", "Decklid Release Circuit Open");
            dtcDictionary.Add("B1553", "Decklid Release Circuit Short To Battery");
            dtcDictionary.Add("B1554", "Decklid Release Circuit Short To Ground");
            dtcDictionary.Add("B1555", "Ignition Run/Start Circuit Failure");
            dtcDictionary.Add("B1556", "Ignition Run/Start Circuit Open");
            dtcDictionary.Add("B1557", "Ignition Run/Start Circuit Short To Battery");
            dtcDictionary.Add("B1558", "Ignition Run/Start Circuit Short To Ground");
            dtcDictionary.Add("B1559", "Door Lock Cylinder Circuit Failure");
            dtcDictionary.Add("B1560", "Door Lock Cylinder Circuit Open");
            dtcDictionary.Add("B1561", "Door Lock Cylinder Circuit Short To Battery");
            dtcDictionary.Add("B1562", "Door Lock Cylinder Circuit Short To Ground");
            dtcDictionary.Add("B1563", "Door Ajar Circuit Failure");
            dtcDictionary.Add("B1564", "Door Ajar Circuit Open");
            dtcDictionary.Add("B1565", "Door Ajar Circuit Short To Battery");
            dtcDictionary.Add("B1566", "Door Ajar Circuit Short To Ground");
            dtcDictionary.Add("B1567", "Lamp Headlamp High-Beam Circuit Failure");
            dtcDictionary.Add("B1568", "Lamp Headlamp High-Beam Circuit Open");
            dtcDictionary.Add("B1569", "Lamp Headlamp High-Beam Circuit Short To Battery");
            dtcDictionary.Add("B1570", "Lamp Headlamp High-Beam Circuit Short To Ground");
            dtcDictionary.Add("B1571", "Door Ajar LR Circuit Failure");
            dtcDictionary.Add("B1572", "Door Ajar LR Circuit Open");
            dtcDictionary.Add("B1573", "Door Ajar LR Circuit Short To Battery");
            dtcDictionary.Add("B1574", "Door Ajar LR Circuit Short To Ground");
            dtcDictionary.Add("B1575", "Lamp Park Input Circuit Failure");
            dtcDictionary.Add("B1576", "Lamp Park Input Circuit Open");
            dtcDictionary.Add("B1577", "Lamp Park Input Circuit Short To Battery");
            dtcDictionary.Add("B1578", "Lamp Park Input Circuit Short To Ground");
            dtcDictionary.Add("B1579", "Dim Panel Increase Input Circuit Failure");
            dtcDictionary.Add("B1580", "Dim Panel Increase Input Circuit Open");
            dtcDictionary.Add("B1581", "Dim Panel Increase Input Circuit Short To Battery");
            dtcDictionary.Add("B1582", "Dim Panel Increase Input Circuit Short To Ground");
            dtcDictionary.Add("B1583", "Dim Panel Decrease Input Circuit Failure");
            dtcDictionary.Add("B1584", "Dim Panel Decrease Input Circuit Open");
            dtcDictionary.Add("B1585", "Dim Panel Decrease Input Circuit Short To Battery");
            dtcDictionary.Add("B1586", "Dim Panel Decrease Input Circuit Short To Ground");
            dtcDictionary.Add("B1587", "Autolamp Delay Increase Circuit Failure");
            dtcDictionary.Add("B1588", "Autolamp Delay Increase Circuit Open");
            dtcDictionary.Add("B1589", "Autolamp Delay Increase Circuit Short To Battery");
            dtcDictionary.Add("B1590", "Autolamp Delay Increase Circuit Short To Ground");
            dtcDictionary.Add("B1591", "Autolamp Delay Decrease Circuit Failure");
            dtcDictionary.Add("B1592", "Autolamp Delay Decrease Circuit Open");
            dtcDictionary.Add("B1593", "Autolamp Delay Decrease Circuit Short To Battery");
            dtcDictionary.Add("B1594", "Autolamp Delay Decrease Circuit Short To Ground");
            dtcDictionary.Add("B1595", "Ignition Switch Illegal Input Code");
            dtcDictionary.Add("B1596", "Service Continuous Codes");
            dtcDictionary.Add("B1597", "Driver's Seat Seatback Autoglide Forward Switch Circuit Short to Ground");
            dtcDictionary.Add("B1598", "Driver's Seat Seatback Autoglide Rearward Switch Circuit Short to Ground");
            dtcDictionary.Add("B1599", "Passenger's Seatback Autoglide Forward Switch Circuit Short to Ground");
            dtcDictionary.Add("B1600", "PATS Ignition Key Transponder Signal Is Not Received");
            dtcDictionary.Add("B1601", "PATS Received Incorrect Key-Code From Ignition Key Transponder");
            dtcDictionary.Add("B1602", "PATS Received Invalid Format Of Key-Code From Ignition Key Transponder");
            dtcDictionary.Add("B1603", "Lamp Anti-Theft Indicator Circuit Failure");
            dtcDictionary.Add("B1604", "Lamp Anti-Theft Indicator Circuit Open");
            dtcDictionary.Add("B1605", "Lamp Anti-Theft Indicator Circuit Short To Battery");
            dtcDictionary.Add("B1606", "Lamp Anti-Theft Indicator Circuit Short To Ground");
            dtcDictionary.Add("B1607", "Illuminated Entry Input Circuit Failure");
            dtcDictionary.Add("B1608", "Illuminated Entry Input Open Circuit");
            dtcDictionary.Add("B1609", "Illuminated Entry Input Short Circuit To Battery");
            dtcDictionary.Add("B1610", "Illuminated Entry Input Short Circuit To Ground");
            dtcDictionary.Add("B1611", "Wiper Rear Mode Select Switch Circuit Failure");
            dtcDictionary.Add("B1612", "Wiper Rear Mode Select Switch Circuit Open");
            dtcDictionary.Add("B1613", "Wiper Rear Mode Select Switch Circuit Short To Battery");
            dtcDictionary.Add("B1614", "Wiper Rear Mode Select Switch Circuit Short To Ground");
            dtcDictionary.Add("B1615", "Wiper Rear Disable Switch Circuit Failure");
            dtcDictionary.Add("B1616", "Wiper Rear Disable Switch Circuit Open");
            dtcDictionary.Add("B1617", "Wiper Rear Disable Switch Circuit Short To Battery");
            dtcDictionary.Add("B1618", "Wiper Rear Disable Switch Circuit Short To Ground");
            dtcDictionary.Add("B1619", "Wiper Rear Low Limit Input Circuit Failure");
            dtcDictionary.Add("B1620", "Wiper Rear Low Limit Input Circuit Open");
            dtcDictionary.Add("B1621", "Wiper Rear Low Limit Input Circuit Short To Battery");
            dtcDictionary.Add("B1622", "Wiper Rear Low Limit Input Circuit Short To Ground");
            dtcDictionary.Add("B1623", "Lamp Keypad Output Circuit Failure");
            dtcDictionary.Add("B1624", "Lamp Keypad Output Open Circuit");
            dtcDictionary.Add("B1625", "Lamp Keypad Output Short Circuit To Battery");
            dtcDictionary.Add("B1626", "Lamp Keypad Output Short Circuit To Ground");
            dtcDictionary.Add("B1627", "PRNDL Reverse Input Circuit Failure");
            dtcDictionary.Add("B1628", "PRNDL Reverse Input Open Circuit");
            dtcDictionary.Add("B1629", "PRNDL Reverse Input Short To Battery");
            dtcDictionary.Add("B1630", "PRNDL Reverse Input Short Circuit To Ground");
            dtcDictionary.Add("B1631", "Mirror Driver Left Circuit Failure");
            dtcDictionary.Add("B1632", "Mirror Driver Left Circuit Open");
            dtcDictionary.Add("B1633", "Mirror Driver Left Circuit Short To Battery");
            dtcDictionary.Add("B1634", "Mirror Driver Left Circuit Short To Ground");
            dtcDictionary.Add("B1635", "Mirror Driver Right Circuit Failure");
            dtcDictionary.Add("B1636", "Mirror Driver Right Circuit Open");
            dtcDictionary.Add("B1637", "Mirror Driver Right Circuit Short To Battery");
            dtcDictionary.Add("B1638", "Mirror Driver Right Short To Ground");
            dtcDictionary.Add("B1639", "Mirror Passenger Left Circuit Failure");
            dtcDictionary.Add("B1640", "Mirror Passenger Left Circuit Open");
            dtcDictionary.Add("B1641", "Mirror Passenger Left Circuit Short To Battery");
            dtcDictionary.Add("B1642", "Mirror Passenger Left Circuit Short To Ground");
            dtcDictionary.Add("B1643", "Mirror Passenger Right Circuit Failure");
            dtcDictionary.Add("B1644", "Mirror Passenger Right Circuit Open");
            dtcDictionary.Add("B1645", "Mirror Passenger Right Circuit Short To Battery");
            dtcDictionary.Add("B1646", "Mirror Passenger Right Circuit Short To Ground");
            dtcDictionary.Add("B1647", "Seat Driver Recline Forward Circuit Failure");
            dtcDictionary.Add("B1648", "Seat Driver Recline Forward Circuit Open");
            dtcDictionary.Add("B1649", "Seat Driver Recline Forward Circuit Short To Battery");
            dtcDictionary.Add("B1650", "Seat Driver Recline Forward Circuit Short To Ground");
            dtcDictionary.Add("B1651", "Seat Driver Recline Backward Circuit Failure");
            dtcDictionary.Add("B1652", "Seat Driver Recline Backward Circuit Open");
            dtcDictionary.Add("B1653", "Seat Driver Recline Backward Circuit Short To Battery");
            dtcDictionary.Add("B1654", "Seat Driver Recline Backward Circuit Short To Ground");
            dtcDictionary.Add("B1655", "Seat Driver Rear Up Circuit Failure");
            dtcDictionary.Add("B1656", "Seat Driver Rear Up Circuit Open");
            dtcDictionary.Add("B1657", "Seat Driver Rear Up Circuit Short To Battery");
            dtcDictionary.Add("B1658", "Seat Driver Rear Up Circuit Short To Ground");
            dtcDictionary.Add("B1659", "Seat Driver Front Up Circuit Failure");
            dtcDictionary.Add("B1660", "Seat Driver Front Up Circuit Open");
            dtcDictionary.Add("B1661", "Seat Driver Front Up Circuit Short To Battery");
            dtcDictionary.Add("B1662", "Seat Driver Front Up Circuit Short To Ground");
            dtcDictionary.Add("B1663", "Seat Driver Front Up/Down Motor Stalled");
            dtcDictionary.Add("B1664", "Seat Driver Rear Up/Down Motor Stalled");
            dtcDictionary.Add("B1665", "Seat Driver Forward/Backward Motor Stalled");
            dtcDictionary.Add("B1666", "Seat Driver Recline Motor Stalled");
            dtcDictionary.Add("B1667", "Mirror Driver Up/Down Motor Stalled");
            dtcDictionary.Add("B1668", "Mirror Driver Right/Left Motor Stalled");
            dtcDictionary.Add("B1669", "Mirror Passenger Up/Down Motor Stalled");
            dtcDictionary.Add("B1670", "Mirror Passenger Right/Left Motor Stalled");
            dtcDictionary.Add("B1671", "Battery Module Voltage Out Of Range");
            dtcDictionary.Add("B1672", "Seat Driver Occupied Input Circuit Failure");
            dtcDictionary.Add("B1673", "Seat Driver Occupied Input Circuit Open");
            dtcDictionary.Add("B1674", "Seat Driver Occupied Input Circuit Short To Battery");
            dtcDictionary.Add("B1675", "Seat Driver Occupied Input Circuit Short To Ground");
            dtcDictionary.Add("B1676", "Battery Pack Voltage Out Of Range");
            dtcDictionary.Add("B1677", "Alarm Panic Input Circuit Failure");
            dtcDictionary.Add("B1678", "Alarm Panic Input Circuit Open");
            dtcDictionary.Add("B1679", "Alarm Panic Input Circuit Short To Battery");
            dtcDictionary.Add("B1680", "Alarm Panic Input Circuit Short To Ground");
            dtcDictionary.Add("B1681", "PATS Transceiver Module Signal Is Not Received");
            dtcDictionary.Add("B1682", "PATS Is Disabled (Check Link Between PATS And Transponder)");
            dtcDictionary.Add("B1683", "Mirror Driver/Passenger Switch Circuit Failure");
            dtcDictionary.Add("B1684", "Mirror Driver/Passenger Switch Circuit Open");
            dtcDictionary.Add("B1685", "Lamp Dome Input Circuit Failure");
            dtcDictionary.Add("B1686", "Lamp Dome Input Circuit Open");
            dtcDictionary.Add("B1687", "Lamp Dome Input Circuit Short To Battery");
            dtcDictionary.Add("B1688", "Lamp Dome Input Circuit Short To Ground");
            dtcDictionary.Add("B1689", "Autolamp Delay Circuit Failure");
            dtcDictionary.Add("B1690", "Autolamp Delay Circuit Open");
            dtcDictionary.Add("B1691", "Autolamp Delay Circuit Short To Battery");
            dtcDictionary.Add("B1692", "Autolamp Delay Circuit Short To Ground");
            dtcDictionary.Add("B1693", "Autolamp On Circuit Failure");
            dtcDictionary.Add("B1694", "Autolamp On Circuit Open");
            dtcDictionary.Add("B1695", "Autolamp On Circuit Short To Battery");
            dtcDictionary.Add("B1696", "Autolamp On Circuit Short To Ground");
            dtcDictionary.Add("B1697", "Mirror Driver/Passenger Switch Circuit Short To Battery");
            dtcDictionary.Add("B1698", "Mirror Driver/Passenger Switch Circuit Short To Ground");
            dtcDictionary.Add("B1699", "Passenger's Seat Occupied Switch Circuit Short to Ground");
            dtcDictionary.Add("B1700", "Passenger's Seatbelt Tension Reducer Circuit Short to Ground");
            dtcDictionary.Add("B1701", "Seat Driver Recline Forward Switch Circuit Failure");
            dtcDictionary.Add("B1702", "Seat Driver Recline Forward Switch Circuit Open");
            dtcDictionary.Add("B1703", "Seat Driver Recline Forward Switch Circuit Short To Battery");
            dtcDictionary.Add("B1704", "Seat Driver Recline Forward Switch Circuit Short To Ground");
            dtcDictionary.Add("B1705", "Seat Driver Recline Rearward Switch Circuit Failure");
            dtcDictionary.Add("B1706", "Seat Driver Recline Rearward Switch Circuit Open");
            dtcDictionary.Add("B1707", "Seat Driver Recline Rearward Switch Circuit Short To Battery");
            dtcDictionary.Add("B1708", "Seat Driver Recline Rearward Switch Circuit Short To Ground");
            dtcDictionary.Add("B1709", "Seat Driver Front Up Switch Circuit Failure");
            dtcDictionary.Add("B1710", "Seat Driver Front Up Switch Circuit Open");
            dtcDictionary.Add("B1711", "Seat Driver Front Up Switch Circuit Short To Battery");
            dtcDictionary.Add("B1712", "Seat Driver Front Up Switch Circuit Short To Ground");
            dtcDictionary.Add("B1713", "Seat Driver Front Down Switch Circuit Failure");
            dtcDictionary.Add("B1714", "Seat Driver Front Down Switch Circuit Open");
            dtcDictionary.Add("B1715", "Seat Driver Front Down Switch Circuit Short To Battery");
            dtcDictionary.Add("B1716", "Seat Driver Front Down Switch Circuit Short To Ground");
            dtcDictionary.Add("B1717", "Seat Driver Forward Switch Circuit Failure");
            dtcDictionary.Add("B1718", "Seat Driver Forward Switch Circuit Open");
            dtcDictionary.Add("B1719", "Seat Driver Forward Switch Circuit Short To Battery");
            dtcDictionary.Add("B1720", "Seat Driver Forward Switch Circuit Short To Ground");
            dtcDictionary.Add("B1721", "Seat Driver Rearward Switch Circuit Failure");
            dtcDictionary.Add("B1722", "Seat Driver Rearward Switch Circuit Open");
            dtcDictionary.Add("B1723", "Seat Driver Rearward Switch Circuit Short To Battery");
            dtcDictionary.Add("B1724", "Seat Driver Rearward Switch Circuit Short To Ground");
            dtcDictionary.Add("B1725", "Seat Driver Rear Up Switch Circuit Failure");
            dtcDictionary.Add("B1726", "Seat Driver Rear Up Switch Circuit Open");
            dtcDictionary.Add("B1727", "Seat Driver Rear Up Switch Circuit Short To Battery");
            dtcDictionary.Add("B1728", "Seat Driver Rear Up Switch Circuit Short To Ground");
            dtcDictionary.Add("B1729", "Seat Driver Rear Down Switch Circuit Failure");
            dtcDictionary.Add("B1730", "Seat Driver Rear Down Switch Circuit Open");
            dtcDictionary.Add("B1731", "Seat Driver Rear Down Switch Circuit Short To Battery");
            dtcDictionary.Add("B1732", "Seat Driver Rear Down Switch Circuit Short To Ground");
            dtcDictionary.Add("B1733", "Mirror Driver Vertical Switch Circuit Failure");
            dtcDictionary.Add("B1734", "Mirror Driver Vertical Switch Circuit Open");
            dtcDictionary.Add("B1735", "Mirror Driver Vertical Switch Circuit Short To Battery");
            dtcDictionary.Add("B1736", "Mirror Driver Vertical Switch Circuit Short To Ground");
            dtcDictionary.Add("B1737", "Mirror Driver Horizontal Switch Circuit Failure");
            dtcDictionary.Add("B1738", "Mirror Driver Horizontal Switch Circuit Open");
            dtcDictionary.Add("B1739", "Mirror Driver Horizontal Switch Circuit Short To Battery");
            dtcDictionary.Add("B1740", "Mirror Driver Horizontal Switch Circuit Short To Ground");
            dtcDictionary.Add("B1741", "Mirror Passenger Vertical Switch Circuit Failure");
            dtcDictionary.Add("B1742", "Mirror Passenger Vertical Switch Circuit Open");
            dtcDictionary.Add("B1743", "Mirror Passenger Vertical Switch Circuit Short To Battery");
            dtcDictionary.Add("B1744", "Mirror Passenger Vertical Switch Circuit Short To Ground");
            dtcDictionary.Add("B1745", "Mirror Passenger Horizontal Switch Circuit Failure");
            dtcDictionary.Add("B1746", "Mirror Passenger Horizontal Switch Circuit Open");
            dtcDictionary.Add("B1747", "Mirror Passenger Horizontal Switch Circuit Short To Battery");
            dtcDictionary.Add("B1748", "Mirror Passenger Horizontal Switch Circuit Short To Ground");
            dtcDictionary.Add("B1749", "Park/Neutral Switch Circuit Failure");
            dtcDictionary.Add("B1750", "Park/Neutral Switch Circuit Open");
            dtcDictionary.Add("B1751", "Park/Neutral Switch Circuit Short To Battery");
            dtcDictionary.Add("B1752", "Park/Neutral Switch Circuit Short To Ground");
            dtcDictionary.Add("B1753", "Hazard Flash Output Circuit Failure");
            dtcDictionary.Add("B1754", "Hazard Flash Output Circuit Open");
            dtcDictionary.Add("B1755", "Hazard Flash Output Circuit Short Battery");
            dtcDictionary.Add("B1756", "Hazard Flash Output Circuit Short To Ground");
            dtcDictionary.Add("B1757", "Seat Driver Rear Down Circuit Failure");
            dtcDictionary.Add("B1758", "Seat Driver Rear Down Circuit Open");
            dtcDictionary.Add("B1759", "Seat Driver Rear Down Circuit Short To Battery");
            dtcDictionary.Add("B1760", "Seat Driver Rear Down Circuit Short To Ground");
            dtcDictionary.Add("B1761", "Seat Driver Front Down Circuit Failure");
            dtcDictionary.Add("B1762", "Seat Driver Front Down Circuit Open");
            dtcDictionary.Add("B1763", "Seat Driver Front Down Circuit Short To Battery");
            dtcDictionary.Add("B1764", "Seat Driver Front Down Circuit Short To Ground");
            dtcDictionary.Add("B1765", "Seat Driver Forward Circuit Failure");
            dtcDictionary.Add("B1766", "Seat Driver Forward Circuit Open");
            dtcDictionary.Add("B1767", "Seat Driver Forward Circuit Short To Battery");
            dtcDictionary.Add("B1768", "Seat Driver Forward Circuit Short To Ground");
            dtcDictionary.Add("B1769", "Seat Driver Backward Circuit Failure");
            dtcDictionary.Add("B1770", "Seat Driver Backward Circuit Open");
            dtcDictionary.Add("B1771", "Seat Driver Backward Circuit Short To Battery");
            dtcDictionary.Add("B1772", "Seat Driver Backward Circuit Short To Ground");
            dtcDictionary.Add("B1773", "Mirror Driver Up Circuit Failure");
            dtcDictionary.Add("B1774", "Mirror Driver Up Circuit Open");
            dtcDictionary.Add("B1775", "Mirror Driver Up Circuit Short To Battery");
            dtcDictionary.Add("B1776", "Mirror Driver Up Circuit Short To Ground");
            dtcDictionary.Add("B1777", "Driver's Seatbelt Tension Reducer Circuit Short to Ground");
            dtcDictionary.Add("B1778", "Mirror Driver Down Circuit Failure");
            dtcDictionary.Add("B1779", "Mirror Driver Down Circuit Open");
            dtcDictionary.Add("B1780", "Mirror Driver Down Circuit Short To Battery");
            dtcDictionary.Add("B1781", "Mirror Driver Down Short To Ground");
            dtcDictionary.Add("B1782", "Mirror Passenger Up Circuit Failure");
            dtcDictionary.Add("B1783", "Mirror Passenger Up Circuit Open");
            dtcDictionary.Add("B1784", "Mirror Passenger Up Circuit Short To Battery");
            dtcDictionary.Add("B1785", "Mirror Passenger Up Circuit Short To Ground");
            dtcDictionary.Add("B1786", "Mirror Passenger Down Circuit Failure");
            dtcDictionary.Add("B1787", "Mirror Passenger Down Circuit Open");
            dtcDictionary.Add("B1788", "Mirror Passenger Down Circuit Short To Battery");
            dtcDictionary.Add("B1789", "Mirror Passenger Down Circuit Short To Ground");
            dtcDictionary.Add("B1790", "Autolamp Sensor Input Circuit Failure");
            dtcDictionary.Add("B1791", "Autolamp Sensor Input Circuit Open");
            dtcDictionary.Add("B1792", "Autolamp Sensor Input Circuit Short To Battery");
            dtcDictionary.Add("B1793", "Autolamp Sensor Input Circuit Short To Ground");
            dtcDictionary.Add("B1794", "Lamp Headlamp Low-Beam Circuit Failure");
            dtcDictionary.Add("B1795", "Lamp Headlamp Low-Beam Circuit Open");
            dtcDictionary.Add("B1796", "Lamp Headlamp Low-Beam Circuit Short To Battery");
            dtcDictionary.Add("B1797", "Lamp Headlamp Low-Beam Circuit Short To Ground");
            dtcDictionary.Add("B1798", "Lamp Turn Signal Front Output Circuit Failure");
            dtcDictionary.Add("B1799", "Lamp Turn Signal Front Output Circuit Open");
            dtcDictionary.Add("B1800", "Lamp Turn Signal Front Output Circuit Short To Battery");
            dtcDictionary.Add("B1801", "Lamp Turn Signal Front Output Circuit Short To Ground");
            dtcDictionary.Add("B1802", "Lamp Turn Signal Rear Output Circuit Failure");
            dtcDictionary.Add("B1803", "Lamp Turn Signal Rear Output Circuit Open");
            dtcDictionary.Add("B1804", "Lamp Turn Signal Rear Output Circuit Short To Battery");
            dtcDictionary.Add("B1805", "Lamp Turn Signal Rear Output Circuit Short To Ground");
            dtcDictionary.Add("B1806", "Lamp Tail Output Circuit Failure");
            dtcDictionary.Add("B1807", "Lamp Tail Output Circuit Open");
            dtcDictionary.Add("B1808", "Lamp Tail Output Circuit Short To Battery");
            dtcDictionary.Add("B1809", "Lamp Tail Output Circuit Short To Ground");
            dtcDictionary.Add("B1810", "Lamp Backup Switch Input Circuit Failure");
            dtcDictionary.Add("B1811", "Lamp Backup Switch Input Circuit Open");
            dtcDictionary.Add("B1812", "Lamp Backup Switch Input Circuit Short To Battery");
            dtcDictionary.Add("B1813", "Lamp Backup Switch Input Circuit Short To Ground");
            dtcDictionary.Add("B1814", "Wiper Rear Motor Down Relay Coil Circuit Failure");
            dtcDictionary.Add("B1815", "Wiper Rear Motor Down Relay Coil Circuit Open");
            dtcDictionary.Add("B1816", "Wiper Rear Motor Down Relay Coil Circuit Short To Battery");
            dtcDictionary.Add("B1817", "Wiper Rear Motor Down Relay Coil Circuit Short To Ground");
            dtcDictionary.Add("B1818", "Wiper Rear Motor Up Relay Coil Circuit Failure");
            dtcDictionary.Add("B1819", "Wiper Rear Motor Up Relay Coil Circuit Open");
            dtcDictionary.Add("B1820", "Wiper Rear Motor Up Relay Coil Circuit Short To Battery");
            dtcDictionary.Add("B1821", "Wiper Rear Motor Up Relay Coil Circuit Short To Ground");
            dtcDictionary.Add("B1822", "Wiper Rear Park Sense Input Circuit Failure");
            dtcDictionary.Add("B1823", "Wiper Rear Park Sense Input Circuit Open");
            dtcDictionary.Add("B1824", "Wiper Rear Park Sense Input Circuit Short To Battery");
            dtcDictionary.Add("B1825", "Wiper Rear Park Sense Input Circuit Short To Ground");
            dtcDictionary.Add("B1826", "Wiper Rear High Limit Input Circuit Failure");
            dtcDictionary.Add("B1827", "Wiper Rear High Limit Input Circuit Open");
            dtcDictionary.Add("B1828", "Wiper Rear High Limit Input Circuit Short To Battery");
            dtcDictionary.Add("B1829", "Wiper Rear High Limit Input Circuit Short To Ground");
            dtcDictionary.Add("B1830", "Door Unlock Disarm Switch Circuit Failure");
            dtcDictionary.Add("B1831", "Door Unlock Disarm Switch Circuit Open");
            dtcDictionary.Add("B1832", "Door Unlock Disarm Switch Circuit Short To Battery");
            dtcDictionary.Add("B1833", "Door Unlock Disarm Switch Circuit Short To Ground");
            dtcDictionary.Add("B1834", "Door Unlock Disarm Output Circuit Failure");
            dtcDictionary.Add("B1835", "Door Unlock Disarm Output Circuit Open");
            dtcDictionary.Add("B1836", "Door Unlock Disarm Output Circuit Short To Battery");
            dtcDictionary.Add("B1837", "Door Unlock Disarm Output Circuit Short To Ground");
            dtcDictionary.Add("B1838", "Battery Saver Power Relay Circuit Failure");
            dtcDictionary.Add("B1839", "Wiper Rear Motor Circuit Failure");
            dtcDictionary.Add("B1840", "Wiper Front Power Circuit Failure");
            dtcDictionary.Add("B1841", "Wiper Front Power Circuit Open");
            dtcDictionary.Add("B1842", "Wiper Front Power Circuit Short To Battery");
            dtcDictionary.Add("B1843", "Wiper Front Power Circuit Short To Ground");
            dtcDictionary.Add("B1844", "Phone Handset Circuit Failure");
            dtcDictionary.Add("B1845", "Ignition Tamper Circuit Failure");
            dtcDictionary.Add("B1846", "Ignition Tamper Circuit Open");
            dtcDictionary.Add("B1847", "Ignition Tamper Circuit Short To Battery");
            dtcDictionary.Add("B1848", "Ignition Tamper Circuit Short To Ground");
            dtcDictionary.Add("B1849", "Climate Control Temperature Differential Circuit Failure");
            dtcDictionary.Add("B1850", "Climate Control Temperature Differential Circuit Open");
            dtcDictionary.Add("B1851", "Climate Control Temperature Differential Circuit Short To Battery");
            dtcDictionary.Add("B1852", "Climate Control Temperature Differential Circuit Short To Ground");
            dtcDictionary.Add("B1853", "Climate Control Air Temperature Internal Sensor Motor Circuit Failure");
            dtcDictionary.Add("B1854", "Climate Control Air Temperature Internal Sensor Motor Circuit Open");
            dtcDictionary.Add("B1855", "Climate Control Air Temperature Internal Sensor Motor Circuit Short To Battery");
            dtcDictionary.Add("B1856", "Climate Control Air Temperature Internal Sensor Motor Circuit Short To Ground");
            dtcDictionary.Add("B1857", "Climate Control On/Off Switch Circuit Failure");
            dtcDictionary.Add("B1858", "Climate Control A/C Pressure Switch Circuit Failure");
            dtcDictionary.Add("B1859", "Climate Control A/C Pressure Switch Circuit Open");
            dtcDictionary.Add("B1860", "Climate Control A/C Pressure Switch Circuit Short To Battery");
            dtcDictionary.Add("B1861", "Climate Control A/C Pressure Switch Circuit Short To Ground");
            dtcDictionary.Add("B1862", "Climate Control A/C Lock Sensor Failure");
            dtcDictionary.Add("B1863", "Ground ECU Circuit Open");
            dtcDictionary.Add("B1864", "Battery Power Supply ECU Circuit Failure");
            dtcDictionary.Add("B1865", "Battery Power Supply ECU Circuit Open");
            dtcDictionary.Add("B1866", "Battery Power Supply ECU Circuit Short To Battery");
            dtcDictionary.Add("B1867", "Battery Power Supply ECU Circuit Short To Ground");
            dtcDictionary.Add("B1868", "Lamp Air Bag Warning Indicator Circuit Failure");
            dtcDictionary.Add("B1869", "Lamp Air Bag Warning Indicator Circuit Open");
            dtcDictionary.Add("B1870", "Lamp Air Bag Warning Indicator Circuit Short To Battery");
            dtcDictionary.Add("B1871", "Passenger Air Bag Disable Module Fault");
            dtcDictionary.Add("B1872", "Turn Signal / Hazard Power Feed Circuit Short To Battery");
            dtcDictionary.Add("B1873", "Turn Signal / Hazard Power Feed Circuit Short To Ground");
            dtcDictionary.Add("B1874", "Cellular Phone Handset Not Present");
            dtcDictionary.Add("B1875", "Turn Signal / Hazard Switch Signal Circuit Failure");
            dtcDictionary.Add("B1876", "Seatbelt Driver Pretensioner Circuit Failure");
            dtcDictionary.Add("B1877", "Seatbelt Driver Pretensioner Circuit Open");
            dtcDictionary.Add("B1878", "Seatbelt Driver Pretensioner Circuit Short to Battery");
            dtcDictionary.Add("B1879", "Seatbelt Driver Pretensioner Circuit Short to Ground");
            dtcDictionary.Add("B1880", "Seatbelt Passenger Pretensioner Circuit Failure");
            dtcDictionary.Add("B1881", "Seatbelt Passenger Pretensioner Circuit Open");
            dtcDictionary.Add("B1882", "Seatbelt Passenger Pretensioner Circuit Short to Battery");
            dtcDictionary.Add("B1883", "Seatbelt Passenger Pretensioner Circuit Short to Ground");
            dtcDictionary.Add("B1884", "PAD Warning Lamp Inoperative");
            dtcDictionary.Add("B1885", "Seatbelt Driver Pretensioner Circuit Resistance Low on Squib");
            dtcDictionary.Add("B1886", "Seatbelt Passenger Pretensioner Circuit Resistance Low on Squib");
            dtcDictionary.Add("B1887", "Air Bag Driver Circuit Resistance Low or Shorted Together");
            dtcDictionary.Add("B1888", "Air Bag Passenger Circuit Resistance Low or Shorted Together");
            dtcDictionary.Add("B1889", "Passenger Airbag Disable Module Sensor Obstructed");
            dtcDictionary.Add("B1890", "PAD Warning Lamp Circuit Short to Battery");
            dtcDictionary.Add("B1891", "Air Bag Tone Warning Indicator Circuit Short to Battery");
            dtcDictionary.Add("B1892", "Air Bag Tone Warning Indicator Circuit Failure");
            dtcDictionary.Add("B1893", "GPS Antenna Open Circuit");
            dtcDictionary.Add("B1894", "Wiper Rear Motor Speed Sense Circuit Failure");
            dtcDictionary.Add("B1895", "Driver's / Passenger's Door Ajar Output Circuit Failure");
            dtcDictionary.Add("B1896", "Driver's / Passenger's Door Ajar Output Circuit Short to Battery");
            dtcDictionary.Add("B1897", "Horn Switch Circuit Failure");
            dtcDictionary.Add("B1898", "Chime Input #2 Circuit Short to Ground");
            dtcDictionary.Add("B1899", "Microphone Input Signal Circuit Open");
            dtcDictionary.Add("B1900", "Driver Side Airbag Fault");
            dtcDictionary.Add("B1901", "Air Bag Crash Sensor #1 Feed/Return Circuit Short To Ground");
            dtcDictionary.Add("B1902", "Air Bag Crash Sensor #1 Ground Circuit Failure");
            dtcDictionary.Add("B1903", "Air Bag Crash Sensor #1 Ground Circuit Short To Battery");
            dtcDictionary.Add("B1904", "Air Bag Crash Sensor #2 Feed/Return Circuit Failure");
            dtcDictionary.Add("B1905", "Air Bag Crash Sensor #2 Feed/Return Circuit Short To Battery");
            dtcDictionary.Add("B1906", "Air Bag Crash Sensor #2 Feed/Return Circuit Short To Ground");
            dtcDictionary.Add("B1907", "Air Bag Crash Sensor #2 Ground Circuit Failure");
            dtcDictionary.Add("B1908", "Air Bag Crash Sensor #2 Ground Circuit Short To Battery");
            dtcDictionary.Add("B1909", "Air Bag Crash Sensor #2 Ground Circuit Short To Ground");
            dtcDictionary.Add("B1910", "Air Bag Diagnostic Monitor Ground Circuit Failure");
            dtcDictionary.Add("B1911", "Air Bag Diagnostic Monitor Ground Circuit Short To Battery");
            dtcDictionary.Add("B1912", "Air Bag Diagnostic Monitor Ground Circuit Short To Ground");
            dtcDictionary.Add("B1913", "Air Bag Driver/Passenger Circuit Short To Ground");
            dtcDictionary.Add("B1914", "Air Bag Crash Sensors #1 / #2 Circuit Short To Ground");
            dtcDictionary.Add("B1915", "Air Bag Driver Circuit Failure");
            dtcDictionary.Add("B1916", "Air Bag Driver Circuit Short To Battery");
            dtcDictionary.Add("B1917", "Air Bag Memory Clear Circuit Failure");
            dtcDictionary.Add("B1918", "Air Bag Memory Clear Circuit Open");
            dtcDictionary.Add("B1919", "Air Bag Memory Clear Circuit Short To Battery");
            dtcDictionary.Add("B1920", "Air Bag Passenger Circuit Failure");
            dtcDictionary.Add("B1921", "Air Bag Diagnostic Monitor Ground Circuit Open");
            dtcDictionary.Add("B1922", "Air Bag Safing Sensor Output Circuit Short To Battery");
            dtcDictionary.Add("B1923", "Air Bag Memory Clear Circuit Short To Ground");
            dtcDictionary.Add("B1924", "Air Bag Internal Diagnostic Monitor Fault or System Disarm Fault");
            dtcDictionary.Add("B1925", "Air Bag Passenger Circuit Short To Battery");
            dtcDictionary.Add("B1926", "Air Bag Passenger Pressure Switch Circuit Failure");
            dtcDictionary.Add("B1927", "Passenger Side Airbag Fault");
            dtcDictionary.Add("B1928", "Air Bag Safing Sensor Output Circuit Failure");
            dtcDictionary.Add("B1929", "Air Bag Safing Sensor Output Circuit Open");
            dtcDictionary.Add("B1930", "Air Bag Safing Sensor Output Circuit Short To Ground");
            dtcDictionary.Add("B1931", "Air Bag Crash Sensor #1 Feed/Return Circuit Failure");
            dtcDictionary.Add("B1932", "Air Bag Driver Circuit Open");
            dtcDictionary.Add("B1933", "Air Bag Passenger Circuit Open");
            dtcDictionary.Add("B1934", "Air Bag Driver Inflator Circuit Resistance Low on Squib");
            dtcDictionary.Add("B1935", "Air Bag Passenger Inflator Circuit Resistance Low on Squib");
            dtcDictionary.Add("B1936", "Air Bag Driver Circuit Short To Ground");
            dtcDictionary.Add("B1937", "Air Bag Passenger Pressure Switch Circuit Open");
            dtcDictionary.Add("B1938", "Air Bag Passenger Circuit Short To Ground");
            dtcDictionary.Add("B1939", "Air Bag Passenger Pressure Switch Circuit Short To Ground");
            dtcDictionary.Add("B1941", "Air Bag Crash Sensor #1 Feed/Return Circuit Open");
            dtcDictionary.Add("B1942", "Air Bag Crash Sensor #2 Feed/Return Circuit Open");
            dtcDictionary.Add("B1943", "Air Bag Crash Sensor #1 Ground Circuit Short To Ground");
            dtcDictionary.Add("B1944", "Air Bag Crash Sensor #1 Ground Circuit Open");
            dtcDictionary.Add("B1945", "Air Bag Crash Sensor #2 Ground Circuit Open");
            dtcDictionary.Add("B1946", "Climate Control A/C Post Evaporator Sensor Circuit Failure");
            dtcDictionary.Add("B1947", "Climate Control A/C Post Evaporator Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1948", "Climate Control Water Temperature Sensor Circuit Failure");
            dtcDictionary.Add("B1949", "Climate Control Water Temperature Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1950", "Seat Rear Up/Down Potentiometer Feedback Circuit Failure");
            dtcDictionary.Add("B1951", "Seat Rear Up/Down Potentiometer Feedback Circuit Open");
            dtcDictionary.Add("B1952", "Seat Rear Up/Down Potentiometer Feedback Circuit Short To Battery");
            dtcDictionary.Add("B1953", "Seat Rear Up/Down Potentiometer Feedback Circuit Short To Ground");
            dtcDictionary.Add("B1954", "Seat Front Up/Down Potentiometer Feedback Circuit Failure");
            dtcDictionary.Add("B1955", "Seat Front Up/Down Potentiometer Feedback Circuit Open");
            dtcDictionary.Add("B1956", "Seat Front Up/Down Potentiometer Feedback Circuit Short To Battery");
            dtcDictionary.Add("B1957", "Seat Front Up/Down Potentiometer Feedback Circuit Short To Ground");
            dtcDictionary.Add("B1958", "Seat Recline Forward/Backward Potentiometer Feedback Circuit Failure");
            dtcDictionary.Add("B1959", "Seat Recline Forward/Backward Potentiometer Feedback Circuit Open");
            dtcDictionary.Add("B1960", "Seat Recline Forward/Backward Potentiometer Feedback Circuit Short To Battery");
            dtcDictionary.Add("B1961", "Seat Recline Forward/Backward Potentiometer Feedback Circuit Short To Ground");
            dtcDictionary.Add("B1962", "Seat Horizontal Forward/Rearward Potentiometer Feedback Circuit Failure");
            dtcDictionary.Add("B1963", "Seat Horizontal Forward/Rearward Potentiometer Feedback Circuit Open");
            dtcDictionary.Add(
                "B1964",
                "Seat Horizontal Forward/Rearward Potentiometer Feedback Circuit Short To Battery");
            dtcDictionary.Add(
                "B1965",
                "Seat Horizontal Forward/Rearward Potentiometer Feedback Circuit Short To Ground");
            dtcDictionary.Add("B1966", "A/C Post Heater Sensor Circuit Failure");
            dtcDictionary.Add("B1967", "A/C Post Heater Sensor Circuit Short To Ground");
            dtcDictionary.Add("B1968", "A/C Water Pump Detection Circuit Failure");
            dtcDictionary.Add("B1969", "A/C Clutch Magnetic Control Circuit Failure");
            dtcDictionary.Add("B1970", "Passenger Seatback Forward Switch Circuit Short to Ground");
            dtcDictionary.Add("B1971", "Passenger Seatback Rearward Switch Circuit Short to Ground");
            dtcDictionary.Add("B1972", "Passenger Rear Seat Up Switch Circuit Short to Battery");
            dtcDictionary.Add("B1973", "Passenger Rear Seat Down Switch Circuit Short to Battery");
            dtcDictionary.Add("B1974", "Passenger's Seat Recline Forward Switch Circuit Short to Battery");
            dtcDictionary.Add("B1975", "Passenger's Seat Recline Back Switch Circuit Short to Battery");
            dtcDictionary.Add("B1976", "Passenger's Seat Forward Switch Circuit Short to Battery");
            dtcDictionary.Add("B1977", "Passenger's Front Seat Up Switch Circuit Short to Battery");
            dtcDictionary.Add("B1978", "Passenger's Front Seat Down Switch Circuit Short to Battery");
            dtcDictionary.Add("B1979", "Passenger Seat Rearward Switch Circuit Short to Battery");
            dtcDictionary.Add("B1980", "Bulb - Outage Condition Detected");
            dtcDictionary.Add("B1981", "Memory Off Switch Circuit Short to Battery");
            dtcDictionary.Add("B1982", "Driver's Door Unlock Relay Circuit Failure");
            dtcDictionary.Add("B1983", "Driver's Door Unlock Relay Circuit Short to Battery");
            dtcDictionary.Add("B1984", "Seat Switch Lumbar Inflate Circuit Failure");
            dtcDictionary.Add("B1985", "Seat Switch Lumbar Deflate Circuit Failure");
            dtcDictionary.Add("B1986", "Driver's Seat Seatback Autoglide Rearward Switch Circuit Failure");
            dtcDictionary.Add("B1987", "Pedal Forward / Rearward Motor Stalled");
            dtcDictionary.Add("B1988", "Pedal Position Forward Switch Circuit Short to Battery");
            dtcDictionary.Add("B1989", "Pedal Position Rearward Switch Circuit Short to Battery");
            dtcDictionary.Add("B1990", "Pedal Forward / Rearward Potentiometer Feedback Circuit Failure");
            dtcDictionary.Add("B1991", "Pedal Forward / Rearward Potentiometer Feedback Circuit Short to Battery");
            dtcDictionary.Add("B1992", "Driver Side, Side mount Airbag Circuit Short to Vbatt");
            dtcDictionary.Add("B1993", "Driver Sid, Side mount Airbag Circuit Short to Ground");
            dtcDictionary.Add("B1994", "Driver Sid, Side mount Airbag Circuit Open");
            dtcDictionary.Add("B1995", "Driver Sid, Side mount Airbag Low resistance on Squib");
            dtcDictionary.Add("B1996", "Passenger Sid, Side mount Airbag Circuit Short to Vbatt");
            dtcDictionary.Add("B1997", "Passenger Sid, Side mount Airbag Circuit Short to Ground");
            dtcDictionary.Add("B1998", "Passenger Sid, Side mount Airbag Circuit Open");
            dtcDictionary.Add("B1999", "Passenger Sid, Side mount Airbag Low resistance on Squib");
            dtcDictionary.Add("B2100", "Door Driver Key Cylinder Switch Failure");
            dtcDictionary.Add("B2101", "Head Rest Switch Circuit Failure");
            dtcDictionary.Add("B2102", "Antenna Circuit Short to Ground");
            dtcDictionary.Add("B2103", "Antenna Not Connected");
            dtcDictionary.Add("B2104", "Door Passenger Key Cylinder Switch Failure");
            dtcDictionary.Add("B2105", "Throttle Position Input Out of Range Low");
            dtcDictionary.Add("B2106", "Throttle Position Input Out of Range High");
            dtcDictionary.Add("B2107", "Front Wiper Motor Relay Circuit Short to Vbatt");
            dtcDictionary.Add("B2108", "Trunk Key Cylinder Switch Failure");
            dtcDictionary.Add("B2109", "Heated Wind Shield Relay Short to Vbatt (changed from Failure 2/6/97)");
            dtcDictionary.Add("B2110", "Front Wiper Motor Relay Circuit Open (changed from Failure 2/6/97)");
            dtcDictionary.Add("B2111", "All Door Lock Input Short to Ground");
            dtcDictionary.Add("B2112", "Door Driver Set Switch Stuck Failure");
            dtcDictionary.Add("B2113", "Heated Windshield Input Short to Ground");
            dtcDictionary.Add("B2114", "Front Washer Input Short to Ground");
            dtcDictionary.Add("B2115", "Rear Washer Input Short to Ground");
            dtcDictionary.Add("B2116", "Door Driver Reset Switch Stuck Failure");
            dtcDictionary.Add("B2117", "Driver Side, Side mount Airbag Low capacitance on Squib");
            dtcDictionary.Add("B2118", "Passenger Sid, Side mount Airbag Low capacitance on Squib");
            dtcDictionary.Add("B2119", "Compressor Failure");
            dtcDictionary.Add("B2120", "Door Passenger Set Switch Stuck Failure");
            dtcDictionary.Add("B2122", "Driver Side Satellite Communication Circuit Short to Ground");
            dtcDictionary.Add("B2123", "Passenger Side Satellite Communication Circuit Short to Ground");
            dtcDictionary.Add("B2124", "Door Passenger Reset Switch Stuck Failure");
            dtcDictionary.Add("B2128", "Central Lock Motor Failure");
            dtcDictionary.Add("B2129", "Central Lock Feedback Failure");
            dtcDictionary.Add("B2130", "Double Lock Timeout Failure");
            dtcDictionary.Add("B2131", "Double Lock Feedback Failure");
            dtcDictionary.Add("B2132", "Dimmer switch Circuit Short to Gnd");
            dtcDictionary.Add("B2133", "Brake Motor Warning lamp Circuit Failure");
            dtcDictionary.Add("B2134", "Brake Motor Warning lamp Circuit Short to Vbatt");
            dtcDictionary.Add("B2135", "Park Brake Applied Warning Lamp Circuit Failure");
            dtcDictionary.Add("B2136", "Park Brake Applied Warning Lamp Circuit Short To Battery");
            dtcDictionary.Add("B2139", "Data Mismatch (receive data does not match what was expected)");
            dtcDictionary.Add("B2141", "NVM Configuration Failure");
            dtcDictionary.Add("B2142", "NVM TIC Failure");
            dtcDictionary.Add("B2143", "NVM Memory Failure");
            dtcDictionary.Add("B2144", "NVM Alarm Data Failure");
            dtcDictionary.Add("B2145", "NVM RF HR Failure");
            dtcDictionary.Add("B2146", "Seat Recline Motor Position Out of Range");
            dtcDictionary.Add("B2148", "PWM Input Circuit Failure");
            dtcDictionary.Add("B2149", "Seat Front Vertical Motor Position Out of Range");
            dtcDictionary.Add("B2150", "Power Supply #1 Circuit Short to Ground");
            dtcDictionary.Add("B2151", "Power Supply #2 Circuit Short to Ground");
            dtcDictionary.Add("B2152", "Seat Rear Vertical Motor Position Out of Range");
            dtcDictionary.Add("B2153", "Rear Echo Sensor Circuit Failure");
            dtcDictionary.Add("B2154", "Front Echo Sensor Circuit Failure");
            dtcDictionary.Add("B2155", "Seat Horizontal Motor Position Out of Range");
            dtcDictionary.Add("B2156", "Rear Doppler Sensor Circuit Failure");
            dtcDictionary.Add("B2157", "Front Doppler Sensor Circuit Failure");
            dtcDictionary.Add("B2158", "Seat Recline Motor Memory Position Out of Range");
            dtcDictionary.Add("B2159", "Memory #1 output Short to Ground");
            dtcDictionary.Add("B2160", "Memory #1 output Short to VBatt");
            dtcDictionary.Add("B2161", "Seat Front Vertical Motor Memory Position Out of Range");
            dtcDictionary.Add("B2162", "Data Mismatch #2 (receive data does not match what was expected)");
            dtcDictionary.Add("B2163", "Clutch Position Fault");
            dtcDictionary.Add("B2164", "Seat Rear Vertical Motor Memory Position Out of Range");
            dtcDictionary.Add("B2165", "Gear shift position Fault");
            dtcDictionary.Add("B2166", "Gear select position Fault");
            dtcDictionary.Add("B2167", "Seat Horizontal Motor Memory Position Out of Range");
            dtcDictionary.Add("B2168", "Unable to confirm Unlock Condition");
            dtcDictionary.Add("B2169", "Unable to confirm lock Condition");
            dtcDictionary.Add("B2170", "Steering Column Lock Switch Circuit Failure");
            dtcDictionary.Add("B2172", "Inertia Switch input Circuit Open");
            dtcDictionary.Add("B2174", "Window Driver Rear Remote Up Switch Short to Battery");
            dtcDictionary.Add("B2175", "A/C Request Signal Circuit Short to Ground");
            dtcDictionary.Add("B2176", "Overdrive switch circuit short to Vbatt");
            dtcDictionary.Add("B2177", "Interior Scanning Sensor Circuit Failure");
            dtcDictionary.Add("B2178", "Window Driver Rear Remote Down Switch Short to Battery");
            dtcDictionary.Add("B2179", "Front Wiper Select Switch 'A' Short to Ground");
            dtcDictionary.Add("B2180", "Front Wiper Select Switch 'B' Short to Ground");
            dtcDictionary.Add("B2181", "Front Wiper Select Switch 'C' Short to Ground");
            dtcDictionary.Add("B2182", "Window Passenger Front Remote Up Switch Short to Battery");
            dtcDictionary.Add("B2183", "Front Wiper Select Switch 'H' Short to Ground");
            dtcDictionary.Add("B2184", "Front Wiper Select Switch 'W' Short to Ground");
            dtcDictionary.Add("B2185", "Rear Wiper Select Switch 'D' Short to Ground");
            dtcDictionary.Add("B2186", "Window Passenger Front Remote Down Switch Short to Battery");
            dtcDictionary.Add("B2187", "Rear Wiper Select Switch 'B' Short to Ground");
            dtcDictionary.Add("B2188", "Rear Wiper Select Switch 'E' Short to Ground");
            dtcDictionary.Add("B2190", "Window Passenger Rear Remote Up Switch Short to Battery");
            dtcDictionary.Add("B2194", "Window Passenger Rear Remote Down Switch Short to Battery");
            dtcDictionary.Add("B2195", "Driver Window Up / Down Power Circuit Short to Ground");
            dtcDictionary.Add("B2196", "Passenger Window Up / Down Power Circuit Short to Ground");
            dtcDictionary.Add("B2197", "TV Module Error");
            dtcDictionary.Add("B2198", "TrafficMaster Module Error");
            dtcDictionary.Add("B2199", "VICS Module Error");
            dtcDictionary.Add("B2200", "No Communication to TV Module (No Fitting of TV)");
            dtcDictionary.Add("B2201", "No Communication With Traffic MasterModule");
            dtcDictionary.Add("B2202", "No Communication to VICS Module (No Fitting of VICS)");
            dtcDictionary.Add("B2203", "CD-ROM Error");
            dtcDictionary.Add("B2204", "GPS Antenna Connection Open or Short");
            dtcDictionary.Add("B2205", "GPS Receiver Error");
            dtcDictionary.Add("B2206", "Gyroscope Error");
            dtcDictionary.Add("B2207", "ECU ROM Checksum Error");
            dtcDictionary.Add("B2208", "Communication Link to Display and Switch Module Error");
            dtcDictionary.Add("B2209", "Interior Lamp Override Switch Open Circuit");
            dtcDictionary.Add("B2210", "Interior Lamp Override Switch Short to Ground");
            dtcDictionary.Add("B2211", "Low Coolant Lamp Output Circuit Short to Battery");
            dtcDictionary.Add("B2214", "Window Passenger Front Up Switch Short to Battery");
            dtcDictionary.Add("B2215", "Window Passenger Front Down Switch Short to Battery");
            dtcDictionary.Add("B2219", "Window Driver Front Current Feedback Exceeded");
            dtcDictionary.Add("B2220", "Window Driver Rear Current Feedback Exceeded");
            dtcDictionary.Add("B2221", "Window Passenger Front Current Feedback Exceeded");
            dtcDictionary.Add("B2222", "Window Passenger Rear Current Feedback Exceeded");
            dtcDictionary.Add("B2223", "Mirror Driver Drive Circuit Failure");
            dtcDictionary.Add("B2224", "Mirror Passenger Drive Circuit Failure");
            dtcDictionary.Add("B2225", "Front Crash Sensor Mount Fault");
            dtcDictionary.Add("B2226", "Front Crash Sensor Internal Fault");
            dtcDictionary.Add("B2227", "Front Crash Sensor Driver Communications Fault");
            dtcDictionary.Add("B2228", "Air Bag Driver Circuit Short to Ground - Loop #2");
            dtcDictionary.Add("B2229", "Air Bag Passenger Circuit Short to Ground - Loop #2");
            dtcDictionary.Add("B2230", "Air Bag Driver Circuit Short to Battery - Loop #2");
            dtcDictionary.Add("B2231", "Air Bag Passenger Circuit Short to Battery - Loop #2");
            dtcDictionary.Add("B2232", "Air Bag Driver Circuit Open - Loop #2");
            dtcDictionary.Add("B2233", "Air Bag passenger Circuit Open - Loop #2");
            dtcDictionary.Add("B2234", "Air Bag Driver Inflator Circuit Resistance Low on Squib - Loop #2");
            dtcDictionary.Add("B2235", "Air Bag Passenger Inflator Circuit");
            dtcDictionary.Add("B2236", "Weak or Defected Electric Vehicle Battery Module Fault");
            dtcDictionary.Add("B2237", "Vehicle Signal indicating Park While VSS Present");
            dtcDictionary.Add("B2238", "Power Cable For Power Sliding Door Broken");
            dtcDictionary.Add("B2239", "Rear Cargo Door Set Switch Stuck (Short to Ground)");
            dtcDictionary.Add("B2240", "Rear Cargo Door Reset Switch Stuck (Short to Ground)");
            dtcDictionary.Add("B2241", "Rear Cargo Door Lock Circuit Short to Ground");
            dtcDictionary.Add("B2242", "Rear Cargo Door Unlock Circuit Open");
            dtcDictionary.Add("B2243", "Driver Rear Door Ajar Circuit Open");
            dtcDictionary.Add("B2244", "Driver Sliding Door Ajar Circuit Short to GND");
            dtcDictionary.Add("B2245", "Passenger Rear Door Ajar Circuit Open");
            dtcDictionary.Add("B2246", "Passenger Sliding Door Ajar Circuit Short to GND");
            dtcDictionary.Add("B2247", "EV Battery Pack Temperature Fault");
            dtcDictionary.Add("B2248", "Heated Windshield Relay Coil Circuit Failure");
            dtcDictionary.Add("B2249", "Head Lamp Relay Coil Short to Battery");
            dtcDictionary.Add("B2250", "All Doors Unlock Relay Circuit Failure");
            dtcDictionary.Add("B2251", "Parklamp Output Relay Driver Circuit Failure");
            dtcDictionary.Add("B2252", "Parklamp Output Relay Dirver Short to Battery");
            dtcDictionary.Add("B2300", "Seat Driver Memory Position Error");
            dtcDictionary.Add("B2301", "Seat Passenger Memory Position Error");
            dtcDictionary.Add("B2302", "Seat Headrest Feedback Potentiometer Circuit Failure");
            dtcDictionary.Add("B2303", "Seat Headrest Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2304", "Seat Headrest Feedback Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("B2305", "Seat Headrest Feedback Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("B2306", "Seat Headrest Motor Stalled");
            dtcDictionary.Add("B2310", "Mirror Driver Memory Position Error");
            dtcDictionary.Add("B2311", "Mirror Passenger Memory Position Error");
            dtcDictionary.Add("B2312", "Mirror Passenger Horizontal Feedback Potentiometer Circuit Failure");
            dtcDictionary.Add("B2313", "Mirror Passenger Horizontal Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2314", "Mirror Passenger Horizontal Feedback Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("B2315", "Mirror Passenger Horizontal Feedback Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("B2316", "Mirror Passenger Vertical Feedback Potentiometer Circuit Failure");
            dtcDictionary.Add("B2317", "Mirror Passenger Vertical Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2318", "Mirror Passenger Vertical Feedback Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("B2319", "Mirror Passenger Vertical Feedback Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("B2320", "Mirror Driver Horizontal Feedback Potentiometer Circuit Failure");
            dtcDictionary.Add("B2321", "Mirror Driver Horizontal Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2322", "Mirror Driver Horizontal Feedback Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("B2323", "Mirror Driver Horizontal Feedback Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("B2324", "Mirror Driver Vertical Feedback Potentiometer Circuit Failure");
            dtcDictionary.Add("B2325", "Mirror Driver Vertical Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2326", "Mirror Driver Vertical Feedback Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("B2327", "Mirror Driver Vertical Feedback Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("B2328", "Column Reach Feedback Potentiometer Circuit Failure");
            dtcDictionary.Add("B2329", "Column Reach Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2330", "Column Reach Feedback Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("B2331", "Column Reach Feedback Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("B2332", "Column Tilt Feedback Potentiometer Circuit Failure");
            dtcDictionary.Add("B2333", "Column Tilt Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2334", "Column Tilt Feedback Potentiometer Circuit Short to Battery");
            dtcDictionary.Add("B2335", "Column Tilt Feedback Potentiometer Circuit Short to Ground");
            dtcDictionary.Add("B2336", "Mirror Switch Assembly Circuit Failure");
            dtcDictionary.Add("B2337", "Mirror Switch Assembly Circuit Open");
            dtcDictionary.Add("B2338", "Mirror Switch Assembly Circuit Short to Battery");
            dtcDictionary.Add("B2339", "Mirror Switch Assembly Circuit Short to Ground");
            dtcDictionary.Add("B2340", "Column Reach Motor Stalled");
            dtcDictionary.Add("B2341", "Column Tilt Motor Stalled");
            dtcDictionary.Add("B2342", "Seat Switch Reference Voltage Positive Common Open Circuit");
            dtcDictionary.Add("B2343", "Seat Switch Reference Voltage Positive Common Supply Low Voltage");
            dtcDictionary.Add("B2344", "Seat Switch Reference Voltage Positive Common Supply Voltage Fault");
            dtcDictionary.Add("B2345", "Seat Switch Reference Voltage Negative Common Open Circuit");
            dtcDictionary.Add("B2346", "Mirror Switch Reference Voltage Positive Common Open Circuit");
            dtcDictionary.Add("B2347", "Mirror Switch Reference Voltage Positive Common Supply Low Voltage");
            dtcDictionary.Add("B2348", "Mirror Switch Reference Voltage Positive Common Supply Voltage Fault");
            dtcDictionary.Add("B2349", "Mirror Switch Reference Voltage Negative Common Open Circuit");
            dtcDictionary.Add("B2350", "Steering Column Switch Circuit Short to Battery");
            dtcDictionary.Add("B2351", "Steering Column Switch Circuit Failure");
            dtcDictionary.Add("B2352", "Driver Memory Power Switch Indicator Circuit Short to Battery");
            dtcDictionary.Add("B2353", "Driver Mirror Power Driver Circuit Short Ground");
            dtcDictionary.Add("B2354", "Driver Mirror Horizontal / Vertical Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2355", "Passenger Mirror Horizontal / Vertical Feedback Potentiometer Circuit Open");
            dtcDictionary.Add("B2357", "Driver Window Down Current Sense Low Circuit Failure");
            dtcDictionary.Add("B2362", "Remote Open/Close signal Circuit Short to Ground");
            dtcDictionary.Add("B2363", "Optical Sensor System Failure");
            dtcDictionary.Add("B2364", "Fuel Filler Door Circuit Open");
            dtcDictionary.Add("B2365", "B-pillar Power Sliding Door Open/Close Switch Input Ckt Short to Gnd");
            dtcDictionary.Add("B2366", "IP Power Sliding Door Open/Close switch Ckt Short to Gnd");
            dtcDictionary.Add("B2367", "Power Sliding Door Override Switch Input Ckt Short to Gnd");
            dtcDictionary.Add("B2368", "Steering Column Switch Circuit Out of Range");
            dtcDictionary.Add("B2369", "Chime OUTPUT Request Ckt Short to Ground");
            dtcDictionary.Add("B2373", "LED #1 Circuit Short to Battery");
            dtcDictionary.Add("B2374", "Power Sliding Detent (Latch) Circuit Failure");
            dtcDictionary.Add("B2380", "Heater Coolant Temp sensor circuit Short to GND");
            dtcDictionary.Add("B2381", "Heater Coolant Temp sensor circuit Open");
            dtcDictionary.Add("B2384", "Audio Reverse Aid Mute Input Ckt Failure");
            dtcDictionary.Add("B2385", "Audio Navigation Mute Input Ckt Failure");
            dtcDictionary.Add("B2401", "Audio Tape Deck Mechanism Fault");
            dtcDictionary.Add("B2402", "Audio CD/DJ Thermal Shutdown Fault");
            dtcDictionary.Add("B2403", "Audio CD/DJ Internal Fault");
            dtcDictionary.Add("B2404", "Audio Steering Wheel Switch Circuit Fault");
            dtcDictionary.Add("B2405", "Audio Single-Disc CD Player Thermal Shutdown Fault");
            dtcDictionary.Add("B2406", "Audio Single-Disc CD Player Internal Fault");
            dtcDictionary.Add("B2416", "Climate Control Recirculation Actuator Out of Limits");
            dtcDictionary.Add("B2425", "Remote Keyless Entry Out of Synchronization");
            dtcDictionary.Add("B2426", "Passenger Solar Radiation Sensor Circuit Open");
            dtcDictionary.Add("B2427", "Passenger Solar Radiation Sensor Circuit Short to Ground");
            dtcDictionary.Add("B2428", "A/C Post Heater Sensor #2 Circuit Failure");
            dtcDictionary.Add("B2429", "A/C Post Heater Sensor #2 Circuit Short to Ground");
            dtcDictionary.Add("B2431", "Transponder Programming Failed");
            dtcDictionary.Add("B2432", "Drivers Seat Belt Buckle Switch Circuit Open");
            dtcDictionary.Add("B2433", "Drivers Seat Belt Buckle Switch Circuit Short to Battery");
            dtcDictionary.Add("B2434", "Drivers Seat Belt Buckle Switch Circuit Short to Ground");
            dtcDictionary.Add("B2435", "Drivers Seat Belt Buckle Switch Resistance out of Range");
            dtcDictionary.Add("B2436", "Passengers Seat Belt Buckle Switch Circuit Open");
            dtcDictionary.Add("B2437", "Passengers Seat Belt Buckle Switch Circuit Short to Battery");
            dtcDictionary.Add("B2438", "Passengers Seat Belt Buckle Switch Circuit Short to Ground");
            dtcDictionary.Add("B2439", "Passengers Seat Belt Buckle Switch Resistance out of Range");
            dtcDictionary.Add("B2440", "Passenger Sid, Side crash sensor mount fault");
            dtcDictionary.Add("B2441", "Driver Sid, Side crash sensor mount fault");
            dtcDictionary.Add("B2442", "Intrusion Sensor Fault");
            dtcDictionary.Add("B2443", "Powertrain Performance Mode Switch Circuit failure");
            dtcDictionary.Add("B2444", "Driver Side Crash Sensor Internal Fault");
            dtcDictionary.Add("B2445", "Passenger Side Crash Sensor Internal Fault");
            dtcDictionary.Add("B2446", "RESCU/VEMS Input Circuit Open");
            dtcDictionary.Add("B2447", "RESCU/VEMS Input Circuit Short to Battery");
            dtcDictionary.Add("B2448", "RESCU/VEMS Input Circuit Short to Ground");
            dtcDictionary.Add("B2449", "Aux Heater Glow Plug Circuit Short to Ground");
            dtcDictionary.Add("B2450", "Aux Heater Glow Plug Circuit Open");
            dtcDictionary.Add("B2451", "Aux Heater Fuel Pump Circuit Short to Ground");
            dtcDictionary.Add("B2452", "Aux Heater Fuel Pump Circuit Open");
            dtcDictionary.Add("B2453", "Aux Heater Blower Fan Circuit Short to Ground");
            dtcDictionary.Add("B2454", "Aux Heater Blower Fan Circuit Open");
            dtcDictionary.Add("B2455", "Aux Heater Blower Faulted");
            dtcDictionary.Add("B2456", "Aux Heater Coolant Sensor Circuit Short to Ground");
            dtcDictionary.Add("B2457", "Aux Heater Coolant Sensor Circuit Open");
            dtcDictionary.Add("B2458", "Aux Heater Overheat Sensor Circuit Short to Ground");
            dtcDictionary.Add("B2459", "Aux Heater Overheat Sensor Circuit Open");
            dtcDictionary.Add("B2460", "Aux Heater Flame Sensor Circuit Short to Ground");
            dtcDictionary.Add("B2461", "Aux Heater Flame Sensor Circuit Open");
            dtcDictionary.Add("B2462", "Aux Heater Flame Out Fault");
            dtcDictionary.Add("B2463", "Aux Heater Overheat Fault");
            dtcDictionary.Add("B2464", "Aux Heater Start Time Exceeded");
            dtcDictionary.Add("B2465", "Aux Heater Start Counter Overrun/System Locked (same as below ?)");
            dtcDictionary.Add("B2466", "Aux Heater Overheat Counter Overrun/System Locked");
            dtcDictionary.Add("B2467", "Aux Heater Cool Down Time Exceeded (may be con to a453-5 )");
            dtcDictionary.Add("B2468", "Aux Heater Coolant Pump Circuit Short to Ground");
            dtcDictionary.Add("B2469", "Aux Heater Coolant Pump Circuit Open");
            dtcDictionary.Add("B2470", "Interior Fan Control Circuit Short to Ground");
            dtcDictionary.Add("B2471", "Interior Fan Control Circuit Open");
            dtcDictionary.Add("B2472", "Fog Lamp Switch Failure");
            dtcDictionary.Add("B2473", "Passenger Door Disarm Switch ckt Short to Ground");
            dtcDictionary.Add("B2474", "Passenger Door Lock Switch Circuit Shorted to Ground");
            dtcDictionary.Add("B2475", "Passenger Door Unlock Switch Circuit Shorted to Ground");
            dtcDictionary.Add("B2476", "RADIO PRESENT SWITCH Ckt Failure");
            dtcDictionary.Add("B2477", "Module Configuration Failure");
            dtcDictionary.Add("B2478", "ANTI THEFT INPUT SIGNAL SHORT TO GROUND");
            dtcDictionary.Add("B2479", "BRAKE PARK SWITCH CIRCUIT SHORT TO GROUND");
            dtcDictionary.Add("B2480", "LF CORNER LAMP OUTPUT Ckt Short to Battery");
            dtcDictionary.Add("B2481", "Convertible Top Up/Down switch Fault");
            dtcDictionary.Add("B2482", "RF CORNER LAMP OUTPUT Ckt Short to Battery");
            dtcDictionary.Add("B2483", "Enable Signal Open Circuit");
            dtcDictionary.Add("B2484", "Disable Signal Short to Ground");
            dtcDictionary.Add("B2485", "LF SIDE REPEATER LAMP OUTPUT Ckt Short to Battery");
            dtcDictionary.Add("B2487", "RF SIDE REPEATER LAMP OUTPUT Ckt Short to Battery");
            dtcDictionary.Add("B2489", "UNDERHOOD LAMP OUTPUT CIRCUIT Failure");
            dtcDictionary.Add("B2490", "UNDERHOOD LAMP OUTPUT CIRCUIT Short to Battery");
            dtcDictionary.Add("B2491", "RF PARK LAMP OUTPUT CIRCUIT Short to Battery");
            dtcDictionary.Add("B2492", "Already Programmed (Test Mode DTC Only !!!)");
            dtcDictionary.Add("B2493", "LF PARK LAMP OUTPUT CIRCUIT Short to Battery");
            dtcDictionary.Add("B2494", "ANTI THEFT HORN OUTPUT CIRCUIT Short to Batt");
            dtcDictionary.Add("B2495", "ANTI THEFT HORN OUTPUT CIRCUIT Failure");
            dtcDictionary.Add("B2496", "ANTI THEFT HORN OUTPUT CIRCUIT Short to Gnd");
            dtcDictionary.Add("B2499", "COURTESY LAMP OUTPUT Failure");
            dtcDictionary.Add("B2500", "COURTESY LAMP OUTPUT Ckt Short to Battery");
            dtcDictionary.Add("B2501", "LF LAMP LOW BEAM CIRCUIT Failure");
            dtcDictionary.Add("B2502", "LF LAMP LOW BEAM CIRCUIT Short to Battery");
            dtcDictionary.Add("B2503", "RF LAMP LOW BEAM CIRCUIT Failure");
            dtcDictionary.Add("B2504", "RF LAMP LOW BEAM CIRCUIT Short to Battery");
            dtcDictionary.Add("B2505", "LF LAMP HIGH BEAM CIRCUIT Failure");
            dtcDictionary.Add("B2506", "LF LAMP HIGH BEAM CIRCUIT Short to Battery");
            dtcDictionary.Add("B2507", "RF LAMP HIGH BEAM CIRCUIT Failure");
            dtcDictionary.Add("B2508", "RF LAMP HIGH BEAM CIRCUIT Short to Battery");
            dtcDictionary.Add("B2509", "Rear Fog Lamp Switch Circuit Failure");
            dtcDictionary.Add("B2510", "Main Blower Motor Relay Circuit Short to Battery");
            dtcDictionary.Add("B2511", "Horn Output Relay Circuit Short to Battery");
            dtcDictionary.Add("B2512", "Front Fog Lamp Relay Ckt Short to Battery");
            dtcDictionary.Add("B2513", "Blower (Fan) Circuit Failure");
            dtcDictionary.Add("B2514", "Blower (Fan) Circuit Short to Vbatt");
            dtcDictionary.Add("B2515", "Heater Blower Relay Circuit Failure");
            dtcDictionary.Add("B2516", "Blower Control Circuit Failure");
            dtcDictionary.Add("B2517", "Emergency Power Off System Faulted");
            dtcDictionary.Add("B2518", "Compressor Overtemp Fault");
            dtcDictionary.Add("B2519", "High Mount Stop Lamp Circuit Failure");
            dtcDictionary.Add("B2520", "High Mount Stop Lamp Circuit Short to Battery");
            dtcDictionary.Add("B2523", "License Lamp Circuit Failure");
            dtcDictionary.Add("B2524", "License Lamp Circuit Short to Battery");
            dtcDictionary.Add("B2525", "Left Rear Backup Lamp Circuit Failure");
            dtcDictionary.Add("B2526", "Left Rear Backup Lamp Circuit Short to Battery");
            dtcDictionary.Add("B2527", "Left Rear Stop lamp Circuit Failure");
            dtcDictionary.Add("B2528", "Left Rear Stop lamp Circuit Short to Battery");
            dtcDictionary.Add("B2529", "Left Rear Turn Lamp Circuit Failure");
            dtcDictionary.Add("B2530", "Left Rear Turn Lamp Circuit Short to Battery");
            dtcDictionary.Add("B2531", "Right Rear Backup Lamp Circuit Failure");
            dtcDictionary.Add("B2532", "Right Rear Backup Lamp Circuit Short to Battery");
            dtcDictionary.Add("B2533", "Right Rear Stop lamp Circuit Failure");
            dtcDictionary.Add("B2534", "Right Rear Stop lamp Circuit Short to Battery");
            dtcDictionary.Add("B2535", "Right Rear Turn Lamp Circuit Failure");
            dtcDictionary.Add("B2536", "Right Rear Turn Lamp Circuit Short to Battery");
            dtcDictionary.Add("B2539", "Aux A/C Mode Position Reference Circuit Short to Ground");
            dtcDictionary.Add("B2540", "Aux A/C Mode Position Reference Circuit Short to Battery");
            dtcDictionary.Add("B2543", "Aux A/C Control Switch Reference Circuit Short to Ground");
            dtcDictionary.Add("B2544", "Aux A/C Control Switch Reference Circuit Short to Battery");
            dtcDictionary.Add("B2545", "System Power Relay Circuit Short to Battery");
            dtcDictionary.Add("B2546", "System Power Relay Circuit Failure");
            dtcDictionary.Add("B2550", "LAMP DOME OUTPUT Circuit Short to Ground");
            dtcDictionary.Add("B2553", "Disable Signal Output Circuit Short to Battery");
            dtcDictionary.Add("B2554", "LAMP DOME OUTPUT Circuit Failure");
            dtcDictionary.Add("B2555", "LAMP DOME OUTPUT Circuit Short to Battery");
            dtcDictionary.Add("B2556", "Enable Signal Circuit Short to Battery");
            dtcDictionary.Add("B2557", "Left Power Sliding Door Open/Close Output Circuit Short to Battery");
            dtcDictionary.Add("B2558", "Right Power Sliding Door Open/Close Output Circuit Short to Battery");
            dtcDictionary.Add("B2559", "Aux A/C Blower Motor Relay Circuit Short to Battery");
            dtcDictionary.Add("B2560", "Aux A/C Blower Motor Relay Circuit Short to Ground");
            dtcDictionary.Add("B2561", "Aux A/C Blower Speed 1 Circuit Failure");
            dtcDictionary.Add("B2562", "Aux A/C Blower Speed 1 Circuit Short to Ground");
            dtcDictionary.Add("B2563", "Aux A/C Blower Speed 2 Circuit Failure");
            dtcDictionary.Add("B2564", "Aux A/C Blower Speed 2 Circuit Short to Ground");
            dtcDictionary.Add("B2565", "Right Tail Lamp Circuit Failure");
            dtcDictionary.Add("B2566", "Right Tail Lamp Circuit Short to Ground");
            dtcDictionary.Add("B2567", "Reverse Mirror Output Circuit Failure");
            dtcDictionary.Add("B2568", "Reverse Mirror Output Circuit Short to Ground");
            dtcDictionary.Add("B2569", "Liftgate Disarm Switch Circuit Short to Ground");
            dtcDictionary.Add("B2570", "Right Lamp Outage Signal Circuit Short to Ground");
            dtcDictionary.Add("B2571", "Left Lamp Outage Signal Circuit Short to Ground");
            dtcDictionary.Add("B2580", "Aux Blower Sense Switch Circuit Failure");
            dtcDictionary.Add("B2581", "Passenger Seat Occupant Detection Circuit Short To Ground");
            dtcDictionary.Add("B2582", "Passenger Seat Occupant Detection Circuit Open");
            dtcDictionary.Add("B2583", "Child Seat Detection Circuit Short to Ground");
            dtcDictionary.Add("B2584", "Child Seat Detection Circuit Open");
            dtcDictionary.Add("B2585", "Anti Theft Input Signal Circuit Short To Battery");
            dtcDictionary.Add("B2586", "Headlamp Mode Select Circuit Failure");
            dtcDictionary.Add("B2587", "Passenger Seat Occupant Detection Circuit Short To Battery");
            dtcDictionary.Add("B2588", "Child Seat Detection Circuit Short To Battery");
            dtcDictionary.Add("B2589", "Unexpected Door Reversal During Close");
            dtcDictionary.Add("B2590", "Vehicle Park/Speed Signal Circuit Failure");
            dtcDictionary.Add("B2591", "Detent Signal Missing During Unlatch");
            dtcDictionary.Add("B2592", "PSD Not Fully Closed (Module Commanded Successfully)");
            dtcDictionary.Add("B2593", "Power Sliding Door Opened During Module Close Command");
            dtcDictionary.Add("B2594", "No Movement Detected After an Unlatch During Power Open");
            dtcDictionary.Add("B2595", "Anti Theft Input Signal Circuit Failure");
            dtcDictionary.Add("B2596", "Headlamp Aim Output Relay Circuit Short To Battery");
            dtcDictionary.Add("B2597", "Headlamp Aim Output Relay Circuit Failure");
            dtcDictionary.Add("B2598", "Headlamp Relay Circuit Failure");
            dtcDictionary.Add("B2599", "Tailgate Release Open Circuit");
            dtcDictionary.Add("B2600", "Double Locking Door Motor Frozen");
            dtcDictionary.Add("B2601", "No Latch Signal Sensed on Closing and Door Reversed");
            dtcDictionary.Add("B2602", "Missing Latch Signal During Power Sliding Door Unlatch");
            dtcDictionary.Add("B2603", "PSD Not Fully Closed During Self-Test");
            dtcDictionary.Add("B2604", "Power Sliding Door On/Off Switch Open Circuit");
            dtcDictionary.Add("B2605", "Disable Signal Open Circuit");
            dtcDictionary.Add("B2606", "A/C Temperature Sensor Out of Range");
            dtcDictionary.Add("U1000", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1001", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1002", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1003", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1004", "SCP (J1850) Invalid or Missing Data for EEC Programming");
            dtcDictionary.Add("U1005", "SCP (J1850) Invalid or Missing Data for EEC Programming");
            dtcDictionary.Add("U1006", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1007", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1008", "SCP (J1850) Invalid or Missing Data for Engine Torque");
            dtcDictionary.Add("U1009", "SCP (J1850) Invalid or Missing Data for Engine Torque");
            dtcDictionary.Add("U1010", "SCP (J1850) Invalid or Missing Data for Engine Air Intake");
            dtcDictionary.Add("U1011", "SCP (J1850) Invalid or Missing Data for Engine Air Intake");
            dtcDictionary.Add("U1012", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1013", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1014", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1015", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1016", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1017", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1018", "SCP (J1850) Invalid or Missing Data for Throttle");
            dtcDictionary.Add("U1019", "SCP (J1850) Invalid or Missing Data for Throttle");
            dtcDictionary.Add("U1020", "SCP (J1850) Invalid or Missing Data for Air Conditioning Clutch");
            dtcDictionary.Add("U1021", "SCP (J1850) Invalid or Missing Data for Air Conditioning Clutch");
            dtcDictionary.Add("U1022", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1023", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1024", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1025", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1026", "SCP (J1850) Invalid or Missing Data for Engine RPM");
            dtcDictionary.Add("U1027", "SCP (J1850) Invalid or Missing Data for Engine RPM");
            dtcDictionary.Add("U1028", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1029", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1030", "SCP (J1850) Invalid or Missing Data for Experimental #1");
            dtcDictionary.Add("U1031", "SCP (J1850) Invalid or Missing Data for Experimental #1");
            dtcDictionary.Add("U1032", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1033", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1034", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1035", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1036", "SCP (J1850) Invalid or Missing Data for Wheels");
            dtcDictionary.Add("U1037", "SCP (J1850) Invalid or Missing Data for Wheels");
            dtcDictionary.Add("U1038", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1039", "SCP (J1850) Invalid or Missing Data for Primary Id / Vehicle Speed");
            dtcDictionary.Add("U1040", "SCP (J1850) Invalid or Missing Data for Vehicle Speed");
            dtcDictionary.Add("U1041", "SCP (J1850) Invalid or Missing Data for Vehicle Speed");
            dtcDictionary.Add("U1042", "SCP (J1850) Invalid or Missing Data for Traction Control");
            dtcDictionary.Add("U1043", "SCP (J1850) Invalid or Missing Data for Traction Control");
            dtcDictionary.Add("U1044", "SCP (J1850) Invalid or Missing Data for Traction Motor");
            dtcDictionary.Add("U1045", "SCP (J1850) Invalid or Missing Data for Traction Motor");
            dtcDictionary.Add("U1046", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1047", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1048", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1049", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1050", "SCP (J1850) Invalid or Missing Data for Brakes");
            dtcDictionary.Add("U1051", "SCP (J1850) Invalid or Missing Data for Brakes");
            dtcDictionary.Add("U1052", "SCP (J1850) Invalid or Missing Data for Steering / Steering Wheel");
            dtcDictionary.Add("U1053", "SCP (J1850) Invalid or Missing Data for Steering / Steering Wheel");
            dtcDictionary.Add("U1054", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1055", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1056", "SCP (J1850) Invalid or Missing Data for Vehicle Configuration");
            dtcDictionary.Add("U1057", "SCP (J1850) Invalid or Missing Data for Vehicle Configuration");
            dtcDictionary.Add("U1058", "SCP (J1850) Invalid or Missing Data for Transmission / Transaxle / PRNDL");
            dtcDictionary.Add("U1059", "SCP (J1850) Invalid or Missing Data for Transmission / Transaxle / PRNDL");
            dtcDictionary.Add("U1060", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1061", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1062", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1063", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1064", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1065", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1066", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1067", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1068", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1069", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1070", "SCP (J1850) Invalid or Missing Data for Engine Sensors");
            dtcDictionary.Add("U1071", "SCP (J1850) Invalid or Missing Data for Engine Sensors");
            dtcDictionary.Add("U1072", "SCP (J1850) Invalid or Missing Data for Engine Coolant");
            dtcDictionary.Add("U1073", "SCP (J1850) Invalid or Missing Data for Engine Coolant");
            dtcDictionary.Add("U1074", "SCP (J1850) Invalid or Missing Data for Engine Oil");
            dtcDictionary.Add("U1075", "SCP (J1850) Invalid or Missing Data for Engine Oil Temp");
            dtcDictionary.Add("U1076", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1077", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1078", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1079", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1080", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1081", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1082", "SCP (J1850) Invalid or Missing Data for Engine Systems Other");
            dtcDictionary.Add("U1083", "SCP (J1850) Invalid or Missing Data for Engine Systems Other");
            dtcDictionary.Add("U1084", "SCP (J1850) Invalid or Missing Data for Powertrain Status Request");
            dtcDictionary.Add("U1085", "SCP (J1850) Invalid or Missing Data for Powertrain Status Request");
            dtcDictionary.Add("U1086", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1087", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1088", "SCP (J1850) Invalid or Missing Data for Suspension");
            dtcDictionary.Add("U1089", "SCP (J1850) Invalid or Missing Data for Suspension");
            dtcDictionary.Add("U1090", "SCP (J1850) Invalid or Missing Data for Non-Legislated Diagnostics");
            dtcDictionary.Add("U1091", "SCP (J1850) Invalid or Missing Data for Non-Legislated Diagnostics");
            dtcDictionary.Add("U1092", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1093", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1094", "SCP (J1850) Invalid or Missing Data for Experimental #2");
            dtcDictionary.Add("U1095", "SCP (J1850) Invalid or Missing Data for Experimental #2");
            dtcDictionary.Add("U1096", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1097", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1098", "SCP (J1850) Invalid or Missing Data for Vehicle Speed Control");
            dtcDictionary.Add("U1099", "SCP (J1850) Invalid or Missing Data for Vehicle Speed Control");
            dtcDictionary.Add("U1100", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1101", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1102", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1103", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1104", "SCP (J1850) Invalid or Missing Data for Chassis Status Request");
            dtcDictionary.Add("U1105", "SCP (J1850) Invalid or Missing Data for Chassis Status Request");
            dtcDictionary.Add("U1106", "SCP (J1850) Invalid or Missing Data for Legislated Diagnostics");
            dtcDictionary.Add("U1107", "SCP (J1850) Invalid or Missing Data for Legislated Diagnostics");
            dtcDictionary.Add("U1108", "SCP (J1850) Invalid or Missing Data for Electric Traction Drive (Inverter)");
            dtcDictionary.Add("U1109", "SCP (J1850) Invalid or Missing Data for Electric Traction Drive (Inverter)");
            dtcDictionary.Add("U1110", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1111", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1112", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1113", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1114", "SCP (J1850) Invalid or Missing Data for Charging System");
            dtcDictionary.Add("U1115", "SCP (J1850) Invalid or Missing Data for Charging System");
            dtcDictionary.Add("U1116", "SCP (J1850) Invalid or Missing Data for Electrical Energy Management");
            dtcDictionary.Add("U1117", "SCP (J1850) Invalid or Missing Data for Electrical Energy Management");
            dtcDictionary.Add("U1118", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1119", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1120", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1121", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1122", "SCP (J1850) Invalid or Missing Data for Odometer");
            dtcDictionary.Add("U1123", "SCP (J1850) Invalid or Missing Data for Odometer");
            dtcDictionary.Add("U1124", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1125", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1126", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1127", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1128", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1129", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1130", "SCP (J1850) Invalid or Missing Data for Fuel System");
            dtcDictionary.Add("U1131", "SCP (J1850) Invalid or Missing Data for Fuel System");
            dtcDictionary.Add("U1132", "SCP (J1850) Invalid or Missing Data for Vehicle Motion");
            dtcDictionary.Add("U1133", "SCP (J1850) Invalid or Missing Data for Vehicle Motion");
            dtcDictionary.Add("U1134", "SCP (J1850) Invalid or Missing Data for Ignition Switch / Starter");
            dtcDictionary.Add("U1135", "SCP (J1850) Invalid or Missing Data for Ignition Switch / Starter");
            dtcDictionary.Add("U1136", "SCP (J1850) Invalid or Missing Data for Telltales");
            dtcDictionary.Add("U1137", "SCP (J1850) Invalid or Missing Data for Telltales");
            dtcDictionary.Add("U1138", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1139", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1140", "SCP (J1850) Invalid or Missing Data for Gateway");
            dtcDictionary.Add("U1141", "SCP (J1850) Invalid or Missing Data for Gateway");
            dtcDictionary.Add("U1142", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1143", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1144", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1145", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1146", "SCP (J1850) Invalid or Missing Data for Vehicle Security");
            dtcDictionary.Add("U1147", "SCP (J1850) Invalid or Missing Data for Vehicle Security");
            dtcDictionary.Add("U1148", "SCP (J1850) Invalid or Missing Data for Audio Control");
            dtcDictionary.Add("U1149", "SCP (J1850) Invalid or Missing Data for Audio Control");
            dtcDictionary.Add("U1150", "SCP (J1850) Invalid or Missing Data for Audible Warnings");
            dtcDictionary.Add("U1151", "SCP (J1850) Invalid or Missing Data for Audible Warnings");
            dtcDictionary.Add("U1152", "SCP (J1850) Invalid or Missing Data for Experimental #3");
            dtcDictionary.Add("U1153", "SCP (J1850) Invalid or Missing Data for Experimental #3");
            dtcDictionary.Add("U1154", "SCP (J1850) Invalid or Missing Data for Compact Disc");
            dtcDictionary.Add("U1155", "SCP (J1850) Invalid or Missing Data for Compact Disc");
            dtcDictionary.Add("U1156", "SCP (J1850) Invalid or Missing Data for Digital Signal Processing");
            dtcDictionary.Add("U1157", "SCP (J1850) Invalid or Missing Data for Digital Signal Processing");
            dtcDictionary.Add("U1158", "SCP (J1850) Invalid or Missing Data for Antenna");
            dtcDictionary.Add("U1159", "SCP (J1850) Invalid or Missing Data for Antenna");
            dtcDictionary.Add("U1160", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1161", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1162", "SCP (J1850) Invalid or Missing Data for Digital Audio Tape");
            dtcDictionary.Add("U1163", "SCP (J1850) Invalid or Missing Data for Digital Audio Tape");
            dtcDictionary.Add("U1164", "SCP (J1850) Invalid or Missing Data for Tuner / Receiver");
            dtcDictionary.Add("U1165", "SCP (J1850) Invalid or Missing Data for Tuner / Receiver");
            dtcDictionary.Add("U1166", "SCP (J1850) Invalid or Missing Data for Cassette Tape");
            dtcDictionary.Add("U1167", "SCP (J1850) Invalid or Missing Data for Cassette Tape");
            dtcDictionary.Add("U1168", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1169", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1170", "SCP (J1850) Invalid or Missing Data for Cellular Phone / Paging System");
            dtcDictionary.Add("U1171", "SCP (J1850) Invalid or Missing Data for Cellular Phone / Paging System");
            dtcDictionary.Add("U1172", "SCP (J1850) Invalid or Missing Data for Remote Button Control");
            dtcDictionary.Add("U1173", "SCP (J1850) Invalid or Missing Data for Remote Button Control");
            dtcDictionary.Add("U1174", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1175", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1176", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1177", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1178", "SCP (J1850) Invalid or Missing Data for Climate Control (HVAC)");
            dtcDictionary.Add("U1179", "SCP (J1850) Invalid or Missing Data for Climate Control (HVAC)");
            dtcDictionary.Add("U1180", "SCP (J1850) Invalid or Missing Data for Personalization (Memory) Features");
            dtcDictionary.Add("U1181", "SCP (J1850) Invalid or Missing Data for Personalization (Memory) Features");
            dtcDictionary.Add("U1182", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1183", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1184", "SCP (J1850) Invalid or Missing Data for Window Wiper / Washer");
            dtcDictionary.Add("U1185", "SCP (J1850) Invalid or Missing Data for Window Wiper / Washer");
            dtcDictionary.Add("U1186", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1187", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1188", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1189", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1190", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1191", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1192", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1193", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1194", "SCP (J1850) Invalid or Missing Data for Mirrors");
            dtcDictionary.Add("U1195", "SCP (J1850) Invalid or Missing Data for Mirrors");
            dtcDictionary.Add("U1196", "SCP (J1850) Invalid or Missing Data for Door Locks");
            dtcDictionary.Add("U1197", "SCP (J1850) Invalid or Missing Data for Door Locks");
            dtcDictionary.Add("U1198", "SCP (J1850) Invalid or Missing Data for External Access (Doors)");
            dtcDictionary.Add("U1199", "SCP (J1850) Invalid or Missing Data for External Access (Doors)");
            dtcDictionary.Add("U1200", "SCP (J1850) Invalid or Missing Data for Seat Motion / Control");
            dtcDictionary.Add("U1201", "SCP (J1850) Invalid or Missing Data for Seat Motion / Control");
            dtcDictionary.Add("U1202", "SCP (J1850) Invalid or Missing Data for Windows");
            dtcDictionary.Add("U1203", "SCP (J1850) Invalid or Missing Data for Windows");
            dtcDictionary.Add("U1204", "SCP (J1850) Invalid or Missing Data for Steering Column");
            dtcDictionary.Add("U1205", "SCP (J1850) Invalid or Missing Data for Steering Column");
            dtcDictionary.Add("U1206", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1207", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1208", "SCP (J1850) Invalid or Missing Data for Seat Switches");
            dtcDictionary.Add("U1209", "SCP (J1850) Invalid or Missing Data for Seat Switches");
            dtcDictionary.Add("U1210", "SCP (J1850) Invalid or Missing Data for Restraints");
            dtcDictionary.Add("U1211", "SCP (J1850) Invalid or Missing Data for Restraints");
            dtcDictionary.Add("U1212", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1213", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1214", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1215", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1216", "SCP (J1850) Invalid or Missing Data for External Lamp Outage");
            dtcDictionary.Add("U1217", "SCP (J1850) Invalid or Missing Data for External Lamp Outage");
            dtcDictionary.Add("U1218", "SCP (J1850) Invalid or Missing Data for External Lamps");
            dtcDictionary.Add("U1219", "SCP (J1850) Invalid or Missing Data for External Lamps");
            dtcDictionary.Add("U1220", "SCP (J1850) Invalid or Missing Data for Interior Lamp Outage");
            dtcDictionary.Add("U1221", "SCP (J1850) Invalid or Missing Data for Interior Lamp Outage");
            dtcDictionary.Add("U1222", "SCP (J1850) Invalid or Missing Data for Interior Lamps");
            dtcDictionary.Add("U1223", "SCP (J1850) Invalid or Missing Data for Interior Lamps");
            dtcDictionary.Add("U1224", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1225", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1226", "SCP (J1850) Invalid or Missing Data for Body Status Request");
            dtcDictionary.Add("U1227", "SCP (J1850) Invalid or Missing Data for Body Status Request");
            dtcDictionary.Add("U1228", "SCP (J1850) Invalid or Missing Data for Tires");
            dtcDictionary.Add("U1229", "SCP (J1850) Invalid or Missing Data for Tires");
            dtcDictionary.Add("U1230", "SCP (J1850) Invalid or Missing Data for Electric Defrost");
            dtcDictionary.Add("U1231", "SCP (J1850) Invalid or Missing Data for Electric Defrost");
            dtcDictionary.Add("U1232", "SCP (J1850) Invalid or Missing Data for Navigation");
            dtcDictionary.Add("U1233", "SCP (J1850) Invalid or Missing Data for Navigation");
            dtcDictionary.Add("U1234", "SCP (J1850) Invalid or Missing Data for Displays");
            dtcDictionary.Add("U1235", "SCP (J1850) Invalid or Missing Data for Displays");
            dtcDictionary.Add("U1236", "SCP (J1850) Invalid or Missing Data for Memory Storage");
            dtcDictionary.Add("U1237", "SCP (J1850) Invalid or Missing Data for Memory Storage");
            dtcDictionary.Add("U1238", "SCP (J1850) Invalid or Missing Data for Experimental #4");
            dtcDictionary.Add("U1239", "SCP (J1850) Invalid or Missing Data for Experimental #4");
            dtcDictionary.Add("U1240", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1241", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1242", "SCP (J1850) Invalid or Missing Data for Exterior Environment");
            dtcDictionary.Add("U1243", "SCP (J1850) Invalid or Missing Data for Exterior Environment");
            dtcDictionary.Add("U1244", "SCP (J1850) Invalid or Missing Data for Interior Environment");
            dtcDictionary.Add("U1245", "SCP (J1850) Invalid or Missing Data for Interior Environment");
            dtcDictionary.Add("U1246", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1247", "SCP (J1850) Invalid or Missing Data for Primary Id");
            dtcDictionary.Add("U1248", "SCP (J1850) Invalid or Missing Data for Time / Date");
            dtcDictionary.Add("U1249", "SCP (J1850) Invalid or Missing Data for Time / Date");
            dtcDictionary.Add("U1250", "SCP (J1850) Invalid or Missing Data for Vehicle Id (VIN)");
            dtcDictionary.Add("U1251", "SCP (J1850) Invalid or Missing Data for Vehicle Id (VIN)");
            dtcDictionary.Add("U1252", "SCP (J1850) Invalid or Missing Data for Class A Functions");
            dtcDictionary.Add("U1253", "SCP (J1850) Invalid or Missing Data for Class A Functions");
            dtcDictionary.Add("U1254", "SCP (J1850) Invalid or Missing Data for Network Control");
            dtcDictionary.Add("U1255", "SCP (J1850) Invalid or Missing Data for Network Control");
            dtcDictionary.Add("U1260", "SCP (J1850) Single Ended (+) Circuit Failure");
            dtcDictionary.Add("U1261", "SCP (J1850) Single Ended (-) Circuit Failure");
            dtcDictionary.Add("U1262", "SCP (J1850) Communication Bus Fault");
            dtcDictionary.Add("U1308", "SCP (J1850) Invalid or Missing Data for Function Read Engine Torque");
            dtcDictionary.Add("U1341", "SCP (J1850) Invalid or Missing Data for Function Read Vehicle Speed");
            dtcDictionary.Add("U1430", "SCP (J1850) Invalid or Missing Data for Function Read Fuel System");
            dtcDictionary.Add("U1451", "SCP (J1850) Invalid or Missing Data for Function Read Audible Warnings / Anti-Theft Module, Vehicle Immobilized");
            dtcDictionary.Add("U1612", "SCP (J1850) Lack of Acknowledgment for Primary Id");
            dtcDictionary.Add("U1736", "SCP (J1850) Lack of Acknowledgment for Telltales");
            dtcDictionary.Add("U1750", "SCP (J1850) Lack of Acknowledgment for Audible Warnings");
            dtcDictionary.Add("U1794", "SCP (J1850) Lack of Acknowledgment for Mirrors");
            dtcDictionary.Add("U1797", "SCP (J1850) Lack of Acknowledgment for Door Locks");
            dtcDictionary.Add("U1798", "SCP (J1850) Lack of Acknowledgment for External Access (Doors)");
            dtcDictionary.Add("U1806", "SCP (J1850) Lack of Acknowledgment for Primary Id");
            dtcDictionary.Add("U1900", "CAN Communication Bus Fault");
            dtcDictionary.Add("U1950", "UPB Communication Bus Fault");
            dtcDictionary.Add("U2000", "Audio Rear Control Unit is Not Responding");
            dtcDictionary.Add("U2001", "Audio Tape Deck Unit is Not Responding");
            dtcDictionary.Add("U2002", "Audio Bezel is Not Responding");
            dtcDictionary.Add("U2003", "Audio Compact Disk / Disk Jockey Unit is Not Responding");
            dtcDictionary.Add("U2004", "Audio Steering Wheel Control Unit is Not Responding");
            dtcDictionary.Add("U2005", "Audio Rear Integrated Control Panel Unit is Not Responding");
            dtcDictionary.Add("U2006", "Audio Remote Climate Control Unit is Not Responding");
            dtcDictionary.Add("U2007", "Audio Navigation Unit is Not Responding");
            dtcDictionary.Add("U2008", "Audio Phone is Not Responding");
            dtcDictionary.Add("U2009", "Audio Front Control Module (ACM) is Not Responding");
            dtcDictionary.Add("U2010", "Module is Not Responding (Non SCP)");
            dtcDictionary.Add("U2011", "Module Transmitted Invalid Data (Non SCP)");
            dtcDictionary.Add("U2012", "Communication Bus Error (Non SCP)");
            dtcDictionary.Add("U2013", "Compass Module is not Responding");
            dtcDictionary.Add("U2014", "Audio Subwoofer Unit is Not Responding");
            dtcDictionary.Add("U2015", "Signal Link Fault (Non SCP)");
            dtcDictionary.Add("U2016", "Signal Link Short to Ground (Non SCP) / From NGV Module");
            dtcDictionary.Add("U2017", "Driver Side Crash Sensor Communication Fault (Non SCP)");
            dtcDictionary.Add("U2018", "Passenger Side Crash Sensor Communication Fault (Non SCP)");
            dtcDictionary.Add("U2019", "Audio Voice Module Not Responding");
            dtcDictionary.Add("U2020", "Audio Center Amp is not responding");
            dtcDictionary.Add("U2021", "Invalid /fault data received (Non SCP)");
            dtcDictionary.Add("U2150", "SCP (J1850) Invalid Data from REM");
            dtcDictionary.Add("U2152", "SCP (J1850) Invalid Data from GEM");
            dtcDictionary.Add("U2160", "SCP (J1850) Invalid Data from IC");
            dtcDictionary.Add("U2195", "SCP (J1850) Invalid Data from SCLM");
            dtcDictionary.Add("U2500", "(CAN) Lack of Acknowledgement From Engine Management");
        } // InitDtcDescriptions()

        /// <summary>
        /// Gets the description for the specified DTC.
        /// </summary>
        /// <param name="dtc">The DTC.</param>
        /// <returns>A description text for the specified DTC.</returns>
        public static string GetDtcDescription(string dtc)
        {
            if (dtcDictionary == null)
            {
                InitDtcDescriptions();
            } // if

            if ((dtcDictionary != null) && dtcDictionary.ContainsKey(dtc))
            {
                return dtcDictionary[dtc];
            } // if

            // fallback
            if (dtc.StartsWith("P00"))
            {
                return "Fuel and air metering and auxiliary emission controls";
            } // if

            if (dtc.StartsWith("P01"))
            {
                return "Fuel and air metering.";
            } // if

            if (dtc.StartsWith("P02"))
            {
                return "Fuel and air metering (injector circuit)";
            } // if

            if (dtc.StartsWith("P03"))
            {
                return "Ignition system or misfire";
            } // if

            if (dtc.StartsWith("P04"))
            {
                return "Auxiliary emissions controls";
            } // if

            if (dtc.StartsWith("P05"))
            {
                return "Vehicle speed controls and idle control system";
            } // if

            if (dtc.StartsWith("P06"))
            {
                return "Computer output circuit";
            } // if

            if (dtc.StartsWith("P07"))
            {
                return "Transmission";
            } // if

            if (dtc.StartsWith("P08"))
            {
                return "Transmission";
            } // if

            return "N/A";
        } // GetDtcDescription()
    } // DtcDatabase
} // Tethys.OBD
